#include "at_process_frame.h"
#include "lk_error_M.h"
#include "lk_error_STB.h"
#include "track_parameters.h"
#include "tf_errors.h"
#include <malloc.h>
#include <jsfile.h>
#include <tf_kalman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <cimage.h>
#include <ipng.h>
#include <convert.h>
#include <png.h>

options_t at_get_options(int argc, char **argv);
/* Parses the command line options and returns them as an {options_t}. */

void at_print_options(FILE *f, options_t o);

image_t at_read_frame(char *frame_prefix, int frame_num, char *image_type);
/* Reads a frame from image "{frame_prefix}/{NNNNN}/image.{image_type}" where {NNNNN} is the 
  {frame_num} number typeset as 5 digits. */

void at_read_p_i_kalman_models
  ( char *kalman_dir, 
    int n_camera_params, 
    int max_kalman_order, 
    kalman_parameters_t ***K );
/* Allocates an array {*K} with {max_kalman_order+1} rows and 2
  columns.  Then reads into {K[r][axis]} the Kalman model of order {r}
  for image coordinate {axis} (X or Y) of the true position of a
  generic mark.  The file names should be
  "{kalman_dir}/{name}_{NN}.kpar" where {name} id "x_i" or "Y_i", and
  {NN} is the model's order in 2 digits.  */

void at_read_cpar_kalman_models
  ( char *kalman_dir, 
    int n_camera_params, 
    int max_kalman_order, 
    kalman_parameters_t ***K );
/* Allocates an array {*K} with {max_kalman_order+1} rows 
  and {n_camera_params} columns.  Then reads into {K[r][ip]} the Kalman model of order {r} for each 
  calibrated camera parameter number {ip} in {0 .. n_camera_params-1} are the camera parameters as 
  defined by {tf_camera_get_param}.  The file names should be "{kalman_dir}/{name}_{NN}.kpar"
  where {name} is the parameter's name, as defined bu {tf_camera_param_name}, and {NN}
  is the model's order in 2 digits. */

kalman_parameters_t at_read_kalman_model(char *dir, char *pname, int order);
/* Reads a Kalman model from file  "{dir}/{pname}_{NN}.kpar" where
   {NN} is the {order} parameter in 2 digits. */

kalman_parameters_t **alloc_kalman_model_array(int m, int n);
/* Allocates an array of {kalman_parameters_t} with {m} rows and {n} columns. */

r2_t *at_copy_mark_positions(int nmarks, r2_t *p_i);
/* Allocates a new vector of mark positions and copies {p_i[0..nmarks-1]} into it. */

double *at_copy_mark_weights(int nmarks, double *p_wgt);
/* Allocates a new vector of mark weights and copies {p_wgt[0..nmarks-1]} into it. */

int main (int argc, char *argv[])
{
  options_t o = at_get_options(argc, argv);

  /* Frame position of the video sequence: */
  int frame_num; 

  /* Number of camera parameters: */
  int n_camera_params = tf_camera_num_parameters;

  mkdir(o->out_dir, 0777); 

  /* Read the target world positions and first-frame image positions: */
  targets_data_t tdat = tf_targets_data_read(o->p_w_fname, o->tuv_w_fname, o->p_i_fname, o->p_wgt_fname);
  int ntargets = tdat->ntargets;

  image_list_t M_list = NULL; /* Mark images. */
  image_list_t W_list = NULL; /* Weight masks for mark images. */
									      
  if (o->useSyntheticMarks) {
    /* Read the ideal mark images and create their weight masks: */
    image_list_t T_list = image_list_read (o->n_mark_types, o->mark_fname_list);
    M_list = image_list_copy_selected (T_list, ntargets, tdat->type);
    image_list_free(T_list);
    /* Create masks (uniform for STB error, Gaussian otherwise): */
    W_list = image_list_create_weight_masks (M_list, o->useGaussianWeights);
  }

  /* Kalman models for all camera parameters and various orders: */
  kalman_parameters_t **K_p_i = NULL;
  at_read_p_i_kalman_models(o->kalman_dir, n_camera_params, o->max_kalman_order_pos, &K_p_i);
  kalman_parameters_t **K_cpar = NULL;
  at_read_cpar_kalman_models(o->kalman_dir, n_camera_params, o->max_kalman_order_cpar, &K_cpar);

  /* Final camera parameters in the previous frames */
  camera_parameters_t prev_cpar[o->max_kalman_order_cpar];
  int ind;
  for (ind = 0; ind < o->max_kalman_order_cpar; ind++) { prev_cpar[ind] = NULL; }

  /* Final mark positions and weights in the previous frames */
  r2_t *prev_p_i[o->max_kalman_order_pos];
  double *prev_p_wgt[o->max_kalman_order_pos];
  for (ind = 0; ind < o->max_kalman_order_pos; ind++) { prev_p_i[ind] = NULL; prev_p_wgt[ind] = NULL; }

  /* Mark positions computed from the initial-guess {cpar}: */
  r2_t *p_i_cpar = (r2_t *)notnull(malloc(ntargets * sizeof(r2_t)), "no mem");

  /*for each image of the video sequence*/
  for (frame_num = o->first_frame; frame_num <= o->last_frame; frame_num++) {

    fprintf(stderr, "======================================================================\n");
    fprintf(stderr, "Processing frame %05d\n", frame_num);

    image_t I = at_read_frame(o->frame_prefix, frame_num, "pgm"); 

    /* Choose the directory name for all output files of this frame: */
    char *frame_out_dir = NULL;
    asprintf(&frame_out_dir, "%s/%05d", o->out_dir, frame_num);
    /* Creating the frame directory */
    fprintf(stderr, "nome do dir:%s\n", frame_out_dir);
    assert(mkdir(frame_out_dir, 0777)==0);

    /* Step 1 - obtain a guessed {cpar} and {cspec} for this frame. */
    camera_specs_t cspec = tf_camera_copy_specs(o->global_cspec);
    camera_parameters_t cpar = tf_get_mean_camera_parameters(o->global_cspec);
    at_guess_cpar_for_frame
      ( o, I, tdat, prev_cpar, K_cpar, cpar, cspec, frame_num, frame_out_dir );
    
    /* Step 2 - obtain guessed positions and shapes for all marks for this frame. */
    r2_t p_i_dev_finder;  /* Exepcted deviation from guessed mark positions to found ones. */
    at_guess_target_positions_and_shapes
      ( o, 
        cpar, 
        ntargets,
        tdat->p_i,   /* (in) user-given, or final pos of previous frame. */
	tdat->conf,  /* (in) user-given confidence weights, or from prev frame. */
        prev_p_i,    /* Final image positions on previous frames. */
	prev_p_wgt,  /* Mark confidence weights of previous frames. */
	K_p_i,       /* Kalman models for mark image positions. */
        tdat->p_w,
        tdat->u_w,
        tdat->v_w,
        p_i_cpar,        /* (out) mark positions computed from {cpar}. */
        tdat->p_i,       /* (out) guessed pos for this frame. */
	&p_i_dev_finder, /* (out) expected error in the guessed pos. */
        tdat->conf,      /* (out) confidence weights for guessed pos. */
        tdat->u_i,
        tdat->v_i,
	frame_num
      );

    /* Grab the canonical target images {M_list,W_list} from this frame, if so requested: */
    if ((frame_num == o->first_frame) && (! o->useSyntheticMarks)) {
      /* Grab the mark images from the first frame: */
      fprintf(stderr, "@@ Grabbing the canonical target images {M,W} from the first frame.\n");
      M_list = image_list_alloc (ntargets);
      at_extract_feature_images (I, tdat, o->QSubsampling, M_list);
      /* Create masks (uniform for STB error, Gaussian otherwise): */
      W_list = image_list_create_weight_masks (M_list, o->useGaussianWeights);
    }
    else {
      /* Use the canonical images from before: */
      fprintf(stderr, "@@ Preserving the canonical images {M,W}.\n");
    }

    /* Step 3 - Obtain the actual target images {G_lis,V_list} to look for in the frame: */
    fprintf(stderr, "@@ computing mark positions and shape from camera params\n");
    image_list_t G_list;
    image_list_t V_list;
    at_compute_target_images_and_masks
      ( M_list,
        W_list,
        o,
        tdat->u_i,
        tdat->v_i,
        o->useSyntheticMarks,
        &G_list,
        &V_list,
        frame_out_dir );

    /* Write images showing the target positions implied by the guessed {cpar}: */
    at_write_debugging_images
      (I, 1, ntargets, p_i_cpar, G_list, V_list, o->QSubsampling, NULL, NULL, frame_out_dir, "I_cpar_guess" );

    /* Decide which {alpha,beta} parameters to use: */
    /* We should use Kalman models for {alpha} and {beta}. */
    int est_alpha_beta_before = (frame_num == o->first_frame);
    if (est_alpha_beta_before) { 
      fprintf(stderr, "@@ Estimating {alpha,beta} by comparing target with image at guessed positions\n");
      at_estimate_alphas_and_betas
        ( I,
  	  G_list,
  	  V_list,
  	  o,
  	  tdat->p_i,
          tdat->conf,
          DEBUG_ALPHA_BETA,
  	  tdat->alpha,
  	  tdat->beta );
    }
    else {
      fprintf(stderr, "@@ Using the {alpha,beta} given in {tdat}\n");
    }


    if (WRITE_GUESS_POINTS) {
      at_mark_positions_and_weights_write (ntargets, tdat->p_i, tdat->conf, frame_out_dir, "guess"); 
    }

    if (WRITE_ALPHA_BETA) {
      at_alpha_beta_write (ntargets, tdat->alpha, tdat->beta, frame_out_dir, "used"); 
    }
  
    /* Locate the marks in the frame: */
    /* Guesses are {tdat->p_i,tdat->conf}, result will be in {p_i_found,p_wgt_found}: */
    /* Obtain also the a priori uncertainties {p_i_dev_calib}: */
    fprintf(stderr, "@@ Locating the targets in the frame image\n");
    r2_t *p_i_found = (r2_t *)malloc(ntargets*sizeof(r2_t));
    double *p_wgt_found = (double *)malloc(ntargets*sizeof(double));
    r2_t p_i_dev_calib;  /* Expected deviation from found mark positions to the true ones. */
    at_locate_targets_in_frame
      ( I,
	G_list,
	V_list,
	tdat->alpha,
	tdat->beta,
	o,
	tdat->p_i,
	p_i_dev_finder,
	tdat->conf,
	p_i_found,
	&p_i_dev_calib,
	p_wgt_found,
	frame_out_dir );

    if (WRITE_FOUND_POINTS) {
      at_mark_positions_and_weights_write (ntargets, p_i_found, p_wgt_found, frame_out_dir, "found"); 
    }
 
    /* Calibrate camera using the positions {p_i_found}: */
    at_calibrate_camera_for_frame
      ( o,
	cspec,
	ntargets,
	tdat->p_w,
	p_i_found,
	p_i_dev_calib,
	p_wgt_found,
	cpar,
	frame_out_dir,
	frame_num );

    /* Recompute mark positions {p_i_calib} and shapes {tdat->u_i,tdat->v_i} from given {cpar}: */
    r2_t *p_i_calib = (r2_t *)malloc(ntargets*sizeof(r2_t));
    fprintf(stderr, "@@ Recomputing the mark positions from the camera parameters\n");
    int mark;
    for (mark = 0; mark < ntargets; mark++) {
      tf_mark_estimate_position_and_shape 
	( cpar,
	  tdat->p_w[mark], tdat->u_w[mark], tdat->v_w[mark],
	  &(p_i_calib[mark]), &(tdat->u_i[mark]), &(tdat->v_i[mark])
	  );
    }

    /* Write images showing the target positions implied by calibrated {cpar}: */
    at_write_debugging_images
      (I, 1, ntargets, p_i_calib, G_list, V_list, o->QSubsampling, NULL, NULL, frame_out_dir, "I_cpar_calib" );

    if (WRITE_CALIB_POINTS) {
      /* !!! Should set 0 weight for marks that fall outside the image. */
      at_mark_positions_and_weights_write (ntargets, p_i_calib, p_wgt_found, frame_out_dir, "calib"); 
    }
 
    /* Decide which positions are the `official' mark positions for this frame: */
    if (o->useComputedPositions) {
      fprintf(stderr, "@@ Setting the computed mark positions as the official positions for this frame\n");
      for (mark = 0; mark < ntargets; mark++) { 
        tdat->p_i[mark] = p_i_calib[mark];
        bool_t inside = tf_camera_point_is_inside_image(p_i_calib[mark], cpar);
        tdat->conf[mark] = (inside ? 1.0 : 0.0);
      }
    }
    else {
      fprintf(stderr, "@@ Setting the found mark positions as the official positions for this frame\n");
      for (mark = 0; mark < ntargets; mark++) { 
	tdat->p_i[mark] = p_i_found[mark];
	tdat->conf[mark] = p_wgt_found[mark];
      }
    }

    /* Write images showing the final offcial target positions in{tdat}: */
    at_write_debugging_images
      (I, 1, ntargets, tdat->p_i, G_list, V_list, o->QSubsampling, NULL, NULL, frame_out_dir, "I_final" );

    /*disallocating variables and structures*/
    free(p_i_calib);
    free(p_i_found);
    free(p_wgt_found);
  
    /* Update the canonical target images as appropriate. */
    if ((! o->useSyntheticMarks) && (! o->useMarksFromFirstFrame)) {
      fprintf(stderr, "@@ Grabbing the canonical target images from the current frame\n");
      /* Update the mark images: */
      assert(M_list != NULL);
      at_extract_feature_images (I, tdat, o->QSubsampling, M_list);
      /* Re-create masks (uniform for STB error, Gaussian otherwise): */
      assert(W_list != NULL);
      image_list_free(W_list);
      W_list = image_list_create_weight_masks (M_list, o->useGaussianWeights);
    }
  
    /* Update the {alpha,beta} parameters as appropriate. */
    /* We should use Kalman models for {alpha} and {beta}. */
    int est_alpha_beta_after = o->adaptiveAlphaBeta || (frame_num == o->first_frame);
    if (est_alpha_beta_after) {     
      /* Re-estimate {alpha,beta} by comparing mark with image at adjusted positions: */   
      fprintf(stderr, "@@ re-estimating mark colors from current positions\n");
      at_estimate_alphas_and_betas
        ( I,
  	  G_list,
  	  V_list,
  	  o,
  	  tdat->p_i,
          tdat->conf,
  	  DEBUG_ALPHA_BETA,
          tdat->alpha,
  	  tdat->beta );
    }
  
    /* Save the final {cpar} for Kalman predictions: */
    if (prev_cpar[o->max_kalman_order_cpar-1] != NULL) { free(prev_cpar[o->max_kalman_order_cpar-1]); }
    for (ind = (o->max_kalman_order_cpar-1); ind > 0; ind--) {
      prev_cpar[ind] = prev_cpar[ind-1];
    }
    prev_cpar[0] = cpar;

    /* Save final mark positions and confidences for Kalman predictions: */
    if (prev_p_i[o->max_kalman_order_pos-1] != NULL) { free(prev_p_i[o->max_kalman_order_pos-1]); }
    if (prev_p_wgt[o->max_kalman_order_pos-1] != NULL) { free(prev_p_wgt[o->max_kalman_order_pos-1]); }
    for (ind = (o->max_kalman_order_pos-1); ind > 0; ind--) {
      prev_p_i[ind] = prev_p_i[ind-1];
      prev_p_wgt[ind] = prev_p_wgt[ind-1];
    }
    prev_p_i[0] = at_copy_mark_positions(ntargets, tdat->p_i);
    prev_p_wgt[0] = at_copy_mark_weights(ntargets, tdat->conf);
     
    image_list_free(G_list);
    image_list_free(V_list);
   
    free(frame_out_dir);
    image_free (I);     
    free(cspec);
  }

  /*disallocating structures*/
  tf_targets_data_free(tdat);
  free(p_i_cpar);
 
  if (M_list != NULL) { image_list_free(M_list); }
  if (W_list != NULL) { image_list_free(W_list); }

  return 0;
}

kalman_parameters_t **alloc_kalman_model_array(int m, int n)
{
  kalman_parameters_t **K = (kalman_parameters_t **)notnull(malloc(m * sizeof(kalman_parameters_t *)), "no mem");
   int i;
   for (i = 0; i < m; i++) {
     K[i] = (kalman_parameters_t *)notnull(malloc(n * sizeof(kalman_parameters_t)), "no mem");
   }
   return K;
}

void at_read_p_i_kalman_models
  ( char *kalman_dir, 
    int n_camera_params, 
    int max_kalman_order, 
    kalman_parameters_t ***K )
{   
  int m = max_kalman_order+1;
  int n = 2;
  (*K) = alloc_kalman_model_array(m, n);
  int axis;
  for (axis = 0; axis < n; axis++) {
    int order;
    for (order = 0; order < m; order++) {
      char *pname = (axis == 0 ? "x_i" : "y_i"); 
      (*K)[order][axis] = at_read_kalman_model(kalman_dir, pname, order);
    }
  }
}

void at_read_cpar_kalman_models
  ( char *kalman_dir, 
    int n_camera_params, 
    int max_kalman_order, 
    kalman_parameters_t ***K )
{   
  int m = max_kalman_order+1;
  int n = n_camera_params;
  (*K) = alloc_kalman_model_array(m, n);
  int ip;
  for (ip = 0; ip < n; ip++) {
    int order;
    for (order = 0; order < m; order++) {
      char *pname = tf_camera_param_name(ip);
      (*K)[order][ip] = at_read_kalman_model(kalman_dir, pname, order);
    }
  }
}

kalman_parameters_t at_read_kalman_model(char *dir, char *pname, int order)
{
      char *file_name_kalman = NULL;
      asprintf(&file_name_kalman, "%s/%s_%02d.kpar", dir, pname, order);
      FILE *file_kalman = open_read(file_name_kalman, TRUE);
      free(file_name_kalman);
      kalman_parameters_t K =  tf_kalman_read_parameters (file_kalman);
      fprintf(stderr, "%-5s ", pname);
      tf_kalman_show_parameters (K, stderr);
      fclose(file_kalman);
      return K;
}

image_t at_read_frame(char *frame_prefix, int frame_num, char *image_type) 
{
    /* Input the next frame {I} of the video: */
    char *frame_fname = NULL;
    asprintf(&frame_fname, "%s/%05d/frame.%s", frame_prefix, frame_num, image_type);

    image_t I;
    if (strcmp (image_type, "png") == 0) {
       cimage_t cimg = read_png (frame_fname);
       convert_RGB_to_YCbCr (cimg);
       I = convert_Y_cimg_to_img (cimg);
    }  
    else if (strcmp (image_type, "pgm") == 0) {
       I = image_read_pgm(frame_fname);
    }
    else {
       fprintf(stderr, "error: unknown extension = %s\n", image_type);
    }
    
    free(frame_fname);         
    return I;
}

options_t at_get_options(int argc, char **argv) 
{
  options_t o = (options_t)malloc(sizeof(struct _options_t));
  o->useSyntheticMarks = 1;
  o->useMarksFromFirstFrame = 0;
  o->redFactor = 2;
  o->QSubsampling = 1;
  o->niter = 30;
  o->markBlur = 0.0;
  o->minMarkSize = 5;
  o->useWeightsInCalib = 1;
  o->adaptiveAlphaBeta = 1; 
  o->useComputedPositions = 1;
  o->p_w_fname = argv[1];
  o->tuv_w_fname = argv[2];
  o->p_i_fname = argv[3];
  o->p_wgt_fname = argv[4];
  o->frame_prefix = argv[5];
  o->first_frame = atoi(argv[6]);
  o->last_frame = atoi(argv[7]);
  /* Kalman parameters: */
  o->max_kalman_order_pos = atoi(argv[8]);
  o->max_kalman_order_cpar = atoi(argv[9]);
  o->extrapolateMarkPositions = atoi(argv[10]);
  o->useKalmanErrorInFinder = atoi(argv[11]);
  o->useKalmanErrorInCalib = atoi(argv[12]);
  /* Calibration parameters: */
  o->useInternalGuessInCalib = atoi(argv[13]);
  o->twoPassCalib = atoi(argv[14]);
  /* Finder parameters: */
  o->efunc_name = argv[15];
  o->efunc = parse_error_function_arg(o->efunc_name);
  o->minize_name = argv[16];
  o->minize = parse_minimizer_arg(o->minize_name);
  o->step_factor = atof(argv[17]);
  char *camera_model_name = argv[18];
  if (strcmp(camera_model_name,"optura") == 0) {
      o->global_cspec = tf_get_canon_optura_specs ();
  }
  else if (strcmp(camera_model_name,"svga") == 0) {
      o->global_cspec = tf_get_svga_povray_specs (); 
  }
  else if (strcmp(camera_model_name,"sony") == 0) {
      o->global_cspec = tf_get_sony_dv40_specs (); 
  }
  else if (strcmp(camera_model_name,"webcam") == 0) {
      o->global_cspec = tf_get_svga_povray_webcam_specs ();
  }    
  else if (strcmp(camera_model_name,"hvga") == 0) {
      o->global_cspec = tf_get_hvga_povray_specs ();
  }
  else if (strcmp(camera_model_name,"hvga-dist") == 0) {
      o->global_cspec = tf_get_hvga_with_distortion_povray_specs ();
  }
  else {
     fprintf(stderr, "error: define the camera specifications\n");
     exit(1);
  }
  o->kalman_dir = argv[19];
  o->out_dir = argv[20];
  o->useGaussianWeights = (o->efunc != &lk_error_STB);
  o->n_mark_types = atoi(argv[21]); 
  o->mark_fname_list = &(argv[22]);
  return o;
}

void at_print_options(FILE *f, options_t o)
{
  fprintf(f, "niter = %d\n", o->niter);
  fprintf(f, "error function = %s\n", o->efunc_name);
  fprintf(f, "minimizer = %s\n", o->minize_name);
  fprintf(f, "step_factor = %8.6f\n", o->step_factor);
  /* !!! Muito mais !!! */
  fprintf(f, "Generic camera parameters:\n");
  tf_show_camera_specs (o->global_cspec, f);
  track_parameters_print(f);
}

r2_t *at_copy_mark_positions(int ntargets, r2_t *p_i)
{
  r2_t *copy = (r2_t *)malloc(ntargets*sizeof(r2_t));
  int i;
  for (i = 0; i < ntargets; i++) { copy[i] = p_i[i]; }
  return copy;
}

double *at_copy_mark_weights(int ntargets, double *p_wgt)
{
  double *copy = (double *)malloc(ntargets*sizeof(double));
  int i;
  for (i = 0; i < ntargets; i++) { copy[i] = p_wgt[i]; }
  return copy;
}

