#include "at_process_frame.h"
#include "lk_error_M.h"
#include "lk_error_STB.h"
#include "track_parameters.h"
#include "tf_errors.h"
#include <malloc.h>
#include <jsfile.h>
#include <tf_kalman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <cimage.h>
#include <ipng.h>
#include <convert.h>
#include <png.h>

void at_guess_cpar_for_frame
  ( options_t o,
    image_t I,
    targets_data_t tdat,
    camera_parameters_t prev_cpar[],
    kalman_parameters_t **K,
    camera_parameters_t cpar, 
    camera_specs_t cspec,
    int frame_num,
    char *frame_out_dir )
{
    int ntargets = tdat->ntargets;

    /* Max Kalman model order that can be used for this frame: */
    int kalman_order = imin(frame_num - o->first_frame, o->max_kalman_order_cpar); 

    /* Obtain the {cspec} and guessed {cpar} for optimization: */
    if (frame_num == o->first_frame) {
      /* Leave {cspec} as a copy of {global_cspec}: */
      /* Get the basic {cpar} fields {Npx,Npy,dpx,dpy,sx,Cx,Cy} from {cspec}: */
      /* Guess {cpar} from {cdat} using the ad-hoc guesser: */
      fprintf(stderr, "@@ Computing the initial guess for camera parameters\n");
      (*cspec) = (*(o->global_cspec));
      struct _calibration_data_t cdat =
        (struct _calibration_data_t){ ntargets, tdat->p_w, tdat->p_i, tdat->conf };
      fprintf(stderr, "@@ Guessing camera parameters from user-given data.\n");
      tf_calib_guess2_initial_camera_parameters
	(&cdat, cspec, cpar, FALSE, TRUE);

      if (DEBUG_GUESS_CPAR) {
        fprintf(stderr, "++++++++++++++++++++++++++++++\n");
        fprintf(stderr, "Frame %d\n", frame_num);
        fprintf(stderr, "Camera pameters of tf_calib_guess2_initial_camera_parameters \n");
        tf_show_camera_parameters (cpar, stderr); 
        fprintf(stderr, "++++++++++++++++++++++++++++++\n");
        fflush(stderr);
      }

      fprintf(stderr, "@@ Optimization of the camera parameters from user-given data.\n");
      r2_t p_i_dev = (r2_t){{ 1.0, 1.0 }}; /* Expected error in user-given positions. */
      tf_calib_generic
        ( &cdat, cspec, p_i_dev, FALSE, cpar,
          tf_calib_refine2_gather_optimization_params,
          tf_calib_refine2_scatter_optimization_params,
          FALSE); 
      
      fprintf(stderr, "Calibration error summary for user-given data:\n");
      tf_calib_summarize_errors (&cdat, cpar, stderr);
    } 
    else {
      /* Set {cspec} by Kalman extrapolation of the {cpar}s of previous frames: */
      fprintf(stderr, "@@ Extrapolating camera parameters from previous frames\n");
      at_extrapolate_camera_parameters_kalman 
        ( o->global_cspec, kalman_order, 
          prev_cpar, K[kalman_order], 
          DEBUG_KALMAN_CPAR, cpar, cspec );
      if (DEBUG_KALMAN_CPAR_IMAGE) {
        at_write_prev_cpar_image (cpar, prev_cpar, kalman_order, tdat, I, frame_out_dir);
      }
    }

    if (DEBUG_GUESS_CPAR) {
      fprintf(stderr, "++++++++++++++++++++++++++++++\n");
      fprintf(stderr, "Frame %d\n", frame_num);
      fprintf(stderr, "Initial guess of camera parameter ranges\n");
      tf_show_camera_specs (cspec, stderr); 
  
      fprintf(stderr, "Initial guess of mean camera parameters\n");
      tf_show_camera_parameters (cpar, stderr); 
      fprintf(stderr, "++++++++++++++++++++++++++++++\n");
      fflush(stderr);
    }
}

void at_guess_target_positions_and_shapes
  ( options_t o,
    camera_parameters_t cpar,
    int ntargets,
    r2_t p_i_given[],
    double p_wgt_given[],
    r2_t **prev_p_i,
    double **prev_p_wgt,
    kalman_parameters_t **K,
    r3_t p_w[],
    r3_t u_w[],
    r3_t v_w[],
    r2_t p_i_cpar[],
    r2_t p_i_guess[],
    r2_t *p_i_dev_guess,
    double p_wgt_guess[],
    r2_t u_i[],
    r2_t v_i[],
    int frame_num )
{
    /* Compute the mark vectors {u_i,v_i} from world coords and guessed {cpar}: */
    /* Compute also the mark positions {p_i_aux} predicted from guessed {cpar}: */
    int mark; 
    for (mark = 0; mark < ntargets; mark++) {
      tf_mark_estimate_position_and_shape 
        ( cpar,
          p_w[mark], u_w[mark], v_w[mark],
          &(p_i_cpar[mark]), &(u_i[mark]), &(v_i[mark]));
    }

    if (frame_num == o->first_frame) {
      /* Use the given mark positions and weights: */
      fprintf(stderr, "@@ Using the given image coordinates as initial guesses.\n");
      for (mark = 0; mark < ntargets; mark++) { 
         p_i_guess[mark] = p_i_given[mark];
         p_wgt_guess[mark] =  p_wgt_given[mark];
      }
      (*p_i_dev_guess) = (r2_t){{ 2.0, 2.0 }}; /* !!! Should be a parameter. */
    }
    else if (o->extrapolateMarkPositions) {
      /* Max Kalman model order that can be used for this frame: */
      int kalman_order = imin(frame_num - o->first_frame, o->max_kalman_order_pos); 

     /* Get Kalman model for mark image coordinates {K_i[0..1]}: */
      kalman_parameters_t *K_i = K[kalman_order];
      /* Extrapolate {p_i} from the positions in previous frames: */
      fprintf(stderr, "@@ Extrapolating mark positions from previous frames\n");
      for (mark = 0; mark < ntargets; mark++) {
        r2_t p_i_avg;
        double p_wgt;
        at_extrapolate_image_position_kalman
          ( kalman_order, 
            prev_p_i, prev_p_wgt, K_i, 
            mark, DEBUG_KALMAN_POINTS, &p_i_avg, &p_wgt );
        if (p_wgt == 0.0) {
          /* !!! We should not do this fix -- the point is to compare mark tracking vs. camera tracking. */
          fprintf(stderr, "*** Mark %d was lost, Kalman failed - trying computed position.\n", mark); 
          p_i_avg = p_i_cpar[mark];
          p_wgt = 1.0;
	}
	p_i_guess[mark] = p_i_avg;
        p_wgt_guess[mark] = p_wgt;
      } 
      /* Get the mark position uncertainty from the Kalman models: */
      (*p_i_dev_guess) = at_mark_position_uncertainty(K_i); /* Deviation for mark image positions. */
    }
    else {
      fprintf(stderr, "@@ Using the image coordinates computed from {cpar} as initial guesses\n");
      for (mark = 0; mark < ntargets; mark++) 
        { p_i_guess[mark] = p_i_cpar[mark];
          p_wgt_guess[mark] = (double)tf_camera_point_is_inside_image(p_i_guess[mark], cpar);
        }
      (*p_i_dev_guess) = (r2_t){{ 2.0, 2.0 }}; /* !!! Should be a parameter. */
    }
    
    /* Make sure that points outside the image get zero confidence: */
    for (mark = 0; mark < ntargets; mark++) {

        if ((p_wgt_guess[mark] > 0.0)  && (! tf_camera_point_is_inside_image(p_i_guess[mark], cpar))) {
          p_wgt_guess[mark] = 0.0;
	}
        if (p_wgt_guess[mark] == 0.0) {
          fprintf(stderr, "*** Position of mark %d is unknown or outside the image.\n", mark); 
	}
    }

    if (! o->useKalmanErrorInFinder) {
      fprintf(stderr, "@@ Turning off the Kalman error terms in the feature finder.\n");
      /* Make {p_i_dev_finder} very large so that the guessed-{p_i} Kalman error terms vanish: */
      (*p_i_dev_guess) = (r2_t){{ 1.0e8, 1.0e8 }};
    }
}

void at_compute_target_images_and_masks
  ( image_list_t M_list,
    image_list_t W_list,
    options_t o,
    r2_t u_i[],
    r2_t v_i[],
    bool_t deform_images,
    image_list_t *G_list,
    image_list_t *V_list,
    char *frame_out_dir
  )
{
  int ntargets = M_list->nimages;

  *G_list = image_list_alloc(ntargets); 
  *V_list = image_list_alloc(ntargets);

  int mark;
  
  for (mark = 0; mark < ntargets; mark++) {
    
    // Get the mark's image {M}: 
    image_t M = M_list->imgs[mark];
    image_t W = W_list->imgs[mark];

    /* Generate the deformed mark image {G} and corresponding weight mask {V}: */
    if (deform_images) {
      /*  Apply affine deformation to {M,W} based on {u_w,v_w} and {u_i,v_i}: */
      at_compute_auxiliary_mark_images
	( M, W,
	  o->QSubsampling,
	  o->markBlur,
	  u_i[mark], v_i[mark],
	  &((*G_list)->imgs[mark]),
	  &((*V_list)->imgs[mark])
	  );
    } else {
      /* Use undeformed mark images: */
      (*G_list)->imgs[mark] = image_copy(M);
      (*V_list)->imgs[mark] = image_copy(W);
    }
  }
}

void at_estimate_alphas_and_betas
  ( image_t I,
    image_list_t G_list,
    image_list_t V_list,
    options_t o,
    r2_t p_i[],
    double p_wgt[],
    bool_t debug,
    double alpha[],
    double beta[] )
{
  int ntargets = G_list->nimages;

  int mark;
  for (mark = 0; mark < ntargets; mark++) {
    
    if (p_wgt[mark] > 0) {
      at_estimate_alpha_beta
        ( I, 
      	  G_list->imgs[mark], 
      	  V_list->imgs[mark], 
      	  o->QSubsampling, 
      	  p_i[mark], 
          p_wgt[mark],
          debug,
      	  &(alpha[mark]),
      	  &(beta[mark])
      	);
    }
  }
}

void at_locate_targets_in_frame
( image_t I,
  image_list_t G_list,
  image_list_t V_list,
  double alpha[],
  double beta[],
  options_t o,
  r2_t p_i_guess[], 
  r2_t p_i_dev_guess,
  double p_wgt_guess[],
  r2_t p_i_found[],
  r2_t *p_i_dev_found, 
  double p_wgt_found[],
  char *frame_out_dir )
{
  int ntargets = G_list->nimages;

  (*p_i_dev_found) = (r2_t){{ 0.0, 0.0 }}; /* For now. */

  char *pyramid_name = NULL;
  asprintf(&pyramid_name, "%s/I", frame_out_dir);
              
  /* Compute image pyramid: */
  image_pyramid_t I_pyr;
  image_pyramid_from_image (I, NULL, o->redFactor, o->minMarkSize, &I_pyr, NULL);

  /* Write the image pyramid for debugging: */
  if (DEBUG_FRAME_PYRAMID)
    { image_pyramid_write(I_pyr, PGM_RAW_MAGIC, MAXVAL, pyramid_name, "I", "pgm", 0,0, 0,0); }
  
  /* Write images showing guessed target positions: */
  at_write_debugging_images
    (I, 1, ntargets, p_i_guess, G_list, V_list, o->QSubsampling, alpha, beta, frame_out_dir, "I_klt_guess" );

  /* Prepare debugging images for KLT: */
  image_pyramid_t I_pyr_deb1 = NULL; /*Frame images, with the initial guess points*/
  image_pyramid_t I_pyr_deb2 = NULL; /*Frame images, with the KLT adjusted points*/
  if (DEBUG_POINTS_PYRAMID) {
    I_pyr_deb1 = image_pyramid_copy (I_pyr); 
    I_pyr_deb2 = image_pyramid_copy (I_pyr); 
  }

  int mark;
  for (mark = 0; mark < ntargets; mark++) {
    
    if (p_wgt_guess[mark] > 0) {
      char *mark_debug_name = NULL;
      asprintf(&mark_debug_name, "%s/deb_%02d", frame_out_dir, mark);

      double errorSqr; /* Mean quadratic error returned by {at_locate_mark}. */
      r2_t p_i_dev_found_mark;
      at_locate_mark
	( I_pyr, 
	  G_list->imgs[mark], 
	  V_list->imgs[mark], 
	  o,
	  p_i_guess[mark],
	  p_i_dev_guess,
	  alpha[mark],
	  beta[mark],
	  mark_debug_name, 
	  &(p_i_found[mark]),
	  &(p_i_dev_found_mark),
	  &errorSqr, 
	  I_pyr_deb1,
	  I_pyr_deb2 );
           
      double csigma = 0.05;  /* Root-mean-square error threshold. */
      double epsilon = 0.10; /* Probability of a bogus match with 0 error. */
      double lk_conf = lk_compute_confidence_A(errorSqr, alpha[mark], beta[mark], csigma); /* LK confidence. */
      p_wgt_found[mark] = lk_conf*(1.0 - epsilon);

      /* Update the maximum position uncertainty {p_i_dev_found} of all marks: */
      int axis;
      for (axis = 0; axis < 2; axis++) {
        if (p_i_dev_found_mark.c[axis] > p_i_dev_found->c[axis]) {
	  p_i_dev_found->c[axis] = p_i_dev_found_mark.c[axis];
	}
      }

      free(mark_debug_name);
    }
    else {
      /* mark was lost, stays lost: */
      p_i_found[mark] = p_i_guess[mark];
      p_wgt_found[mark] = 0.0; 
    }
  }

  if (DEBUG_POINTS_PYRAMID) {
    image_pyramid_write(I_pyr_deb1, PGM_RAW_MAGIC, MAXVAL, pyramid_name, "klt_guess", "pgm", 0,0, 0,0);
    image_pyramid_write(I_pyr_deb2, PGM_RAW_MAGIC, MAXVAL, pyramid_name, "klt_found", "pgm", 0,0, 0,0);

    image_pyramid_free (I_pyr_deb1);
    image_pyramid_free (I_pyr_deb2);
  }

  /* Write images showing found target positions: */
  at_write_debugging_images
    (I, 1, ntargets, p_i_found, G_list, V_list, o->QSubsampling, alpha, beta, frame_out_dir, "I_klt_found" );

  image_pyramid_free(I_pyr);
  free(pyramid_name);
}

void at_calibrate_camera_for_frame
( options_t o,
  camera_specs_t cspec,
  int ntargets,
  r3_t p_w[],
  r2_t p_i[],
  r2_t p_i_dev,
  double p_wgt[],
  camera_parameters_t cpar,
  char *frame_out_dir,
  int frame_num )
{
    if (! o->useKalmanErrorInCalib) {
      fprintf(stderr, "@@ Turning off the Kalman error terms in the camera calibration.\n");
    }

    int mark;

    if (! o->useWeightsInCalib) { 
      fprintf(stderr, "@@ Setting the calibration weights of all located marks to 1.\n");
      for (mark = 0; mark < ntargets; mark++) { 
	if (p_wgt[mark] != 0.0) { p_wgt[mark] = 1.0; }
      }
    }
    else {
      fprintf(stderr, "@@ Using the feature-finder confidence as calibration weights.\n");
    }

  if (WRITE_GUESS_CPAR) {
    /* Write the input {cpar}, presumably the mean of {cspec}: */
    char *file_cpar_fname = NULL;
    asprintf(&file_cpar_fname, "%s/guess.cpar", frame_out_dir);
    FILE *file_cpar = open_write(file_cpar_fname, TRUE);
    tf_write_camera_parameters (file_cpar, frame_num, cpar);
    fclose(file_cpar);
    free(file_cpar_fname);
  }

  struct _calibration_data_t cdat =
    (struct _calibration_data_t){ ntargets, p_w, p_i, p_wgt };

  if(o->useInternalGuessInCalib) {
    fprintf(stderr, "@@ Re-guessing camera parameters from found mark positions.\n");
    tf_calib_guess2_initial_camera_parameters
      (&cdat, cspec, cpar, FALSE, TRUE);
  }
  else {
    fprintf(stderr, "@@ Using previously guessed camera parameters.\n");
  }

  fprintf(stderr, "@@ Nonlinear optimization of the camera parameters (FIRST PASS).\n");
  tf_calib_generic
    ( &cdat, cspec, p_i_dev, o->useKalmanErrorInCalib, cpar,
      tf_calib_refine2_gather_optimization_params,
      tf_calib_refine2_scatter_optimization_params,
      FALSE); 
  
  fprintf(stderr, "Calibration error summary (FIRST PASS):\n");
  tf_calib_summarize_errors (&cdat, cpar, stderr);

  if (o->twoPassCalib) {

    if (WRITE_CALIB_CPAR) {
      char *cpar_fname = NULL;
      asprintf(&cpar_fname, "%s/semiopt.cpar", frame_out_dir);
      FILE *cpar_file = open_write(cpar_fname, TRUE);
      free(cpar_fname);
      tf_write_camera_parameters (cpar_file, frame_num, cpar);
      fclose(cpar_file);
    }

    /* Recompute the confidence weights based on the mark position errors: */
    fprintf(stderr, "@@ Recomputing the confidence weights based on the mark position errors.\n");
    /* Estimate the standard deviation {dev_bad} of outliers: */
    r2_t dev_bad;
    double cook1 = 5.0; /* Guessed ratio of outlier-dev/inlier-dev. */
    r2_scale(cook1, &p_i_dev, &dev_bad);
    /* Estimate the standard deviation {dev_gud} of inliers: */
    r2_t dev_gud; /* Standard deviation for inliers. */
    double cook2 = 0.10; /* Guessed relative effect of outliers on predicted {p_i}s. */
    r2_mix(1.0, &p_i_dev, cook2, &dev_bad, &dev_gud);
    tf_recompute_target_weights(cpar, ntargets, p_w, p_i, p_wgt, dev_gud, dev_bad, FALSE);

    at_mark_weights_write(ntargets, p_wgt, frame_out_dir, "recalc");
    
    fprintf(stderr, "@@ Nonlinear optimization of the camera parameters (SECOND PASS).\n");
    tf_calib_generic
      ( &cdat, cspec, p_i_dev, o->useKalmanErrorInCalib, cpar,
        tf_calib_refine2_gather_optimization_params,
        tf_calib_refine2_scatter_optimization_params,
        FALSE); 

    fprintf(stderr, "Calibration error summary (SECOND PASS):\n");
    tf_calib_summarize_errors (&cdat, cpar, stderr);
  }
    
  if (WRITE_CALIB_CPAR) {
    char *cpar_fname = NULL;
    asprintf(&cpar_fname, "%s/opt.cpar", frame_out_dir);
    FILE *cpar_file = open_write(cpar_fname, TRUE);
    free(cpar_fname);
    tf_write_camera_parameters (cpar_file, frame_num, cpar);
    fclose(cpar_file);
  }
  
  if (DEBUG_CALIB_CPAR) {
    fprintf(stderr, "++++++++++++++++++++++++++++++\n");
    fprintf(stderr, "Frame %d\n", frame_num);
    fprintf(stderr, "Calibrated camera parameters\n");
    tf_show_camera_parameters (cpar, stderr); 
    fprintf(stderr, "++++++++++++++++++++++++++++++\n");
    fflush(stderr);
  }
}

void at_write_debugging_images
  ( image_t I,
    int magnify,
    int ntargets,
    r2_t p_i[],
    image_list_t G_list,
    image_list_t V_list,
    int subsampling,
    double alpha[],
    double beta[],
    char *frame_out_dir,
    char *tag )
{

  if (DEBUG_PAINTED_MARKS_IMAGE) {
    at_write_debugging_image_with_painted_targets
      ( I, ntargets, p_i, G_list, V_list, subsampling, alpha, beta, frame_out_dir, tag );
  }

  if (DEBUG_CROSSES_IMAGE) {
    at_write_debugging_image_with_crosses
      ( I, magnify, ntargets, p_i, frame_out_dir, tag );
  }
}

void at_write_debugging_image_with_painted_targets
  ( image_t I,
    int ntargets,
    r2_t p_i[],
    image_list_t G_list,
    image_list_t V_list,
    int subsampling,
    double alpha[],
    double beta[],
    char *frame_out_dir,
    char *tag )
{
  assert(ntargets == G_list->nimages);
  assert(ntargets == V_list->nimages);
 
  /*Image overpainted with target images at the given positions: */  
  image_t J = image_copy(I);

  int mark;
  for (mark = 0; mark < ntargets; mark++) {
    image_t G_mark = G_list->imgs[mark];
    image_t V_mark = V_list->imgs[mark];
    double alpha_mark = (alpha == NULL ? 1.0 : alpha[mark]);
    double beta_mark = (beta == NULL ? 0.0 : beta[mark]);
    image_paint_translated_mark(G_mark, V_mark, J, subsampling, alpha_mark, beta_mark, p_i[mark]);
  }
    
  char *J_fname = NULL;
  asprintf(&J_fname, "%s/%s_t.pgm", frame_out_dir, tag);
  image_write_pgm (J, PGM_RAW_MAGIC, MAXVAL, J_fname);
  image_free (J);
  free(J_fname);
}

void at_write_debugging_image_with_crosses
  ( image_t I,
    int magnify,
    int ntargets,
    r2_t p_i[],
    char *frame_out_dir,
    char *tag )
{
  /*Image with crosses at the given positions: */  
  
  char *J_fname = NULL;
  asprintf(&J_fname, "%s/%s_c.pgm", frame_out_dir, tag);
  image_ppm_write_targets_with_crosses
      (I, p_i, ntargets, 6, MAXVAL, magnify, J_fname);
  free(J_fname);
}

void at_write_prev_cpar_image
  ( camera_parameters_t cpar,
    camera_parameters_t prev_cpar[],
    int kalman_order,
    targets_data_t tdat, 
    image_t I,
    char *frame_out_dir
  )
{
  int ntargets = tdat->ntargets; 
  r2_t u_i[ntargets], v_i[ntargets], p_i[ntargets];

  image_t I_tmp = image_copy (I);
 
  int size = 2;

  int ind;
  int mark;
    
  for (ind = (kalman_order-1); ind >= 0; ind--) {
    camera_parameters_t cp = prev_cpar[ind]; 

    for (mark = 0; mark < ntargets; mark++) {
      tf_mark_estimate_position_and_shape 
        ( cp, tdat->p_w[mark], tdat->u_w[mark], tdat->v_w[mark], &(p_i[mark]), &(u_i[mark]), &(v_i[mark]));
      if ( tf_camera_point_is_inside_image(p_i[mark], cp) ) {
	image_print_cross (I_tmp, size*(ind+2), p_i[mark]);
      }
    }
  }

  for (mark = 0; mark < ntargets; mark++) {
   tf_mark_estimate_position_and_shape 
     ( cpar, tdat->p_w[mark], tdat->u_w[mark], tdat->v_w[mark], &(p_i[mark]), &(u_i[mark]), &(v_i[mark]));
     if ( tf_camera_point_is_inside_image(p_i[mark], cpar) ) {
       image_print_dot (I_tmp, 3, p_i[mark]);
     }
  }

  char *I_fname = NULL;
  asprintf(&I_fname, "%s/I_cpar_prev.pgm", frame_out_dir);
  image_write_pgm (I_tmp, PGM_RAW_MAGIC, MAXVAL, I_fname);
  free(I_fname);
  image_free(I_tmp);
}

void at_locate_mark
  ( image_pyramid_t I_pyr,
    image_t G,
    image_t V,
    options_t o,
    r2_t p_i_guess,
    r2_t p_i_dev,
    double alpha,
    double beta, 
    char *mark_debug_name,
    r2_t *p_i_found,
    r2_t *p_i_dev_found,
    double *errorSqr, 
    image_pyramid_t I_pyr_deb1,
    image_pyramid_t I_pyr_deb2
  )
{
  /*Verify memory*/
  struct mallinfo info;
  info = mallinfo();
  int MemDinInicial = info.uordblks;
  /*End*/

  fprintf(stderr, "-------------------------------------\n");
  fprintf(stderr, "processing mark %s\n", mark_debug_name);

  int minGSize = o->QSubsampling * o->minMarkSize;

  image_pyramid_t G_pyr;

  image_pyramid_t V_pyr;

  image_pyramid_from_image (G, V, o->redFactor, minGSize, &G_pyr, &V_pyr);

  r2_t p_i = p_i_guess;
                 
  klt_find_mark_position_in_image 
    ( mark_debug_name, 
      I_pyr,
      G_pyr, 
      V_pyr,
      o->QSubsampling,
      alpha, beta, 
      &p_i,
      &p_i_dev,
      o->efunc,
      o->minize,
      o->step_factor,
      MINIMIZE_ITERS,
      errorSqr,
      I_pyr_deb1,
      I_pyr_deb2
      );

  /* Save the adjusted mark position: */
  (*p_i_found) = p_i;
  (*p_i_dev_found) = (r2_t){{ 1.0, 1.0 }}; /* !!! Must figure out the correct formula. */

  image_pyramid_free(G_pyr);
  image_pyramid_free(V_pyr);
  fprintf(stderr, "done with mark %s processing\n", mark_debug_name);
  fprintf(stderr, "-------------------------------------\n");

  /*Verify memory*/
  info = mallinfo();
  int MemDinFinal = info.uordblks;
  if (MemDinInicial != MemDinFinal)
    { fprintf(stderr, "*** at_locate_mark: memory leak %d block\n", MemDinFinal - MemDinInicial); }
}

void at_compute_auxiliary_mark_images
  ( image_t M,
    image_t W,      
    int QSubsampling,
    double markBlur,
    r2_t u_i, r2_t v_i,
    image_t *G, image_t *V
  )
{
  /* Generate the deformed mark image {G} and the deformed weight mask {V}: */
  image_deform(M, W, &u_i, &v_i, QSubsampling, G, V);
    
  if (markBlur != 0.0) {
    image_t GB = image_blur (*G, *V, markBlur);
    image_free (*G); 
    *G = GB;
  }
}



void at_estimate_alpha_beta
  ( image_t I, 
    image_t M, 
    image_t W, 
    int subsampling, 
    r2_t p_i, 
    double wgt,
    bool_t debug,
    double *alpha, 
    double *beta
  ) 
{
  int M_nx = M->nx;
  int M_ny = M->ny;

  // Find best-matching {alpha_b,beta_b} of {M} in {I} at given position: 
  double alpha_b, beta_b;
  double corr_b; // Correlation coefficient of {I} relative to {M}. 
  r2_t trivial_p_i = (r2_t){{ M_nx/2.0, M_ny/2.0 }};  // Point of {M} to align with {p_i}. 
  image_weighted_regression(W, M, 1, trivial_p_i, I, subsampling, p_i, debug, &alpha_b, &beta_b, &corr_b);
  if (debug) {
    fprintf(stderr, "%s: pi.c[0] = %f, pi.c[1] = %f\n", __FUNCTION__, p_i.c[0], p_i.c[1]);
    fprintf(stderr, "raw alpha = %f beta = %f corr = %f\n", alpha_b, beta_b, corr_b);
  }

  // Sanity adjustments: 
  if (alpha_b < 0.1) {alpha_b = 0.1;} // Mark cannot be blank or negative. //
  if (alpha_b > 1.0) {alpha_b = 1.0;} // Mark cannot have higher contrast in {I} than {M}. //
  if (beta_b < 0.0) {beta_b = 0.0;}  // !!! Not needed? //

  // Mix with {alpha,beta} according to confidence weight {wgt}: //
  (*alpha) = (isnan(*alpha) ? alpha_b : (1-wgt)*(*alpha) + wgt*alpha_b);
  (*beta)  = (isnan(*beta)  ? beta_b  : (1-wgt)*(*beta) + wgt*beta_b);
  
  if (debug) {
    fprintf(stderr, "adjusted alpha = %f beta = %f\n", *alpha, *beta);
  }
  
} 


/*void at_estimate_alpha_beta
  ( image_t I,
    image_t M,
    image_t W,
    int subsampling,
    r2_t p_i,
    double wgt,
    bool_t debug,
    double *alpha,
    double *beta
  )
{
  int M_nx = M->nx;
  int M_ny = M->ny;

  // Compute mean and average of mark in the reference image {M}: 
  r2_t trivial_p_i = (r2_t){{ M_nx/2.0, M_ny/2.0 }};

  double Mavg, Mvar;
  image_weighted_avg_var (M, W, 1, trivial_p_i, &Mavg, &Mvar);
  fprintf(stderr, "Mavg %f Mvar %f\n", Mavg, Mvar);

  // Compute mean and average of mark in the frame image {I}: 

  double Iavg, Ivar;
  image_weighted_avg_var (I, W, subsampling, p_i, &Iavg, &Ivar);
  fprintf(stderr, "Iavg %f Ivar %f\n", Iavg, Ivar);

  *alpha = sqrt (Ivar/Mvar);
  if (*alpha < 0.1) {*alpha = 0.1;}
  if (*alpha > 1.0) {*alpha = 1.0;}

  *beta = Iavg - (*alpha)*Mavg;

  if (*beta < 0.0) {*beta = 0.0;}

  *alpha = 5.0;
  *beta = 0.0;

  fprintf(stderr, "Estimated alpha = %f beta = %f\n", *alpha, *beta);
} */

void at_extract_feature_images (image_t R, targets_data_t tdat, int QSubsampling, image_list_t M_list)
{
  
  assert(M_list->nimages == tdat->ntargets);

  M_list->imgs = (image_t *)malloc(M_list->nimages * sizeof(image_t));

  int mark;
  for (mark = 0; mark < M_list->nimages; mark++) {
    if (tdat->conf[mark] > 0.0) {
      /* !!! Should also check that the target is wholly inside {I} !!! */
      r2_t p_i = tdat->p_i[mark];
      r2_t u_i = tdat->u_i[mark];
      r2_t v_i = tdat->v_i[mark];
      int hwx = (int)ceil(fabs(u_i.c[0]) + fabs(v_i.c[0]));
      int hwy = (int)ceil(fabs(u_i.c[1]) + fabs(v_i.c[1]));
      fprintf(stderr, "mark %2d", mark);
      fprintf(stderr, "  u_i = (%8.3lf %8.3lf)", u_i.c[0], u_i.c[1]);
      fprintf(stderr, "  v_i = (%8.3lf %8.3lf)", v_i.c[0], v_i.c[1]);
      fprintf(stderr, "  size = %d x %d\n", 2*hwx+1, 2*hwy+1);
      if (M_list->imgs[mark] != NULL) { image_free(M_list->imgs[mark]); }
      M_list->imgs[mark] = at_extract_feature(R, p_i, hwx, hwy, QSubsampling);
    }
  }
}

image_t at_extract_feature(image_t R, r2_t p_i, int hwx, int hwy, int QSubsampling)
{
  int nx = (2*hwx + 1)*QSubsampling;
  int ny = (2*hwy + 1)*QSubsampling;
  image_t M = image_alloc(nx, ny);
  image_extract_translated_mark(R, M, QSubsampling, 1.0, 0.0, p_i);
  return M;
}

void at_extrapolate_camera_parameters_kalman 
  ( camera_specs_t global_cspec, 
    int order, 
    camera_parameters_t prev_cpar[], 
    kalman_parameters_t K[],
    bool_t debug,
    camera_parameters_t cpar,
    camera_specs_t cspec )
{
  int nparams = tf_camera_num_parameters;
  assert(order > 0);

  double data[order]; /* Previous parameter values. */

  int iparam;
  for (iparam = 0; iparam < nparams; iparam++) { 
    /* Choose a reference value (useful only for the {R} parameters): */
    double vref = tf_camera_get_param(prev_cpar[0], iparam, NAN);
    /* Get the parameter's range from {global_cspec}: */
    interval_t global_range = tf_camera_get_param_range(global_cspec, iparam, vref);

    if (interval_is_trivial(&global_range)) {
      /* There is nothing to predict: */
      tf_camera_set_param_range(cspec, iparam, &global_range);
    }
    else {
      /* Get the Kalman model for this parameter: */
      kalman_parameters_t Ki = K[iparam];
      int this_order = Ki->order;
      assert(this_order <= order);
      /* Get the data for the Kalman prediction: */
      if (debug) { fprintf(stderr, "%s: data =", tf_camera_param_name(iparam)); }
      int i;
      for (i = 0; i < this_order; i++)
	{ data[i] = tf_camera_get_param(prev_cpar[i], iparam, vref);
          if (debug) { fprintf(stderr, " %.8f", data[i]); }
        }
      /* Evaluate the Kalman model: */
      double avg, dev;
      tf_kalman_predict_parameter(Ki, data, &avg, &dev);
      if (debug) { fprintf(stderr, "  avg  = %.8f dev = %.8f", avg, dev); }
      /* Make sure that the predicted value {avg} is inside the global range: */
      avg = interval_project(&global_range, avg);
      /* Convert the prediction to interval format: */
      interval_t kalman_range = (interval_t){{ avg-3*dev, avg+3*dev }};
      if (debug) { 
         fprintf(stderr, "  range = [ %.8f _ %.8f ]\n", LO(kalman_range), HI(kalman_range));
      }
      /* Store the prediction in {cspec}: */
      tf_camera_set_param_range(cspec, iparam, &kalman_range);
    }
  }
  /* Set {cpar} to be the mean of that {cspec}: */
  tf_store_mean_camera_parameters(cspec, cpar);
}

r2_t at_mark_position_uncertainty(kalman_parameters_t K_i[])
{
  return (r2_t){{ K_i[0]->dev, K_i[1]->dev }};
}

void at_extrapolate_image_position_kalman
  ( int order, 
    r2_t *prev_p_i[],
    double *prev_p_wgt[],
    kalman_parameters_t K_i[],
    int mark,
    bool_t debug,
    r2_t *p_i_avg,
    double *p_wgt )
{
  (*p_wgt) = +INF; /* For now. */
  double data[order];
  int axis;
  for (axis = 0; axis < 2; axis++) {
    /* Get the Kalman model for this axis: */
    kalman_parameters_t Kax = K_i[axis];
    int axis_order = Kax->order;
    assert(axis_order <= order);
    /* Gather data for ths axis: */
    int i;
    for (i = 0; i < axis_order; i++) {
      double coord = prev_p_i[i][mark].c[axis];
      double weight = prev_p_wgt[i][mark];
      data[i] = coord;
      if (weight < (*p_wgt)) { (*p_wgt) = weight; }
    }
    double avg, dev;
    tf_kalman_predict_parameter(Kax, data, &avg, &dev);
    p_i_avg->c[axis] = avg;
  }
  if (debug) {
    fprintf(stderr, "%s: p_i_avg = ( %f  %f ) p_wgt = %f\n", 
      __FUNCTION__, p_i_avg->c[0], p_i_avg->c[1], *p_wgt); 
  }
}

void at_mark_positions_and_weights_write
  ( int ntargets,
    r2_t p_i[],
    double p_wgt[],
    char *dir, 
    char *tag )
{
  at_mark_positions_write(ntargets, p_i, dir, tag);
  at_mark_weights_write(ntargets, p_wgt, dir, tag);
} 

void at_mark_positions_write
  ( int ntargets,
    r2_t p_i[],
    char *dir, 
    char *tag )
{
  int mark;

  char *p_i_fname = NULL;
  asprintf(&p_i_fname, "%s/p_i_%s.txt", dir, tag);
  FILE *p_i_file = open_write(p_i_fname, TRUE);
  free(p_i_fname);

  fprintf(p_i_file, "%d\n", ntargets);
  for (mark = 0; mark < ntargets; mark++) {
    fprintf(p_i_file, "%f %f\n", p_i[mark].c[0], p_i[mark].c[1]);
  }

  fclose(p_i_file);
} 

void at_mark_weights_write
  ( int ntargets,
    double p_wgt[],
    char *dir, 
    char *tag )
{
  int mark;

  char *p_wgt_fname = NULL;
  asprintf(&p_wgt_fname, "%s/p_wgt_%s.txt", dir, tag);
  FILE *p_wgt_file = open_write(p_wgt_fname, TRUE);
  free(p_wgt_fname);

  fprintf(p_wgt_file, "%d\n", ntargets);
  for (mark = 0; mark < ntargets; mark++) {
    fprintf(p_wgt_file, "%f\n", (p_wgt == NULL ? 1.0 : p_wgt[mark]));
  }

  fclose(p_wgt_file);
} 

void at_alpha_beta_write
  ( int ntargets,
    double alpha[],
    double beta[],
    char *dir, 
    char *tag )
{
  int mark;

  char *alpha_beta_fname = NULL;
  asprintf(&alpha_beta_fname, "%s/alpha_beta_%s.txt", dir, tag);
  FILE *alpha_beta_file = open_write(alpha_beta_fname, TRUE);
  free(alpha_beta_fname);

  fprintf(alpha_beta_file, "%d\n", ntargets);
  for (mark = 0; mark < ntargets; mark++) {
    fprintf(alpha_beta_file, "%f %f\n", alpha[mark], beta[mark]);
  }

  fclose(alpha_beta_file);
} 

