#ifndef __at_process_frame_H__
#define __at_process_frame_H__

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <image.h>
#include <image_list.h>
#include <image_pyramid.h>
#include <image_pyramid_list.h>
#include <tf_calib.h> 
#include <tf_kalman.h> 
#include <tf_calib_guess1.h> 
#include <tf_calib_guess2.h> 
#include <tf_calib_refine2.h> 
#include <tf_math.h> 
#include <tf_targets.h> 
#include <r2x2.h> 
#include <r2.h>
#include "kltM.h" 
#include <time.h> 
#include "ransac.h" 

#define DEBUG_FRAME_PYRAMID 0 

#define DEBUG_GUESS_POINTS 1 
#define DEBUG_FOUND_POINTS 1 
#define DEBUG_CALIB_POINTS 1 

#define DEBUG_GUESS_CPAR 1 
#define DEBUG_CALIB_CPAR 1 
#define DEBUG_KALMAN_CPAR 1

#define DEBUG_ALPHA_BETA 1
#define DEBUG_KALMAN_POINTS 0 

#define WRITE_GUESS_POINTS 1 
#define WRITE_FOUND_POINTS 1 
#define WRITE_CALIB_POINTS 1 
#define WRITE_ALPHA_BETA 1 
#define WRITE_GUESS_CPAR 1 
#define WRITE_CALIB_CPAR 1 

#define DEBUG_KALMAN_CPAR_IMAGE 1
#define DEBUG_PAINTED_MARKS_IMAGE 1
#define DEBUG_CROSSES_IMAGE 1
#define DEBUG_POINTS_PYRAMID 0

typedef struct _options_t 
{ 
  char *out_dir;                   /* Output directory. */
  bool_t useSyntheticMarks;        /* Use synthetic mark images or real mark images extracted from video? */
  bool_t useMarksFromFirstFrame;   /* Extract real mark images from the first frame, or from previous frame? */
  int redFactor;                   /* Reduction factor for image pyramid */
  int QSubsampling;                /* Ideal subsamples per pixel when computing image integrals? */
  int niter;                       /* Number of Newton iterations per level. */
  double markBlur;                 /* Extra blurring factor when comparing targets. */
  int minMarkSize;                 /* Minimum dimension of a reduced and deformed mark along any axis. */
  bool_t useWeightsInCalib;        /* Use mark weights in Tsai's algorithm? */
  bool_t adaptiveAlphaBeta;        /* Recompute alpha and beta after each camera calibration? */
  bool_t extrapolateMarkPositions; /* Get the KLT guess {tdat->p_i} from Kalman of positions, not {cpar}? */
  bool_t useKalmanErrorInFinder;   /* Use the Kalman position-error terms in the feature finder? */
  bool_t useKalmanErrorInCalib;    /* Use the Kalman {cpar}-error terms in the camera calibrator? */
  bool_t twoPassCalib;             /* Repeat the calibration twice for each frame? */
  bool_t useInternalGuessInCalib;  /* Shall we ignore the extrapolated {cpar} in calibration? */
  bool_t useComputedPositions;     /* Compute final {tdat->p_i} from {cpar} instead of the KLT output? */
  int max_kalman_order_pos;        /* Maximum order of Kalman models to use for mark positions in image. */
  int max_kalman_order_cpar;       /* Maximum order of Kalman models to use for camera parameters. */
  char *p_w_fname;                 /* File name of mark positions (world coordinates). */
  char *tuv_w_fname;               /* File name of mark types and shapes (world coordinates). */
  char *p_i_fname;                 /* File name of mark positions on first frame. */
  char *p_wgt_fname;               /* File name of mark confidence weights for first frame. */
  char *frame_prefix;              /* Prefix for frame file names: */
  int first_frame;                 /* Number of first frame to analyze: */
  int last_frame;                  /* Number of last frame to analyze: */
  char *efunc_name;                /* Name of mark finder's error function. */
  lk_error_function_t *efunc;      /* Mark finder's error function. */
  char *minize_name;               /* Name of mark finder's minimizer. */ 
  numeric_minimizer_t *minize;     /* Mark finder's minimizer. */ 
  double step_factor;              /* Choose the minimization step factor: */
  camera_specs_t global_cspec;     /* Basic camera specifications. */
  char *kalman_dir;                /* Directory for Kalman models: */
  bool_t useGaussianWeights;       /* Shall we use Gaussian or uniform weights? */ 
  int n_mark_types;                /* Number of distinct marks: */
  char **mark_fname_list;          /* Array of mark image names. */
} *options_t;

void at_guess_cpar_for_frame
  ( options_t o,
    image_t I,
    targets_data_t tdat,
    camera_parameters_t prev_cpar[],
    kalman_parameters_t **K,
    camera_parameters_t cpar, 
    camera_specs_t cspec,
    int frame_num,
    char *frame_out_dir );
/* Computes an initial estimate of the camera parameters {cpar} and their
  ranges {cpspec} for the frame {frame_num}. Uses the world positions
  {tdat->p_w}. 

  On input, the ranges of {o->global_cspec} determine which parameters are fixed
  and which are variable.  Fixed parameters are set to their fixed value.
  
  For the first frame, the variable parameters are obtained by the camera
  calibration procedure, using the mark image positions in {tdat->p_i} 
  and {tdat->conf}.  The ranges in {cspec} are copied from {o->global_cspec}.

  For the second and following frames, the variable camera parameters are
  extrapolated using the Kalman models {K[order][iparam]} and the 
  previous parameter sets {prev_cpar[0..nprev-1]} where {nprev} is 
  {min(o->max_kalman_order_cpar,frame_num - o->first_frame)}.
  The ranges in {cspec} are set according to the {.dev} fields of
  the Kalman models. */

void at_guess_target_positions_and_shapes
  ( options_t o,
    camera_parameters_t cpar,
    int ntargets,
    r2_t p_i_given[],
    double p_wgt_given[],
    r2_t **prev_p_i,
    double **prev_p_wgt,
    kalman_parameters_t **K,
    r3_t p_w[],
    r3_t u_w[],
    r3_t v_w[],
    r2_t p_i_cpar[],
    r2_t p_i_guess[],
    r2_t *p_i_dev_guess,
    double p_wgt_guess[],
    r2_t u_i[],
    r2_t v_i[],
    int frame_num );
/* Obtains guesses for the positions {p_i_guess[mark]} of each mark in the frame,
  for use by the feature tracker. Also obtains guessed shape
  vectors {u_i[mark],v_i[mark]} for each shape, an /a priori/
  confidence weight {p_wgt_guess[mark]}, and the expected position error
  {*p_i_dev_guess}. 

  The shape vectors are always computed from the given {cpar}.  The
  positions computed from the {cpar} are returned in {p_i_cpar}.  Note
  that these may or may not be identical to the guessed positions
  {p_i_guess}, depending on the options.
 
  For the first frame, the guessed positions and weights are copied from 
  the given vectors {p_i_given,p_wgt_given}; and the uncertainty
  {p_i_dev_guess} is set to an appropriate guess. 

  For the other frames, if {o->extrapolateMarkPositions} is true, the positions are
  extrapolated from the positions in previous frames {prev_p_i[0..nprev-1][mark]},
  and their confidences {prev_p_wgt[nprev-1][mark]}}, where {nprev} is 
  {min(o->max_kalman_order_pos,frame_num - o->first_frame)}.

  For the other frames, if {o->extrapolateMarkPositions} is false, the positions are
  computed from the given {cpar}.

  In any case, the confidence {p_wgt_guess[mark]} is set to zero whenever
  the guessed position {p_i_guess[mark]} is outside the image's domain. */
  
void at_compute_target_images_and_masks
  ( image_list_t M_list,
    image_list_t W_list,
    options_t o,
    r2_t u_i[],
    r2_t v_i[],
    bool_t deform_images,
    image_list_t *G_list,
    image_list_t *V_list,
    char *frame_out_dir
  );
/* Computes the expected image {(*G_list)->imgs[mark]} and
  corresponding mask {(*V_list)->imgs[mark]}.  Allocates the lists
  {*G_list} and {*V_list}.

  If {deform_images} is TRUE the output images and masks
  {G_list,V_list} are deformed copies of the canonical images {M_list}
  and canonical masks {W_list}, according to the computed {u_i,v_i}.
  Othwerwise they are just undeformed copies of {M_list,W_list}. */

 void at_estimate_alphas_and_betas
  ( image_t I,
    image_list_t G_list,
    image_list_t V_list,
    options_t o,
    r2_t p_i[],
    double p_wgt[],
    bool_t debug,
    double alpha[],
    double beta[] );
/* Estimates the lighting parameters {alpha[mark]} and {beta[mark]} of
  every target in the frame image {I}, given the template
  {G_list->imgs[mark]}, its mask {V_list->imgs[mark]}, and its
  position {p_i[mark]}.  See {at_estimate_alpha_beta} below.

  The estimation is performed only if the mark has positive weight {p_wgt[mark]}.
  Otherwise {alpha} and {beta} are left unchanged. */

void at_locate_targets_in_frame
  ( image_t I,
    image_list_t G_list,
    image_list_t V_list,
    double alpha[],
    double beta[],
    options_t o,
    r2_t p_i_guess[],
    r2_t p_i_dev_guess,
    double p_wgt_guess[],
    r2_t p_i_found[], 
    r2_t *p_i_dev_found,
    double p_wgt_found[],
    char *frame_out_dir
  );
/* Locates each target {G_list->imgs[mark]} in the frame image {I},
  starting at position {p_i_guess[mark]}, with a priori confidence
  {p_wgt_guess[mark]} and position uncertainty {p_i_dev_guess[mark]}.
  Assumes that the targets are modified by the lighting parameters
  {alpha[mark]} and {beta[mark]}.

  Returns the found position in {p_i_found[mark]}, its confidence
  weight in {p_wgt_found[mark]}, and the estimated position error 
  (not considering bogus matches) in {*p_i_dev_found}.
  
  If {p_wgt_guess[mark]} is zero (meaning that the mark has been
  lost), returns {p_i_found[mark] = p_i_guess[mark]} and
  {p_wgt_found[mark] = 0}. */

void at_calibrate_camera_for_frame
  ( options_t o,
    camera_specs_t cspec,
    int ntargets,
    r3_t p_w[],
    r2_t p_i[],
    r2_t p_i_dev,
    double p_wgt[],
    camera_parameters_t cpar,
    char *frame_out_dir,
    int frame_num );
/* Calibrates the camera for some frame, given the world coordinates
  {p_w[mark]} of each target, its found position {p_i[mark]} in the
  frame image, the confidence weight {p_wgt[mark]} of the location,
  the a priori uncertainty {p_i_dev} on the image position, and the a
  priori distribution on the parameters implied by the {cspec} record.

  On input, assumes that {cpar} is the mean of {cspec}, as produced by
  {tf_get_mean_camera_parameters}.  On output, {cpar} contains the
  calibrated camera parameters.

  Writes files "guess.cpar" and "opt.cpar" to directory "{frame_out_dir}".
  The frame number {frame_num} is written into those files.

  Ig {o->useWeightsInCalib} is FALSE, sets all nonzero weights {p_wgt_guess[mark]}
  to 1.0 right away.  */

void at_extrapolate_camera_parameters_kalman 
  ( camera_specs_t global_cspec, 
    int order, 
    camera_parameters_t prev_cpar[], 
    kalman_parameters_t K[],
    bool_t debug,
    camera_parameters_t cpar,
    camera_specs_t cspec );
/* Extrapolates the camera parameters from previous parameter sets
  {prev_cpar[0..order-1]} using the Kalman models {K[0..n-1]} where {n
  = tf_camera_num_parameters}.  

  The parameter {order} must be positive. All the Kalman models must
  have the given {order} (or a smaller order).  The parameters are
  indexed as in {tf_camera_get_param}.  The extrapolated values are
  stored in {cspec} in interval format, namely each interval is {avg
  +/- 3*dev}. The mean values of those parameters are stored in {cpar}.

  The {global_cspec} defines the valid parameter ranges for the camera.
  Each extrapolated interval is clipped to the corresponding interval
  in {global_cspec} before being stored into {cspec}. In particular, if the interval
  in {global_cspec} is trivial, the parameter is considered fixed, and no 
  extrapolation is done. */

r2_t at_mark_position_uncertainty(kalman_parameters_t K_i[]);
/* Returns the residual uncertainty (standard deviation) in the X and Y image coordinates
   of mark positions, given the respective Kalman models {K_i[0]} and {K_i[1]}. */

void at_extrapolate_image_position_kalman
  ( int order, 
    r2_t *prev_p_i[],
    double *prev_p_wgt[],
    kalman_parameters_t K_i[],
    int mark,
    bool_t debug,
    r2_t *p_i_avg,
    double *p_wgt );
/* Extrapolates the image coordinates of mark number {mark} from previous 
  coordinates {prev_p_i[0..order-1][mark]}, using the Kalman models {K[0]}
  for the X coordinate and {K_i[1]} for the Y coordinate.  Both
  Kalman models must have the given {order} (or a smaller order).
  
  The predicted position is returned in {p_i_avg}.
  Also returns in {p_wgt} the minimum of the weights of all previous
  positions of this mark that were used in the prediction.  Thus,
  if any of those positions was outside the image (which implies
  zero weight) the value of {p_wgt} will be zero.
  (To get the uncertainty, use {at_mark_position_uncertainty}). */

void at_create_list_of_pyramids_from_marks (image_list_t ilist, image_pyramid_list_t plist, int redFactor, int nlevels);

image_pyramid_t at_create_diff_pyramid_from_image_pyramid (image_pyramid_t pI, char *type);

void at_locate_mark
  ( image_pyramid_t I_pyr,
    image_t G,
    image_t V,
    options_t o,
    r2_t p_i_guess,
    r2_t p_i_dev,
    double alpha,
    double beta, 
    char *mark_debug_name,
    r2_t *p_i_found,
    r2_t *p_i_dev_found,
    double *errorSqr, 
    image_pyramid_t I_pyr_deb1,
    image_pyramid_t I_pyr_deb2
  );

void at_write_debugging_images
  ( image_t I,
    int magnify,
    int ntargets,
    r2_t p_i[],
    image_list_t G_list,
    image_list_t V_list,
    int subsampling,
    double alpha[],
    double beta[],
    char *frame_out_dir,
    char *tag );
/* Writes two debugging images that show the positions
  {p_i[0..ntargets-1]} on the frame image {I}.  In the first image,
  the target images {G_list->imgs[0..ntargets-1]} are painted over {I}
  at those points, using
  {at_write_debugging_image_with_painted_targets} with the given
  {subsampling} and the given {alpha} and {beta} vectors
  (which may be NULL).  In the second image, the positions are shown by
  white crosses, using {at_write_debugging_image_with_crosses} with
  the given {magnify} parameter. The images are written to 
  {frame_out_dir}. */

void at_write_debugging_image_with_painted_targets
  ( image_t I,
    int ntargets,
    r2_t p_i[],
    image_list_t G_list,
    image_list_t V_list,
    int subsampling,
    double alpha[],
    double beta[],
    char *frame_out_dir,
    char *tag );
/* Writes a debuging image that consists of a copy of {I}, 
  in natural size, with the target images {G_list->imgs[0..ntargets-1]}
  painted over {I} at the points {p_i[0..ntargets-1]}.
  The target images are reduced in size by the given {subsampling}
  factor and have their colors adjusted by the parameters {alpha,beta}.
  If {alpha,beta} are null, assumed {alpha=1} and {beta=0} for all marks.
  The file name will be "{frame_out_dir}/{tag}_t.pgm". */

  void at_write_debugging_image_with_crosses
  ( image_t I,
    int magnify,
    int ntargets,
    r2_t p_i[],
    char *frame_out_dir,
    char *tag );
/* Writes a debuging image that consists of a copy of {I}, 
  magnified bt {magnify}, with crosses painted at the 
  positions that correspond to the points {p_i[0..ntargets-1]}.
  The file name will be "{frame_out_dir}/{tag}_c.pgm". */

void at_write_prev_cpar_image
  ( camera_parameters_t cpar,
    camera_parameters_t prev_cpar[],
    int kalman_order,
    targets_data_t tdat, 
    image_t I,
    char *frame_out_dir );
/* Writes a image with all the camera parameter guesses {prev_cpar} until the kalman order.*/  

void at_extract_feature_images (image_t R, targets_data_t tdat, int QSubsampling, image_list_t M_list);
/* For every target such that {tdat->conf[mark]} is positive, 
   grabs the corresponding piece of {I} and stores it as {M_list->imgs[mark]}. */

image_t at_extract_feature(image_t R, r2_t p_i, int hwx, int hwy, int QSubsampling);
  /* The parameters {hwx,hwy} are the half-width and half-height of the feature 
    in the reference image.  The extracted image will have size {(2*hwx+1)*QSubsampling} 
    by {(2*hwy+1)*QSubsampling}. */

void at_compute_auxiliary_mark_images
  ( image_t M,
    image_t W,
    int QSubsampling,
    double markBlur,
    r2_t u_i, r2_t v_i,
    image_t *G, image_t *V
  );

void at_estimate_alpha_beta
  ( image_t I, 
    image_t M, 
    image_t W, 
    int QSubsampling, 
    r2_t p_i, 
    double wgt,
    bool_t debug, 
    double *alpha, 
    double *beta
  );
/* Estimates the {*alpha} and {*beta} of a target by comparing its 
  maximum-contrast image {M} with the region of {I} centered at {p_i}.
  
  The new estimate is combined with the previous values of {*alpha}
  and {*beta} according to the weight {wgt}.  In particular, if {wgt}
  is 1.0 (or if the previous values are NAN), the old values are
  ignore, and the new estimates are stored into {*alpha,*beta}; if
  {wgt} is 0.0, the new estimates are ignored and the previous values
  are retained in {*alpha,*beta}. */

void at_mark_positions_and_weights_write
  ( int ntargets,
    r2_t p_i[],
    double p_wgt[],
    char *dir, 
    char *tag );
/* Writes the mark positions {p_i[0..ntargets]} and weights {p_wgt[0..ntargets]}
   to files "{dir}/p_i_{tag}.txt" and "{dir}/p_wgt_{tag}.txt", respectively. */

void at_mark_positions_write 
  ( int ntargets,
    r2_t p_i[],
    char *dir, 
    char *tag );
/* Writes the mark positions {p_i[0..ntargets]}
   to file "{dir}/p_i_{tag}.txt". */

  void at_mark_weights_write
  ( int ntargets,
    double p_wgt[],
    char *dir, 
    char *tag );
/* Writes the mark weights {p_wgt[0..ntargets]}
   to file "{dir}/p_wgt_{tag}.txt". */

void at_alpha_beta_write
  ( int ntargets,
    double alpha[],
    double beta[],
    char *dir, 
    char *tag );
/* Writes the lighting parameters {alpha[0..ntargets]} and {beta[0..ntargets]}
   to file "{dir}/alpha_beta_{tag}.txt". */

#endif
