#define _GNU_SOURCE
#include <stdio.h>
#include <math.h>
#include <r2.h>
#include <malloc.h>
#include <assert.h>
#include <tf_camera.h>
#include <tf_calib.h>
#include <tf_calib_data.h>
#include <tf_kalman.h>
#include <image.h>

void initialize_vetor (double *v, int n);

camera_parameters_t *alloc_all_camera_parameters (int number_of_frames);

int read_camera_parameters (char *arq_name, camera_parameters_t cpar);

void print_all_camera_parameters (camera_parameters_t cpar[], int number_of_frames);

void solve_system (double *A, double *x, double *b, int size);

double extrapolate_camera_parameter 
  ( camera_parameters_t cpar[], 
    kalman_parameters_t K,
    int iparam, 
    int frame
  );
/* Given a list of camera parameters {cpar[0..num_frames-1]}, 
  predicts the value of camera parameter {iparam} for frame number {frame},
  using the Kalman model {K}. Assumes that {frame >= K->order}. */

void show_extrapolated_positions 
 ( camera_parameters_t cpar[], 
   int number_of_params, 
   int number_of_frames, 
   kalman_parameters_t K[], 
   char *image_prefix, 
   char *out_dir,
   char *word_coords_name );
/* Shows the extapolated marks as crosses on the frame images.
  For each frame number {frame} in {order..number_of_frames-1},
  uses the Kalman model {K[iparam]} for each parameter {iparam}
  in {0..number_of_params} to predict the camera parameters {cpar_pred}
  of tyhat frame from {cpar[frame-1-j]} for each {j} in {0..order-1}.

  Then uses {cpar_pred} and the world positions {p_w[0..number_of_marks-1]}
  to compute the image positions of all marks, and draws crosses at those
  positions on top of frame number {frame}.  */


void print_cross_from_cpar ( 
   camera_parameters_t cpar, 
   int frame, 
   char *image_prefix, 
   char *out_dir,
   char *word_coords_name );


kalman_parameters_t build_kalman_model 
  ( camera_parameters_t *cpar, 
    int number_of_params,   
    int number_of_frames,  
    int iparam,  
    int order );
/* Build a Kalman model of given {order} for camera parameter number {iparam},
  given camera parameters {cpar[0..number_of_frames-1]}. */

int main (int argc, char *argv[])
{
   char *prefix = argv[1];
   int number_of_frames = atoi(argv[2]);
   int max_order = atoi(argv[3]);  assert(max_order >= 0);
   char *image_prefix = argv[4];
   char *cpar_filename = argv[5];
   char *word_coords_name = argv[6];
   bool_t show_result_images = (atoi(argv[7]) != 0);


   char *out_dir = "kalman_out";

   int number_of_params = tf_camera_num_parameters;

   camera_parameters_t *cpar = alloc_all_camera_parameters(number_of_frames);

   int i;
   for (i = 0; i < number_of_frames; i++) {
       char *in_name = NULL;
       asprintf(&in_name, "%s/%05d/%s", prefix, i+1, cpar_filename);
       read_camera_parameters (in_name, cpar[i]);
       free(in_name);
   }

   //print_all_camera_parameters (cpar, number_of_frames);

   kalman_parameters_t K[number_of_params];
   
   int order;
   for (order = 0; order <= max_order; order++) {
     int iparam;
     for (iparam = 0; iparam < number_of_params; iparam++) {
       K[iparam] = build_kalman_model (cpar, number_of_params, number_of_frames, iparam, order);
       char *pname = tf_camera_param_name(iparam);
       fprintf(stderr, "Kalman coefficients for parameter %d (%s) \n", iparam, pname);
       tf_kalman_show_parameters (K[iparam], stderr);

       char *file_name_kalman = NULL;
       asprintf(&file_name_kalman, "%s/%s_%02d.kpar", out_dir, pname, order);
       FILE *file_kalman = open_write(file_name_kalman, TRUE);
       tf_kalman_write_parameters (K[iparam], file_kalman);
       fclose(file_kalman);
       free(file_name_kalman);
     }
   }
   if (show_result_images) {
     /* Show extrapolated results for max order: */
     show_extrapolated_positions 
      ( cpar, 
        number_of_params, 
        number_of_frames, 
        K, 
        image_prefix, 
        out_dir,
        word_coords_name );
   }

   return 0; 
}

int read_camera_parameters (char *arq_name, camera_parameters_t cpar)
{
  FILE *f = fopen(arq_name, "r");
  int iframe = tf_read_camera_parameters(f, cpar);
  fclose(f);
  return iframe;
}  

void print_all_camera_parameters (camera_parameters_t cpar[], int number_of_frames)
{
   int frame;
   for (frame = 0; frame < number_of_frames; frame++) {
      tf_write_camera_parameters(stderr, frame, cpar[frame]);
   }
}

kalman_parameters_t build_kalman_model 
  ( camera_parameters_t *cpar, 
    int number_of_params,  
    int number_of_frames,  
    int iparam,  
    int order )
{
  int i, j, frame;

  kalman_parameters_t K = tf_kalman_new_parameters (order);

  /*Find coefficient {c[i]} of data {p[frame-1-i]}, for {i} in {0..order-1}. */
  /*Coefficient {c[order]} is the independent term. */
  double *c = (double *)malloc((order+1) * sizeof(double));

  if (order > 0) {
    /*Find coefficients {c[0..order]} by solvin a least squares system: */

    double *A = (double *)malloc(((order+1)*(order+1)) * sizeof(double));
    double *b = (double *)malloc((order+1) * sizeof(double));
 
    /*We first build a model subtracting the previous value to avoid large numbers. */
    initialize_vetor (A, (order+1)*(order+1));
    initialize_vetor (b, (order+1));
    for (frame = order; frame < number_of_frames; frame++) {
      double vref = tf_camera_get_param(cpar[frame], iparam, NAN);
      double vprev = tf_camera_get_param(cpar[frame-1], iparam, vref);
      for (i = 0; i < order; i++) {
        double qi = tf_camera_get_param(cpar[frame-i-1],iparam, vref) - vprev;
	for (j = 0; j < order; j++) {
	  double qj = tf_camera_get_param(cpar[frame-j-1],iparam, vref) - vprev;
	  A[i*(order+1) + j] += (qi * qj);
	  //fprintf(stderr, "pos = %d\n", i*order + j);
	} 	
	double qf =  tf_camera_get_param(cpar[frame], iparam, vref) - vprev;
	b[i] += qf * qi; 
      }   
    }

    /*Adding a col of 1's to allow for constant bias. */
    /*Adding a row of 1's to force the sum of the coefficients to be 0. */
    fprintf(stderr, "System:\n");
    for (i = 0; i <= order; i++) {
      for (j = 0; j <= order; j++) {
	int ij = i*(order+1) + j;
	if ((i == order) || (j == order)) { A[ij] = 1.0; }
	fprintf(stderr, " %12.2f", A[ij]);
      }
      if (i == order) { b[i] = 0.0; }
      fprintf(stderr, "   %12.2f\n", b[i]);
    }
    /*end*/

    solve_system(A, c, b, order+1);

    /* Adjust coefficient {c[0]} to account for the subtraction of frame {i-1}: */
    c[0] = c[0] + 1.0;
  }
    
  char *pname = tf_camera_param_name(iparam);
  fprintf(stderr, "Coeficientes do iparam %d (%s) = (", iparam, pname);
  double coef_sum = 0.0;
  for (i = 0; i < order; i++) {
    K->coeffs[i] = c[i];
    fprintf(stderr, " %+9.6f*Z[k-%d]", c[i], i+1); 
    coef_sum += c[i];
  }
  fprintf(stderr, " %+9.6f", c[order]);
  fprintf(stderr, " )\n");
  fprintf(stderr, "Soma dos coeficientes dependentes = %f\n", coef_sum);
      
  double data[order];
  double avg, dev, npred;
  
  /* Compute the average residual: */
  double sume = 0.0;
  npred = 0;
  K->avg = 0.0;
  for (frame = order; frame < number_of_frames; frame++) {
    double vref = tf_camera_get_param(cpar[frame], iparam, NAN);
    for (j = 0; j < order; j++) {
      data[j] = tf_camera_get_param(cpar[frame-j-1], iparam, vref);
    }
    tf_kalman_predict_parameter(K, data, &avg, &dev);
    double e = tf_camera_get_param(cpar[frame], iparam, vref) - avg;
    sume += e;
    npred++;
  }
  K->avg = sume/npred;

  /* Compute the deviation of the residual: */
  double sume2 = 0.0;
  npred = 0;
  for (frame = order; frame < number_of_frames; frame++) {
    double vref = tf_camera_get_param(cpar[frame], iparam, NAN);
    for (j = 0; j < order; j++) {
    	data[j] = tf_camera_get_param(cpar[frame-j-1], iparam, vref);
    }
    tf_kalman_predict_parameter(K, data, &avg, &dev);
    double e = tf_camera_get_param(cpar[frame], iparam, vref) - avg;
    sume2 += e*e;
    npred++;
  }
  K->dev = sqrt(sume2/(npred-1));
  if (K->dev < 0.0) {
    fprintf(stderr, "Fudging the uncertainty.\n");
    K->dev = 0.0;
  }
 
  return K;
}

double extrapolate_camera_parameter 
  ( camera_parameters_t cpar[],
    kalman_parameters_t K,
    int iparam,
    int frame )
{
  int j;

  int order = K->order;
  double data[order];
  double vref = tf_camera_get_param(cpar[frame], iparam, NAN);

  double avg, dev;
  
  for (j = 0; j < order; j++) {
    data[j] = tf_camera_get_param(cpar[frame-j-1], iparam, vref);
  }
  tf_kalman_predict_parameter(K, data, &avg, &dev);
  return avg;
} 

void show_extrapolated_positions 
 ( camera_parameters_t cpar[], 
   int number_of_params, 
   int number_of_frames,
   kalman_parameters_t K[], 
   char *image_prefix, 
   char *out_dir,
   char *word_coords_name )
{
  int frame, iparam;

  double params[number_of_params];

  for (frame = K[0]->order; frame < number_of_frames; frame++) {

    for (iparam = 0; iparam < number_of_params; iparam++) {
         params[iparam]= extrapolate_camera_parameter (cpar, K[iparam], iparam, frame); 
 
      //camera_parameters_t p_ext = extrapolate_camera_parameter (cpar, K, iparam, frame);
      //image_print_cross (img, 5, p_ext);
      //image_print_dot (img, 2, p[frame][iparam]);
    }
 
    camera_parameters_t cpar_extrapolated = (camera_parameters_t)malloc(sizeof(struct _camera_parameters_t));
 
    tf_camera_set_params_from_vector (cpar_extrapolated, params);

    print_cross_from_cpar (cpar_extrapolated, frame, image_prefix, out_dir, word_coords_name);

  }
}


void print_cross_from_cpar ( 
   camera_parameters_t cpar, 
   int frame, 
   char *image_prefix, 
   char *out_dir,
   char *word_coords_name ) {
 
  int i;

  
  char *frame_name = NULL;
  asprintf(&frame_name, "%s%05d.pgm", image_prefix, frame+1);
  image_t img = image_read_pgm (frame_name);
  free(frame_name);

  int n_p_w;
  calibration_data_t cdat = (calibration_data_t)notnull(malloc(sizeof(struct _calibration_data_t)), "no mem");
  tf_read_world_positions (word_coords_name, &n_p_w, &(cdat->world_coords));

  for (i = 0; i < n_p_w; i++) {

    r2_t p_i = tf_world_coords_to_image_coords (cpar, cdat->world_coords[i]);

    int mark_radius = 5;

    if ( (p_i.c[0] > 5) && 
         (p_i.c[1] > 5) && 
         (p_i.c[0] < img->nx - mark_radius) && 
         (p_i.c[1] < img->ny - mark_radius) ) {
      image_print_cross (img, mark_radius, p_i);
    }
  }

  char *cross_fname = NULL;
  asprintf(&cross_fname, "%s/out%05d_cross.pgm", out_dir, frame+1);
  image_write_pgm (img, PGM_RAW_MAGIC, MAXVAL, cross_fname);
  free(cross_fname);
  image_free(img); 
}


camera_parameters_t *alloc_all_camera_parameters (int number_of_frames) 
{
    camera_parameters_t *cpar;
    int frame;

    cpar = (camera_parameters_t *)malloc(number_of_frames * sizeof(camera_parameters_t));
    if (cpar == NULL){
        fprintf( stderr, "error: allocating array" );
        return NULL;
    }

    for (frame = 0; frame < number_of_frames; frame++) {
      cpar[frame] = (camera_parameters_t)malloc(sizeof(struct _camera_parameters_t));
      if ( cpar[frame] == NULL ){
            fprintf( stderr, "error: allocating array" );
            return NULL;
        }
    }
    return cpar;
}

/*Solves Ax = b system, where {x} is the unknows coefficients */
void solve_system (double *A, double *x, double *b, int size)
{
    double *ftmp;

    ftmp = (double *)malloc((size*size) * sizeof(double));

    rmxn_inv (size, A, ftmp);

    rmxn_map_col (size, size, ftmp, b, x);

    free(ftmp);
}

void initialize_vetor (double *v, int n)
{
   int i;
   for (i = 0; i < n; i++) {
      v[i] = 0.0;
   }
}








