#include <r2.h>
#include <stdio.h>
#include <malloc.h>
#include <assert.h>
#include <tf_calib.h>
#include <tf_calib_data.h>
#include <tf_kalman.h>
#include <image.h>

void initialize_vetor (double *v, int n);

r2_t ** alloc_all_points (int number_of_marks, int number_of_frames);

int read_points (char *arq_name, r2_t **p);

void print_all_points (r2_t **p, int number_of_marks, int number_of_frames);

void solve_system (double *A, double *x, double *b, int size);

r2_t extrapolate_mark_position (r2_t **p, kalman_parameters_t K[], int mark, int frame);

void show_extrapolated_positions 
 ( r2_t **p, 
   int number_of_marks, 
   int number_of_frames, 
   kalman_parameters_t K[], 
   char *image_prefix, 
   char *out_dir );
/* Shows the extapolated marks as crosses on the frame images. */

kalman_parameters_t build_kalman_model 
(r2_t **p, 
 int number_of_marks,  
 int number_of_frames,  
 int eixo,  
 int order);

double *compute_x_y_covariance_matrix 
  ( r2_t **p,
    kalman_parameters_t K[],
    int number_of_marks,
    int number_of_frames );

int main (int argc, char *argv[])
{
   char *prefix = argv[1];

   int number_of_frames = atoi(argv[2]);
   
   int number_of_marks = atoi(argv[3]);
   
   int max_order = atoi(argv[4]);
  
   char *image_prefix = argv[5];

   bool_t show_result_images = (atoi(argv[6]) != 0);

   assert(max_order >= 0);

   char *out_dir = "kalman_out";

   r2_t **p = alloc_all_points (number_of_marks, number_of_frames);

   int i;
   for (i = 0; i < number_of_frames; i++) {
       char *in_name = NULL;
       asprintf(&in_name, "%s/%05d/p_i.txt", prefix, i+1);
       int n_p = read_points (in_name, &(p[i]));
       assert (number_of_marks = n_p);
       free(in_name);
   }

   //print_all_points (p, number_of_marks, number_of_frames);
 
   kalman_parameters_t K[2];
   
   int order;
   for (order = 0; order <= max_order; order++) {
     int eixo;
     for (eixo = 0; eixo < 2; eixo++) {
       K[eixo] = build_kalman_model (p, number_of_marks, number_of_frames, eixo, order);
       fprintf(stderr, "Kalman coefficients for axis %d \n", eixo);
       tf_kalman_show_parameters (K[eixo], stderr);

       char *file_name_kalman = NULL;
       asprintf(&file_name_kalman, "%s/%s_i_%02d.kpar", out_dir, (eixo == 0 ? "x" : "y"), order);
       FILE *file_kalman = open_write(file_name_kalman, TRUE);
       tf_kalman_write_parameters (K[eixo], file_kalman);
       fclose(file_kalman);
       free(file_name_kalman);
     }
     double *E = compute_x_y_covariance_matrix (p, K, number_of_marks, number_of_frames);
     free(E);
   }
   if (show_result_images) {
     /* Show extrapolated results for max order: */
     show_extrapolated_positions (p, number_of_marks, number_of_frames, K, image_prefix, out_dir);
   }

   return 0; 
}

int read_points (char *arq_name, r2_t **p)
{
  int nmarks;
  tf_read_image_positions (arq_name, &nmarks, &(*p));
  return nmarks;
}  

void print_all_points (r2_t **p, int number_of_marks, int number_of_frames)
{
   int mark, frame;
   for (frame = 0; frame < number_of_frames; frame++) {
      fprintf(stderr, "frame = %d\n", frame);
      for (mark = 0; mark < number_of_marks; mark++) {
          fprintf(stderr, "mark = %d, p_x = %f, p_y = %f\n", mark, p[frame][mark].c[0], p[frame][mark].c[1]);
      }
   }
}

kalman_parameters_t build_kalman_model 
(r2_t **p, 
 int number_of_marks,  
 int number_of_frames,  
 int eixo,  
 int order)
{
  int i, j, frame, mark;

  kalman_parameters_t K = tf_kalman_new_parameters (order);

  /*Find coefficient {c[i]} of data {p[frame-1-i]}, for {i} in {0..order-1}. */
  /*Coefficient {c[order]} is the independent term. */
  double *c = (double *)malloc((order+1) * sizeof(double));

  if (order > 0) {
    /*Find coefficients {c[0..order]} by solvin a least squares system: */

    double *A = (double *)malloc(((order+1)*(order+1)) * sizeof(double));
    double *b = (double *)malloc((order+1) * sizeof(double));
 
    /*We first build a model subtracting the previous position to avoid large numbers. */
    initialize_vetor (A, (order+1)*(order+1));
    initialize_vetor (b, (order+1));
    for (i = 0; i < order; i++) {
      for (j = 0; j < order; j++) {
	for (frame = order; frame < number_of_frames; frame++) {
	  for (mark = 0; mark < number_of_marks; mark++) {
	    double qirs = p[frame-i-1][mark].c[eixo] - p[frame-1][mark].c[eixo];
	    double qjrs = p[frame-j-1][mark].c[eixo] - p[frame-1][mark].c[eixo];
	    A[i*(order+1) + j] += (qirs * qjrs);
	    //fprintf(stderr, "pos = %d\n", i*order + j);
	  } 	
	} 
      }
      for (frame = order; frame < number_of_frames; frame++) {
	for (mark = 0; mark < number_of_marks; mark++) {
	  double qirs = p[frame-i-1][mark].c[eixo] - p[frame-1][mark].c[eixo];
	  double qrs =  p[frame][mark].c[eixo] - p[frame-1][mark].c[eixo];
	  b[i] += qrs * qirs; 
	}
      }   
    }

    /*Adding a col of 1's to allow for constant bias. */
    /*Adding a row of 1's to force the sum of the coefficients to be 0. */
    fprintf(stderr, "System:\n");
    for (i = 0; i <= order; i++) {
      for (j = 0; j <= order; j++) {
	int ij = i*(order+1) + j;
	if ((i == order) || (j == order)) { A[ij] = 1.0; }
	fprintf(stderr, " %12.2f", A[ij]);
      }
      if (i == order) { b[i] = 0.0; }
      fprintf(stderr, "   %12.2f\n", b[i]);
    }
    /*end*/

    solve_system(A, c, b, order+1);

    /* Adjust coefficient {c[0]} to account for the subtraction of {p[i-1]}: */
    c[0] = c[0] + 1.0;
  }
    
  fprintf(stderr, "Coeficientes do eixo %d = (", eixo);
  double coef_sum = 0.0;
  for (i = 0; i < order; i++) {
    K->coeffs[i] = c[i];
    fprintf(stderr, " %+9.6f*Z[k-%d]", c[i], i+1); 
    coef_sum += c[i];
  }
  fprintf(stderr, " %+9.6f", c[order]);
  fprintf(stderr, " )\n");
  fprintf(stderr, "Soma dos coeficientes dependentes = %f\n", coef_sum);
      
  double data[order];
  double avg, dev, npred;
  
  /* Compute the average residual: */
  double sume = 0.0;
  npred = 0;
  K->avg = 0.0;
  for (frame = order; frame < number_of_frames; frame++) {
    for (mark = 0; mark < number_of_marks; mark++) {
      for (j = 0; j < order; j++) {
	data[j] = p[frame-j-1][mark].c[eixo];
      }
      tf_kalman_predict_parameter(K, data, &avg, &dev);
      double e = p[frame][mark].c[eixo] - avg;
      sume += e;
      npred++;
    } 
  }
  K->avg = sume/npred;

  /* Compute the deviation of the residual: */
  double sume2 = 0.0;
  npred = 0;
  for (frame = order; frame < number_of_frames; frame++) {
    for (mark = 0; mark < number_of_marks; mark++) {
      for (j = 0; j < order; j++) {
	data[j] = p[frame-j-1][mark].c[eixo];
      }
      tf_kalman_predict_parameter(K, data, &avg, &dev);
      double e = avg - p[frame][mark].c[eixo];
      sume2 += e*e;
      npred++;
    } 
  }
  K->dev = sqrt(sume2/(npred-1));
  if (K->dev < 0.0) {
    fprintf(stderr, "Fudging the uncertainty.\n");
    K->dev = 0.0;
  }
 
  return K;
}

r2_t extrapolate_mark_position (r2_t **p, kalman_parameters_t K[], int mark, int frame)
{
  int j;
  r2_t p_pred;

  int order = K[0]->order;
  assert(K[0]->order == K[1]->order);
  double data[order];

  int eixo;
  for (eixo = 0; eixo < 2; eixo++) {
    double dev;
    for (j = 0; j < order; j++) {
      data[j] = p[frame-j-1][mark].c[eixo];
    }
    tf_kalman_predict_parameter(K[eixo], data, &(p_pred.c[eixo]), &dev);
  }
  return p_pred;
} 

void show_extrapolated_positions 
( r2_t **p, 
  int number_of_marks, 
  int number_of_frames, 
  kalman_parameters_t K[], 
  char *image_prefix, 
  char *out_dir )
{
  int order = K[0]->order;
  assert(K[0]->order == K[1]->order);
  int frame, mark;

  for (frame = order; frame < number_of_frames; frame++) {

    char *frame_name = NULL;
    asprintf(&frame_name, "%s%05d.pgm", image_prefix, frame+1);
    image_t img = image_read_pgm (frame_name);
    free(frame_name);

    for (mark = 0; mark < number_of_marks; mark++) {
      r2_t p_pred = extrapolate_mark_position (p, K, mark, frame);

      fprintf(stderr, "p_pred = ( %f %f )\n", p_pred.c[0], p_pred.c[1]);

      image_print_cross (img, 5, p_pred);
      //image_print_dot (img, 2, p[frame][mark]);
    }
 
    char *cross_fname = NULL;
    asprintf(&cross_fname, "%s/out%05d_cross.pgm", out_dir, frame+1);
    image_write_pgm (img, PGM_RAW_MAGIC, MAXVAL, cross_fname);
    free(cross_fname);
    image_free(img); 
  }
}

double *compute_x_y_covariance_matrix 
  ( r2_t **p,
    kalman_parameters_t K[],
    int number_of_marks,
    int number_of_frames )
{  
   int order = K[0]->order;
   assert(K[0]->order == K[1]->order);
   int r, s, frame, mark;
   int num_eixos = 2;
   double V[num_eixos*num_eixos];

   for (r = 0; r < num_eixos; r++) {
      for (s = 0; s < num_eixos; s++) {
         double sum = 0.0;
         int npred = 0;
         for (frame = order; frame < number_of_frames; frame++) {
            for (mark = 0; mark < number_of_marks; mark++) {
              r2_t p_pred = extrapolate_mark_position (p, K, mark, frame);
              double dr = p[frame][mark].c[r] - p_pred.c[r];
              double ds = p[frame][mark].c[s] - p_pred.c[s];
	      sum += dr*ds;
	      npred++;
            }
         }
	 V[r*num_eixos + s] = sum/npred;
      }
   }

   fprintf(stderr, "V[0][0] = %f, V[0][1] = %f\n", V[0], V[1]);
   fprintf(stderr, "V[1][0] = %f, V[1][1] = %f\n", V[2], V[3]);

   double *E;

   E = (double *)malloc((num_eixos*num_eixos) * sizeof(double));

   rmxn_inv (num_eixos, V, E);

   fprintf(stderr, "E[0][0] = %f, E[0][1] = %f\n", E[0], E[1]);
   fprintf(stderr, "E[1][0] = %f, E[1][1] = %f\n", E[2], E[3]);

   return E;
}

r2_t ** alloc_all_points (int number_of_marks, int number_of_frames) 
{
    r2_t **m;
    int  frame;

    m = (r2_t **)malloc(number_of_frames * sizeof(r2_t *));
    if ( m == NULL ){
        fprintf( stderr, "error: allocating array" );
        return NULL;
    }

    for (frame = 0; frame < number_of_frames; frame++){
        m[frame] = (r2_t *)malloc(number_of_marks * sizeof(r2_t));
        if ( m[frame] == NULL ){
            fprintf( stderr, "error: allocating array" );
            return NULL;
        }
    }
    return m;
}

/*Solves Ax = b system, where {x} is the unknows coefficients */
void solve_system (double *A, double *x, double *b, int size)
{
    double *ftmp;

    ftmp = (double *)malloc((size*size) * sizeof(double));

    rmxn_inv (size, A, ftmp);

    rmxn_map_col (size, size, ftmp, b, x);

    free(ftmp);
}

void initialize_vetor (double *v, int n)
{
   int i;
   for (i = 0; i < n; i++) {
      v[i] = 0.0;
   }
}
