#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <image.h>
#include <tf_calib.h>
#include <r2x2.h>
#include <r2.h>
#include "lk.h" 
#include "lk_error_generic.h" 
#include "lk_error_STB.h"
#include "track_parameters.h"
#include "numeric.h" 
#include <argparser.h>
#include <jsmath.h>
#include <affirm.h>
#include <bool.h>



#define PROG_NAME "test_generic"
#define PROG_DESC "program to test the best discrepancy function by choosing: \
                    -smoothing;\
                    -preprocessing;\
                    -interpolation;\
                    -integration"
#define PROG_VERS "1.0"

#define test_generic_C_COPYRIGHT "Copyright © 2008 by Rodrigo Minetto (UNICAMP)"

/* 
** Written by Rodrigo Minetto 
**
** Copyright notice:
**
** Permission to use this software for any purpose is hereby granted,
** provided that any substantial copy or mechanically derived version
** of this file that is made available to other parties is accompanied
** by this copyright notice in full, and is distributed under these same
** terms. 
**
** DISCLAIMER: This software is provided "as is" with no explicit or
** implicit warranty of any kind.  Neither the authors nor their
** employers can be held responsible for any losses or damages
** that might be attributed to its use.
**
** End of copyright notice.
*/

#define PROG_HELP \
  PROG_NAME " \\\n" \
  "  -I {FRAME_IMG_NAME} \\\n" \
  "  -M {TARGET_IMG_NAME} \\\n" \
  "  -W {WEIGHTS_IMG_NAME} \\\n" \
  "  -minimizer {MINIMIZER_NAME} \\\n" \
  "  -interpolation {INTERPOLATION_NAME} \\\n" \
  "  -step_factor {MIMINIZATION_STEP_FACTOR} \\\n" \
  "  -out {OUT_NAME}"

#define PROG_INFO \
  "NAME\n" \
  "  " PROG_NAME " - " PROG_DESC "\n" \
  "\n" \
  "SYNOPSIS\n" \
  PROG_HELP  "\n" \
  "\n" \
  "DESCRIPTION\n" \
  "  Reads two or more frames of a video.  For each pair of consecutive" \
  " frames, computes the optical flow at certain pixels of the first" \
  " frame, and writes that flow to a file called \"{OUT_PREFIX}_{FRAME_NUM}.opf\".\n" \
  "\n" \
  "OPTIONS\n" \
  "  -framePrefix {FRAME_PREFIX}\n" \
  "    This mandatory parameter specifies the prefix of frame file names.\n" \
  "\n" \
  "  -frames {FIRST_FR} {NUM_FR} {STEP_FR}\n" \
  "    This mandatory parameter specifies the set of frame numbers to" \
  " process, namely {FIRST_FR}, {FIRST_FR+STEP_FR}, {FIRST_FR+2*STEP_FR}," \
  " ..., {FIRST_FR+(NUM_FR-1)*STEP_FR}.\n" \
  "\n" \
  "  -frameExtension {FR_EXT}\n" \
  "    This optional parameter specifies the extension (and format) of" \
  " the frame files.  At present, only the \"pgm\" extension is supported.  The default is \"pgm\".\n" \
  "\n" \
  "  -pixelStep {STEP_PIX}\n" \
  "    This optional parameter specifies the spacing between the" \
  " `source' pixels in the first frame of each pair.  The default is 1" \
  " (the flow is computed for all pixels).\n" \
  "\n" \
  "  -altKLT\n" \
  "    This optional flag requests the alternative implementation" \
  " of the KLT algorithm.  The default is Bicherfield's implementation.\n" \
  "\n" \
  "  -outPrefix {OUT_PREFIX}\n" \
  "    This mandatory parameter specifies the common prefix for" \
  " all output file names.\n" \
  "\n" \
  "DOCUMENTATION OPTIONS\n" \
  argparser_help_info_HELP_INFO "\n" \
  "\n" \
  "SEE ALSO\n" \
  "  opflow_analyze(1), opflow_segment(1).\n" \
  "\n" \
  "AUTHOR\n" \
  "  Created sept/2008 by Rodrigo Minetto, IC-UNICAMP.\n" \
  "\n" \
  "MODIFICATION HISTORY\n" \
  "\n" \
  "WARRANTY\n" \
  argparser_help_info_NO_WARRANTY "\n" \
  "\n" \
  "RIGHTS\n" \
  "  " test_generic_C_COPYRIGHT ".\n" \
  "\n" \
  argparser_help_info_STANDARD_RIGHTS



typedef struct options_t
  {
    char *I;             /* Frame image filename.  */
    char *M;             /* Mask image filename.   */ 
    char *W;             /* Weight image filename. */
    char *minimizer;     /* Minimizer type. */
    char *interpolation; /* Interpolation function type. */
    double step_factor;  /* Minimization step factor. */
    char *out;           /* Output filename. */
  } options_t;

lk_interpolation_function *parse_interpolation (char *interpolation, int *iord);

options_t *parse_options(int argc, char **argv);

int main (int argc, char *argv[])
{
  /*Verify memory*/
  void *trash = malloc(1);
  struct mallinfo info;
  int MemDinInicial, MemDinFinal;
  /*End*/

  free(trash);
  info = mallinfo();
  MemDinInicial = info.uordblks;

  
  options_t *o = parse_options(argc, argv);

  image_t I = image_read_pgm (o->I);
  image_t M = image_read_pgm (o->M);
  image_t W = image_read_pgm (o->W);
  
  /* Choose the minimizer: */
  numeric_minimizer_t *minize = parse_minimizer_arg(o->minimizer);

  /* Choose the interpolation: */
  int iord;
    
  lk_interpolation_function *interpolation = parse_interpolation (o->interpolation, &iord);

  /* Choose the minimization step factor: */
  double step_factor = o->step_factor;
  printf("step_factor = %8.6f\n", step_factor);

  /* Output filename prefix: */
  char *outname = o->out;

  /* Choose the central parameters: */
  int nz = 2;
  double z[nz];
  r2_t p_i;
  p_i.c[0] = I->nx/2.0;
  p_i.c[1] = I->ny/2.0;
  lk_gather_params (nz, z, p_i);

  /* Choose the number of plotting steps: */ 
  double alpha = 1.0;
  double beta = 0.0; 
  int niter = 10;

  int color_flag = 1;

  double sigma = 1.0;

  int hgn, i;
  double *gw;

  image_compute_gaussian_weights(sigma, &gw, &hgn);

  for (i = 0; i <= 2*hgn; i++) {printf("%d: %f\n", i, gw[i]);}

  printf("%s %s %s %s %s iord %d ######## %s\n", o->I, o->M, o->W, o->minimizer, o->interpolation, iord, o->out);

  fprintf(stderr, "before generic\n");
  auto void goal_func(int nz, double z[], double *Q, double *J, double *H);

  void goal_func(int nz, double z[], double *Q, double *J, double *H)
     { assert(nz == 2);
       r2_t q_i;
       lk_scatter_params (nz, z, &q_i);
       /*TODO - ACERTAR PARAMETROS*/
       lk_error_generic (  I, 
                           M, 
                           W, 
                           MARK_SUBSAMPLING, 
                           alpha, 
                           beta, 
                           color_flag, 
                           q_i, 
                           Q,  
                           J, 
                           H, 
                           interpolation,  
                           iord, 
                           gw, 
                           hgn  ); 
       //lk_error_STB (I, M, W, MARK_SUBSAMPLING, alpha, beta, q_i, Q, J, H);

     }
  fprintf(stderr, "after generic\n");

  track_parameters_print(stderr);

  lk_gather_params (nz, z, p_i);
  double plot_step[nz];
  lk_choose_plot_steps(nz, plot_step);
  double rplot = 5.0; /* Number of pixels on each side of the center in plot. */ 
  int N = (int)(rplot / fmin(plot_step[0], plot_step[1]));

   /* Max displacement allowed from initial guess [pixels]: */
   /* !!! SHOULD COME FROM A COMMON INTERFACE */
   double max_d = 2.0;
   double max_disp[nz];
   int k;
   for (k = 0; k < nz; k++) { max_disp[k] = max_d; } 
   double min_step = 0.001/max_d; /* Convergence step length (rel to {max_disp}): */

   fprintf(stderr, "before numeric\n");
   numeric_plot_2D_goal_func_and_derivatives
    ( outname, 
      nz, z,
      plot_step,  
      goal_func, 
      *minize,
      step_factor,
      max_disp,
      niter,
      min_step,
      N
    );
   fprintf(stderr, "after numeric\n");

  image_free (I);
  image_free (M);
  image_free (W);

  info = mallinfo();
  MemDinFinal = info.uordblks;
  if (MemDinInicial!=MemDinFinal)
    fprintf(stderr, "\n\nMain loop: Dinamic memory was not completely deallocated (%d, %d)\n", MemDinInicial,MemDinFinal);
  return 0;
} 

lk_interpolation_function *parse_interpolation (char *interpolation, int *iord)
{
  if (strcmp(interpolation, "linear") == 0) {
    *iord = 1;
     return &interpol_linear;
  }
  else if (strcmp(interpolation, "2B") == 0) {
     *iord = 2;
     return &interpol_spline_2_B;
  }
  else if (strcmp(interpolation, "3A") == 0) {
     *iord = 3;
     return &interpol_spline_3_A;
  }
  else if (strcmp(interpolation, "3B") == 0) {
     *iord = 3;
     return &interpol_spline_3_B;
  }
  else if (strcmp(interpolation, "3C") == 0) {
     *iord = 3; 
     return &interpol_spline_3_C;
  }
  else {
     fprintf(stderr, "error: unrecognized interpolation option.\n");
     exit(1);
  }
}

options_t *parse_options(int argc, char **argv)
  {
    argparser_t *pp = argparser_new(stderr, argc, argv);
    
    argparser_set_help(pp, PROG_NAME " version " PROG_VERS ", usage:\n" PROG_HELP);

    argparser_set_info(pp, PROG_INFO);
    argparser_process_help_info_options(pp);

    /* Parse keyword arguments: */
     
    options_t *o = (options_t *)notnull(malloc(sizeof(options_t)), "no mem");
    
    argparser_get_keyword(pp, "-I");
    o->I = argparser_get_next(pp);

    argparser_get_keyword(pp, "-M");
    o->M = argparser_get_next(pp);
    
    argparser_get_keyword(pp, "-W");
    o->W = argparser_get_next(pp);
    
    argparser_get_keyword(pp, "-minimizer");
    o->minimizer = argparser_get_next(pp);
 
    argparser_get_keyword(pp, "-interpolation");
    o->interpolation = argparser_get_next(pp);

    argparser_get_keyword(pp, "-step_factor");
    o->step_factor = argparser_get_next_double (pp, -HUGE, HUGE);

    argparser_get_keyword(pp, "-out");
    o->out = argparser_get_next(pp);    
       
    argparser_skip_parsed(pp);

    /* Parse positional arguments: */
     
    /* Check for spurious args: */
    argparser_finish(pp);

    return o;
  }
