#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <image.h>
#include <tf_calib.h>
#include <r2.h>
#include <jsrandom.h>
#include "lk.h"
#include "lk_error_M.h"
#include "track_parameters.h"
#include "at_process_frame.h"
#include "lk_error_STB.h"
#include "numeric.h"



void fit_line_with_lmdif (r2_t *p, 
                          int npoints, 
                          double A_guessed,
                          double B_guessed,
                          double *A_calc,
                          double *B_calc,
                          double sigma, 
                          double lambda, 
                          double epsilon);
/* Finds the straith line coefficients {A_calc, B_calc} to the data points 
   {p[0..npoints-1]} by non-linear  Bayesian likelyhood maximization. Assumes 
   a Y probability distribution ProbCZ with parameters {sigma, lambda, epsilon}.
   The initial guess is {A_guessed, B_guessed}.  
 */


void optimization_error
  ( doublereal *err,
    r2_t *p,
    int npoints,
    double A,
    double B,
    double sigma,
    double lambda,
    double epsilon );

void fit_line_with_weighted_least_square_opt 
  ( r2_t *p, 
    int npoints, 
    double sigma,  
    double lambda,
    double epsilon,
    double log_prob (double A, double B, double x, double y),
    double *y_orig,  
    double *y_calc );

double err_out (double lambda, double y);

double err_in (double sigma, double lambda, double epsilon, double yc, double y);

double ProbCZ (double sigma, double lambda, double epsilon, double yc, double y);

double ProbGG (double sigma, double lambda, double epsilon, double yc, double y);

double log_gaussian (double delta, double sigma);
  /*Logarithm of the Gaussian probability distribution with deviation
    {sigma}, at displacement {delta} from the mean.*/

double log_prob_gauss_out (double A, double B, double x, double y, double epsilon, double sigma);
  /*Logarithm of the probability of {y} given {A}, {B} and {x},
    assuming a Gaussian distribution of errors with deviation {sigma},
    and outliers uniformly distributed in [-1,+1] with probability
    {epsilon}. */


double Q_lmdif 
  ( double A, 
    double B, 
    r2_t p[], 
    int npoints,  
    double sigma,  
    double lambda, 
    double epsilon );


void plot_Q_lmdif 
  ( double minA, 
    double maxA, 
    int nA, 
    double minB, 
    double maxB, 
    int nB, 
    double sigma, double lambda, double epsilon,
    r2_t p[], 
    int npoints, 
    FILE *arq );

double Q_weighted_least_squares (double A, double B, double log_prob (double A, double B, double x, double y), r2_t p[], int npoints);
  /*Logarithm of the likelihood of {A} and {B}, given the data points
    {p[0...npoints-1]} assuming the measurement pdf {log_prob}. */

void initialize_points (r2_t *p, int npoints, double A, double B, double sigma, double lambda, double epsilon); /*  Generates a set of "data" points {p[0..npoints-1]}.  The {x}
    coordinate is selected uniformly in {[-1 _ +1]}.  For inliers, the {y}
    coordinate is taken from a straight line {A x + B}, modified by
    Gaussian errors with deviation {sigma}. For outliers, the y coordinate is a 
    Gaussian variable with zero mean and deviation {lambda}. The
    oultiers have probability {epsilon}. */

void fit_line (r2_t p[], double w[], int npoints, double *A, double *B);
  /* Fits a line {y = A x + B} to the data points {p[0..npoints-1]}, by weighted least squares
    ("linear regression").  Each point {p[i]} has weight {w[i]}, the goal function is
    the sum of {w[i]*(y[i] - A*x[i] - B)^2} for all {i}, where {x[i] = p[i].c[0]}
    and {y[i] = p[i].c[1]}. */

void plot_error_function (void);
/* Function error to discards outliers. */

double gaussProb (double dev, double arg);
/* Gaussian pdf with deviation {dev} at argument {arg} */

int main (int argc, char *argv[]);

void plot_Q_weighted_least_squares (double minA, double maxA, int nA, double minB, double maxB, int nB, double log_prob (double A, double B, double x, double y), r2_t p[], int npoints, FILE *arq);

double recompute_weights
  ( r2_t p[], 
    int npoints, 
    double log_prob (double A, double B, double x, double y),
    double sigma, 
    double A, double B, double w[]
  );

double recompute_weights_lmdif
  ( r2_t p[], 
    int npoints, 
    double sigma,
    double lambda,
    double epsilon, 
    double A, double B, double w[]
    );


double log_prob_gauss_out (double A, double B, double x, double y, double epsilon, double sigma)
{
  double delta = y - (A * x + B);
  double z = delta/sigma;
  double C = sigma*sqrt(2*M_PI);
  if ((y >= -1.0) && (y <= +1.0) && (epsilon > 0.0)) {
    return log((1 - epsilon) * exp(-z*z/2.0)/C + epsilon/2.0);
  }
  else {
    return -z*z/2.0 - log(C);
  }
}

double log_gaussian (double delta, double sigma)
{
  double z = delta/sigma;
  double C = sigma*sqrt(2*M_PI);
  return -z*z/2.0 - log(C);
}


double Q_lmdif 
  ( double A, 
    double B, 
    r2_t p[], 
    int npoints,  
    double sigma,  
    double lambda, 
    double epsilon )
{
  int i;
  double sum = 0.0;

  double err[2*npoints];

  optimization_error (err, p, npoints, A, B, sigma, lambda, epsilon); 

  for (i = 0; i < 2*npoints; i++) {
    sum += err[i];
  }

  return sum;
}

void plot_Q_lmdif 
  ( double minA, 
    double maxA, 
    int nA, 
    double minB, 
    double maxB, 
    int nB, 
    double sigma, double lambda, double epsilon,
    r2_t p[], 
    int npoints, 
    FILE *arq
  )
{

  int iA, iB;
  
  for (iA = 0; iA <= +nA; iA++) {
    for (iB = 0; iB <= +nB; iB++) {
      double A = minA + iA*(maxA - minA)/nA;
      double B = minB + iB*(maxB - minB)/nB;
      double Q_AB = Q_lmdif (A, B, p, npoints, sigma, lambda, epsilon);
      fprintf(arq, "%+10.7f %+10.7f %+13.7f\n", A, B, Q_AB); 
    }
    fprintf(arq, "\n");
  }
}


double Q_weighted_least_squares (double A, double B, double log_prob (double A, double B, double x, double y), r2_t p[], int npoints)
{
  int i;
  double sum = 0.0;

  for (i = 0; i < npoints; i++)
    {
      sum += -log_prob(A, B, p[i].c[0], p[i].c[1]);
    }
  return sum;

}




void plot_Q_weighted_least_squares 
  ( double minA, 
    double maxA, 
    int nA, 
    double minB, 
    double maxB, 
    int nB, 
    double log_prob (double A, double B, double x, double y), 
    r2_t p[], 
    int npoints, 
    FILE *arq
  )
{

  int iA, iB;
  
  for (iA = 0; iA <= +nA; iA++) {
    for (iB = 0; iB <= +nB; iB++) {
      double A = minA + iA*(maxA - minA)/nA;
      double B = minB + iB*(maxB - minB)/nB;
      double Q_AB = Q_weighted_least_squares (A, B, log_prob, p, npoints);
      fprintf(arq, "%+10.7f %+10.7f %+13.7f\n", A, B, Q_AB); 
    }
    fprintf(arq, "\n");
  }
}

void initialize_points (r2_t *p, int npoints, double A, double B, double sigma, double lambda, double epsilon)
{
    int i;
    FILE *pts;
    pts = fopen("points.txt", "w");
    for (i = 0; i < npoints; i++)
    {
      double x = 2*drandom() - 1;
      double y;
      if (drandom() >= epsilon) {
	/*gaussian error*/ 
        y = A*x + B + sigma*dgaussrand();   
      }
      else {
        y = 0 + lambda*dgaussrand();
      }
      p[i].c[0] = x;
      p[i].c[1] = y;
      fprintf(pts, "%f %f\n", p[i].c[0], p[i].c[1]);
    }
    fclose(pts);

}


void fit_line (r2_t p[], double w[], int npoints, double *A, double *B)
{
  int i;
  double sum_w = 0, sum_w_x = 0.0, sum_w_y = 0.0, sum_w_x2 = 0.0, sum_w_x_y = 0.0;
  for (i = 0; i < npoints; i++)
  {
    double x = p[i].c[0];
    double y = p[i].c[1];  
    sum_w += w[i];
    sum_w_x += x*w[i];
    sum_w_y += y*w[i];
    sum_w_x2 += x*x*w[i];
    sum_w_x_y += x*y*w[i];
  }
  double det = (sum_w_x2 * sum_w) - (sum_w_x * sum_w_x);
  *A = ((sum_w_x_y * sum_w) - (sum_w_y * sum_w_x)) / det; 
  *B = ((sum_w_x2 * sum_w_y) - (sum_w_x_y * sum_w_x)) / det;
}


double recompute_weights_lmdif
  ( r2_t p[], 
    int npoints, 
    double sigma,
    double lambda,
    double epsilon, 
    double A, double B, double w[]
  )
{
   int i;
   double w_new[npoints];
   double w_max = 0.0;
   
   for (i = 0; i < npoints; i++)
   {
     double x = p[i].c[0];
     double y = p[i].c[1]; 

     double y_calc = A * x + B;
     double ei = err_in (sigma, lambda, epsilon, y_calc, y);
     double eo = err_out (lambda, y);


     double delta = (y - (A * x + B))*(y - (A * x + B));
     w_new[i] = (ei*ei + eo*eo)/delta;  
     if (w_new[i] < 0)
       { fprintf(stderr, "w_new[%3d] = %14.8f", i, w_new[i]);
         //double prG = exp(log_gaussian(delta, sigma));
         //double prO = exp(log_prob(A, B, x, y)); 
         //fprintf(stderr, "PrG = %14.8f  PrO = %14.8f\n", prG, prO);
       }
     if (fabs(w_new[i]) > w_max){ w_max = fabs(w_new[i]); }
   }

   assert(w_max > 0);
   double sum_dw2 = 0.0;
   
   for (i = 0; i < npoints; i++)
   {
     w_new[i] /= w_max; 
     double d_w = (w[i] - w_new[i]);  
     sum_dw2 += d_w*d_w; 
     w[i] = w_new[i];
   }
   return sqrt(sum_dw2); 
}


/*    for (i = 0; i < npoints; i++) {
      double y_calc = A * p[i].c[0] + B;
      double ei = err_in (sigma, lambda, epsilon, y_calc, p[i].c[1]);
      double eo = err_out (lambda, p[i].c[1]); 
      err[2*i]   = ei*ei;
      err[2*i+1] = eo*eo;
    }
*/


double recompute_weights
  ( r2_t p[], 
    int npoints, 
    double log_prob (double A, double B, double x, double y),
    double sigma, 
    double A, double B, double w[]
  )
{
   int i;
   double w_new[npoints];
   double w_max = 0.0;
   
   for (i = 0; i < npoints; i++)
   {
     double x = p[i].c[0];
     double y = p[i].c[1]; 
     double delta = (y - (A * x + B));
     w_new[i] = log_prob(A, B, x, y)/log_gaussian(delta, sigma);  
     if (w_new[i] < 0)
       { fprintf(stderr, "w_new[%3d] = %14.8f", i, w_new[i]);
         double prG = exp(log_gaussian(delta, sigma));
         double prO = exp(log_prob(A, B, x, y)); 
         fprintf(stderr, "PrG = %14.8f  PrO = %14.8f\n", prG, prO);
       }
     if (fabs(w_new[i]) > w_max){ w_max = fabs(w_new[i]); }
   }

   assert(w_max > 0);
   double sum_dw2 = 0.0;
   
   for (i = 0; i < npoints; i++)
   {
     w_new[i] /= w_max; 
     double d_w = (w[i] - w_new[i]);  
     sum_dw2 += d_w*d_w; 
     w[i] = w_new[i];
   }
   return sqrt(sum_dw2); 
}

double gaussProb (double sigma, double arg) //Gaussian pdf with deviation {dev} at argument {arg}
{
    return (exp(-((arg/sigma)*(arg/sigma))/2.0) / (sigma*sqrt(2*M_PI)));
}

double err_in (double sigma, double lambda, double epsilon, double yc, double y)
{
  double W = sqrt(2);
  double B = log(epsilon * gaussProb(lambda,yc) + (1 - epsilon)*gaussProb(sigma,0.0));
  double C = B + 0.5*(yc/lambda)*(yc/lambda);
  double D = sqrt(2 * (C - log(epsilon/(lambda * sqrt(2 * M_PI)))));
  double e = D * sqrt(1 - W * sigma * sqrt(2 * M_PI) * gaussProb(W * sigma, y - yc));
  if (y < yc)
    e = -e;
  return e; 
}

double err_out (double lambda, double y)
{
  return y / lambda;
}


double ProbCZ (double sigma, double lambda, double epsilon, double yc, double y)
{
  double B = log(epsilon * gaussProb(lambda,yc) +  (1 - epsilon) * gaussProb(sigma,0.0));
  double C = B + 0.5*(yc/lambda)*(yc/lambda);
  double eo2 = err_out (lambda, y)*err_out (lambda, y);
  double ei2 = err_in (sigma, lambda, epsilon, yc, y)*err_in (sigma, lambda, epsilon, yc, y); 
  return exp(C - 0.5*eo2 - 0.5*ei2);
}

double ProbGG (double sigma, double lambda, double epsilon, double yc, double y)
{
  return ((1 - epsilon)*gaussProb(sigma,y-yc) + epsilon*gaussProb(lambda,y)); 
}


void plot_error_function (void)
{
  /* Parameters of the "GG" error model: */
  double y_calc = 0.5;  /* "Correct" data value. */
  double sigma = 0.05;   /* Deviation for inliers (mean {y_calc}). */
  double lambda = 1.0;  /* Deviation for outliers (mean 0). */
  double epsilon = 0.5;     /* Probability of outliers. */


  /* Plotting parameters: */
  int npoints = 100; /* Actually, number of intervals. */
  double y_min =  -3.0;
  double y_max =   3.0;
  double step = (y_max - y_min)/(npoints);
 

  char *f_name = NULL;
  asprintf(&f_name, "error_plot.txt");
  FILE *f = fopen(f_name, "w");
  
  int i;
  for (i = 0; i <= npoints; i++) {
    double y = y_min + (i*step);
    
    /* The "GG" error model: */
    double probCZ = ProbCZ(sigma, lambda, epsilon, y_calc, y); 
    
    /* The "Cz" error model: */
    double err_inlier = err_in (sigma, lambda, epsilon, y_calc, y);
    double err_outlier = err_out (lambda, y);
     
    double probGG = ProbGG(sigma, lambda, epsilon, y_calc, y);
    fprintf(f, "%10.6f %15.8f %15.8f %15.12f %15.12f\n", y, err_inlier, err_outlier, probCZ, probGG);
    
  }

  free(f_name);
  fclose(f);

}

void fit_line_with_lmdif (r2_t *p, 
                          int npoints, 
                          double A_guessed,
                          double B_guessed,
                          double *A_calc,
                          double *B_calc,
                          double sigma, 
                          double lambda, 
                          double epsilon)
{
  /* parameters needed by MINPACK's lmdif() */
  /* Parameters controlling MINPACK's lmdif() optimization routine. */
  /* See the file lmdif.f for definitions of each parameter.        */  
  #define REL_SENSOR_TOLERANCE_ftol    1.0E-5      /* [pix] */  
  #define REL_PARAM_TOLERANCE_xtol     1.0E-7
  #define ORTHO_TOLERANCE_gtol         0.0
  #define LMDIF_EPSFCN                 1.0E-12     /* Do not set to zero! */
  #define LMDIF_MODE                   1           /* variables are scalled internally */
  #define LMDIF_FACTOR                 100.0 
  #define NPARAMS 2  //Miminize the (A,B) of (Ax + B)

  integer     m = 2*npoints; 
  integer     n = NPARAMS;   


  doublereal  ftol = REL_SENSOR_TOLERANCE_ftol;
  doublereal  xtol = REL_PARAM_TOLERANCE_xtol;
  doublereal  gtol = ORTHO_TOLERANCE_gtol;
  integer     maxfev = 1000*n;
  doublereal  epsfcn = LMDIF_EPSFCN;

  integer     mode = LMDIF_MODE;
  doublereal  factor = LMDIF_FACTOR;
  integer     nprint = 0;
  integer     info;
  integer     nfev;

  integer     ldfjac = m;
  doublereal  x[NPARAMS];
  doublereal  diag[NPARAMS];
  integer     ipvt[NPARAMS];
  doublereal  qtf[NPARAMS];
  doublereal  wa1[NPARAMS];
  doublereal  wa2[NPARAMS];
  doublereal  wa3[NPARAMS];

  doublereal *fvec;
  doublereal *fjac;
  doublereal *wa4;

  fvec = (doublereal *)notnull(malloc((unsigned int)m*(unsigned int)sizeof(doublereal)), "no mem for fvec");
  fjac = (doublereal *)notnull(malloc((unsigned int)(m*n)*(unsigned int)sizeof(doublereal)),"no mem for fjac");
  wa4 =  (doublereal *)notnull(malloc((unsigned int)m*(unsigned int)sizeof(doublereal)), "no mem for wa4");

  if (mode == 2) {
    int i;
    for (i = 0; i < NPARAMS; i++)
      diag[i] = 1.0; // some user-defined values 
  }

  auto void error (integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err, integer *iflag);
  // goal function for MINPACK's lmdif. 

  void error (integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err, integer *iflag) {
    assert(*m_ptr == m); assert(*n_ptr == n);
    double A = params[0];  
    double B = params[1]; 
    optimization_error (err, p, npoints, A, B, sigma, lambda, epsilon);
  }

  x[0] = A_guessed;
  x[1] = B_guessed;

  // perform the optimization 
  lmdif (error,  &m, &n, x, fvec, &ftol, &xtol, &gtol, &maxfev, &epsfcn,
         diag, &mode, &factor, &nprint, &info, &nfev, fjac, &ldfjac, ipvt, qtf, wa1, wa2, wa3, wa4);

  *A_calc = x[0];
  *B_calc = x[1];
   
  /* release allocated workspace */
  free (fvec);
  free (fjac);
  free (wa4);
 

}

void optimization_error
  ( doublereal *err,
    r2_t *p,
    int npoints,
    double A,
    double B,
    double sigma,
    double lambda,
    double epsilon)
{   
    int i;

    for (i = 0; i < npoints; i++) {
      double y_calc = A * p[i].c[0] + B;
      double ei = err_in (sigma, lambda, epsilon, y_calc, p[i].c[1]);
      double eo = err_out (lambda, p[i].c[1]); 
      err[2*i]   = ei*ei;
      err[2*i+1] = eo*eo;
    }
}

void fit_line_with_weighted_least_square_opt 
  ( r2_t *p, 
    int npoints, 
    double sigma,  
    double lambda,
    double epsilon,
    double log_prob (double A, double B, double x, double y),
    double *y_orig,  
    double *y_calc )
{



  /* Set the initial data point weights: */
  int i;
  double w[npoints];
  for (i = 0; i < npoints; i++) { w[i] = 1.0; } 
 
  /* The iterative weight adjustment loop: */    
  double A, B;         /* Current values of the adjustable model parameters to be fitted. */
  double d_w;          /* Change in weight vector between iterations. */
  double tol = 0.001;  /* Max weight change for stopping criterion. */
  int max_iters = 100; /* Give up after this many iterations. */
  int iters = 0;
  do {
    /* Recompute {A,B} by weighted least squares: */
    fit_line (p, w, npoints, &A, &B);
    fprintf(stdout, "A = %f, B = %f\n", A, B);

    /* Recompute the weights based on data and current model: */
    d_w = recompute_weights (p, npoints, log_prob, sigma, A, B, w);

    d_w = recompute_weights_lmdif (p, npoints, sigma, lambda, epsilon, A, B, w);
    
    /* Write the discrepancies, weights and other info: */
    char *f_A_B_name = NULL;
    asprintf(&f_A_B_name, "min%05d.txt", iters);
    FILE *min = fopen(f_A_B_name, "w"); 

    for (i = 0; i < npoints; i++) {
      double x = p[i].c[0];
      double y = p[i].c[1]; 
      double y_model = (A * x + B);
      double delta = (y - y_model);
      double lg_poi = log_prob (A, B, x, y);
      fprintf(min, "%+10.7f %+10.7f %+10.7f %+10.7f %9.7f %+10.7f\n", x, y, y_model, delta, w[i], lg_poi);
      y_orig[i] = y;
      y_calc[i] = y_model;
    }     



    free(f_A_B_name);
    fclose(min);
    iters++;
  } while ((iters < max_iters) && (d_w > tol));
}



int main (int argc, char *argv[])
{
  
  auto double lp_go (double A, double B, double x, double y);

  /* Get the "true" model parameters, for data generation: */
  int npoints = atoi(argv[1]);
  double sigma_true = atof(argv[2]);
  double lambda_true = atof(argv[3]);
  double epsilon_true = atof(argv[4]);
  double A_true = atof(argv[5]);
  double B_true = atof(argv[6]);

  double A_guessed = atof(argv[7]);
  double B_guessed = atof(argv[8]);

  /* Get the "assumed" fixed parameters of model to be fitted: */
  double sigma_assumed = atof(argv[9]);
  double lambda_assumed = atof(argv[10]);
  double epsilon_assumed = atof(argv[11]);
  char *opt = argv[12];

  /* Generate the data points: */
  r2_t *p = (r2_t *)malloc(npoints*sizeof(r2_t));

  double *y_orig = (double *)malloc(npoints*sizeof(double));
  double *y_calc = (double *)malloc(npoints*sizeof(double));

  initialize_points (p, npoints, A_true, B_true, sigma_true, lambda_true, epsilon_true);

  double A_calc, B_calc;

  if (strcmp(opt,"wls") == 0) {
    fit_line_with_weighted_least_square_opt (p, npoints, sigma_assumed, lambda_assumed, epsilon_assumed, lp_go, y_orig, y_calc);
  }
  else if (strcmp(opt,"lmdif") == 0) {
    fit_line_with_lmdif (p, npoints, A_guessed, B_guessed, &A_calc, &B_calc, sigma_assumed, lambda_assumed, epsilon_assumed);
  }
 
  fprintf(stderr, "A_calc = %f, B_calc = %f, A_true = %f, B_true = %f\n", A_calc, B_calc, A_true, B_true);

  FILE *Qerr = fopen("Qerror.txt", "w");

  if (strcmp(opt,"wls") == 0) {
     plot_Q_weighted_least_squares (-1, +1, 30, -1, +1, 30, lp_go, p, npoints, Qerr);
  }
  else if (strcmp(opt,"lmdif") == 0) {
     plot_Q_lmdif (-1, +1, 30, -1, +1, 30, sigma_assumed, lambda_assumed, epsilon_assumed, p, npoints, Qerr); 
  }

  fclose (Qerr);
 
  plot_error_function();

  return 0;

  double lp_go (double A, double B, double x, double y) {
    return log_prob_gauss_out (A, B, x, y, epsilon_assumed, sigma_assumed);
  }

}


