#ifndef __image_H__
#define __image_H__

#define _GNU_SOURCE
#include <stdio.h>
#include <tf_camera.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <assert.h>
#include <math.h>
#include <stdarg.h>
#include <r3x3.h>
#include "interpolate.h"

//#define MAXVAL 65535
#define MAXVAL 255
/* Maxval for output images. */

#define PGM_RAW_MAGIC 5
/* Magic number of PGM raw files. */

#define LENGTH 80

#define GAMMA (2.0)


/* --------------    image structures   --------------*/

typedef struct _image_t {
  int nx;    /* Image width */
  int ny;    /* Image height */
  double **val; /* Image pixels matrix pointer */
} *image_t;
/* Struct that defines an image */

void image_eval_bilinear
  ( double x,
    double y,
    image_t img,
    double *v
  );

void image_eval_3_A
  ( double x,
    double y,
    image_t img,
    double *v
  );

void image_eval_3_B
  ( double x,
    double y,
    image_t img,
    double *v
  );

void image_eval_3_C
  ( double x,
    double y,
    image_t img,
    double *v
  );

/* --------------   I/O image functions   --------------*/

image_t image_read_pgm (char *fname);
/* Return an image structure with the data contained in the image disk file {fname}.
  Pixel values are rescaled from {[0 .. maxval]} to {[0 _ 1]}. */

void image_write_pgm (image_t img, int magic, int maxval, char *fname);
/* Writes a PGM (grayscale) image file {fname} with the data of {img}.
  Pixel values are rescaled from {[0 _ 1]} to {[0 .. maxval]},
  and clipped to {[0 .. maxval]}. */

void image_ppm_write_targets_with_crosses (image_t img, r2_t *p_i, int ntargets, int magic, int maxval, int magnify, char *fname);
/* Writes a PPM (color) image file called {fname} with the data of {img}, with red crosses
  at the points {p_i[0..ntargets-1]}.
  Pixel values are rescaled from {[0 _ 1]} to {[0 .. maxval]},
  and clipped to {[0 .. maxval]}. */

void image_read_pgm_header (FILE *fp, int *magic, int *ncols, int *nrows, int *maxval);
/* Read the header (which consists of: magic number (P1, P2, ... [see PGM format]), 
   rows, cols and the maximum grey level) of pgm files. */

void image_get_next_string (FILE *fp, char *line);
/* Function to ignore string commentaries that some 
   programs as GIMP puts in pgm files after save them.*/

/* --------   Basic image and auxiliary functions  --------*/

void image_print_cross (image_t img, int size, r2_t pi);

void image_print_dot (image_t img, int size, r2_t pi);

image_t image_alloc (int nx, int ny);
/* Returns a P{magic} image with {nx} columns and {ny} rows. */

void image_free (image_t img);
/* Disallocates the {img} structure */

image_t image_copy (image_t img);
/* Returns the copy a {img} image */

void image_message_error (char *fmt, ...);
/* Show the {fmt} message when occurs an error and 
  interrupts (exit(1)) the execution of the program */

image_t image_diff_x_direction (image_t img);
/* Returns the image gradient in x direction of {img}. */
/* gx(x,y) = ( f(x, y+1) - f(x, y-1) )/ 2 */

image_t image_diff_y_direction (image_t img);
/* Returns the image diff in y direction of {img}. */
/* gy(x,y) = ( f(x+1, y) - f(x-1, y) )/ 2 */

image_t image_diff_xx_direction (image_t img);
/* Returns the second derivative of {img} in x direction. */
/* gxx(x,y) = f(x, y+1) - 2*f(x, y) + f(x, y-1) */

image_t image_diff_xy_direction (image_t img);
/* Returns the first derivative of {img} in x direction and the snd in y. */
/* gxy(x,y) = f(x+1, y+1) - f(x-1, y+1) - f(x+1, y-1) + f(x-1, y-1)  */

image_t image_diff_yy_direction (image_t img);
/* Returns the second derivative of {img} in y direction. */
/* gyy(x,y) = f(x+1, y) - 2*f(x, y) + f(x-1, y)  */


/* --------------  Others image functions --------------*/

void image_pixel_center (int ix, int iy, double *x, double *y);

void image_map_coords (double ax, double ay, r3x3_t *map, double *bx, double *by);
/* Applies the {map} to the point {(ax,ay)}, stores the result in {(*bx,*by)}*/

void image_map_mark_coords_to_image_coords (double ax, double ay, int Mnx, int Mny, int subsampling, r2_t p_i, double *bx, double *by);
/* Assumes that {(ax,ay)} are the coordinates of a point {a} in the domain of a mark image {M} with
  size {Mnx,Mny}.  Returns in {(*bx,*by)} the coordinates of the corresponding point {b}
  in the plane when {M} is reduced by a factor {1/subsampling} and translated 
  so that its center ends up at {p_i}. */

void image_map_image_coords_to_mark_coords (double bx, double by, int Mnx, int Mny, int subsampling, r2_t p_i, double *ax, double *ay);
/* Assumes that some mark image {M} with
  size {Mnx,Mny} is reduced by a factor {1/subsampling} and translated 
  so that its center ends up at {p_i}. Given the coordinates {(bx,by)} of a point {b} in the plane,
  returns in {(*ax,*ay)} the coordinates of the corresponding point {a}
  in the domain of {M}. */

void image_map (image_t M, image_t U, r3x3_t *map, image_t N, image_t V);
/* Paints the image {M} into image {N}, with its domain deformed by {map}.

  If {U} is not NULL, it must be a weight mask for {M}.  The procedure
  sets a pixel of {N} to 0.5 if all its source pixels have zero
  {U}-weight.

  If {V} is not NULL, it must be an image with the same size as {N}.
  The procedure sets each pixel of {V} to the total weight 
  (sampling kernel times mask) of the pixels that contributed to 
  the corresponding pixel of {N}. In particular, pixels of {N} that 
  have no source pixel with positive weight will get their {V} pixel
  set to zero. */  

void image_compute_gaussian_weights(double sigma, double **wp, int *hwp);
/* Allocates and fills a table {(*wp)[0..nw-1]} of weights taken from a Gaussian
  curve with deviation {sigma}.  The table size {nw} is {2*(*hwp)+1} where
  {*hwp} is chosen by the procedure.  The central (maximum) weight is therefore 
  {(*wp)[*hwp]}. The weights are normalized so that their sum is 1. */


void image_compute_map_weights(double dx, double dy, double **wp, int *hwp, double *stepp);
/* Computes the number of subsamples, the sampling step, and the
  sample weights to compute the line integral of an image.  Assumes
  that a unit step in the integration variable results in a
  displacement {{(dx,dy)} in the image.  The subsampling arguments to
  be used in the summation are {t*(*stepp)} for {t} in {-(*hwp)
  .. +(*hwp)}.  The weight of each subsample is {(*wp)[t +
  (*hwp)]}. */

void image_paint_mapped_mark (image_t gab, image_t wmk, image_t img, double alpha, double beta, r3x3_t *map);
/* Paints the image {gab}, masked by {wmk}, into the image {img}, with
  its domain deformed by {map}.  Pixels of {img} that are not covered
  by the deformed domain are left unchanged.  Each color value {v} of
  {gab} is mapped to {alpha*v + beta}. */  

void image_paint_translated_mark (image_t gab, image_t wmk, image_t img, int subsampling, double alpha, double beta, r2_t p);
/* Paints into the image {img} the image {gab}, masked by {wmk}, scaled by {1/subsampling} and translated so that
  the center of {gab}'s domain is at point {p} of {img}'s domain.
  Each color value {v} of {gab} is mapped to {alpha*v + beta}. */  

void image_compute_blur_weights(double blur, int *hwp, double **wp);
/* Computes the number of subsamples and the sample weights to apply a
   gaussian blur with standard deviation {blur} to a signal.  The
   input samples to be used for each blurred sample {i} are {i+t} for
   {t} in {-(*hwp) .. +(*hwp)}.  The weight of each subsample is
   {(*wp)[t + (*hwp)]}. */


image_t image_blur (image_t img, image_t msk, double blur);
/* Applies a gaussian blur in {img} with radius {blur}. If {msk} if not NULL
   it must have the same size as {img} and each pixel of {img} is weighted by the 
   corresponding pixel of {msk}, besides the blur kernel weights. */

void image_change_interval (image_t img, double vmin, double vmax, double min, double max);
/* Maps every pixel in {img} from the range [vmin..vmax] to the range [min..max]. 
  If vmin = vmax, uses the actual range of pixel values from the {img}*/

void image_change_interval_and_write
  ( image_t img, 
    int magic, 
    int maxval, 
    char *iname, 
    char *type, 
    double vmin,
    double vmax, 
    double min, 
    double max
  );
/* Writes a copy of {img} as a PNM file of type {magic} with specified {maxval}.
  If {min!=max}, performs {image_change_interval(cpy,vmin,vmax,min,max)} on the copy. 
  The image {img} is not modified. */

void image_soft_resize (image_t M, image_t U, int redFactor, image_t *N, image_t *V);
/*
  Returns a copy of the image {img} reduced by {1/redFactor}.
  If {useAvg} is true, the original pixels are averaged, else
  they are added.  If {wholePixels} is true, a reduced pixel 
  is 0 if any original pixel falls outside the image; else
  the pixels outside the image are taken to be 0. */

r3x3_t image_resize_map(int nx, int ny, int redFactor);
/* The affine map that takes a point in the domain of a an image {I} with size {nx,ny} to 
  a copy of {I} reduced by {redFactor}.  (This is the map used by {image_soft_resize}
  above.) */

r2_t image_map_point_to_smaller_image (r2_t p_i, int dscale, int redFactor, int nx, int ny);
/* 
  Computes the position {p_r} in a reduced-scale image {R} that corresponds to 
  point {p_i} in the original image {I}, assuming that {R} was obtained from {I}
  by {dscale} calls to {image_soft_resize} with scale factor {redFactor}.
  Assumes that the larger image {I} has size {nx} by {ny}. */

r2_t image_map_point_to_larger_image (r2_t p_r, int dscale, int redFactor, int nx, int ny);
/* 
  Computes the position {p_i} in the original image {I} that corresponds to 
  point {p_r} in a reduced-scale image {R}, assuming that {R} was obtained from {I}
  by {dscale} calls to {image_soft_resize} with scale factor {redFactor}.
  Assumes that the larger image {I} has size {nx} by {ny}. */

image_t image_gaussian_mask (int nx, int ny, int normalize);
/* A mask image with size {nx,ny}, whose pixels are a two-dimensional gaussian function.
  The gaussian is centered in the image and has radius  */

void image_extract_subimage(image_t img, image_t gab, int minx, int miny);
/* Fills the image {gab} with the sub-image of {img} whose upper left pixel
  is at column {minx} and row {miny}. Pixel values outside the image are 
  extapolated from the nearest existing pixel. */  

void image_extract_mapped_mark (image_t img, image_t gab, double alpha, double beta, r3x3_t *map);
/* Fills the image {gab} with the part of {img} that is contained in the domain of {gab} deformed by {map}.
  Each color value {v} of {gab} is mapped to {(v - beta)/alpha}. */  

void image_extract_translated_mark (image_t img, image_t gab, int subsampling, double alpha, double beta, r2_t p);
/* Fills the image {gab} with the part of {img} that is contained in the domain of {gab}, 
  scaled by {1/subsampling} and translated
  so that the center of {gab}'s domain is at point {p} of {img}'s domain.
  Each color value {v} of {gab} is mapped to {(v - beta)/alpha}. */  

void image_fill (image_t img, double value);


void image_weighted_avg_var 
  ( image_t I, 
    image_t W, 
    int subsampling,
    r2_t p_i, 
    double *avg, 
    double *var );
/* Computes the average and variance of {I} over a region {R}, weighted by {W}.
  The region {R} is the domain of {W}, reduced by {1/subsampling},
  and translated so that its center is located at {p_i}. */

void image_weighted_regression
  ( image_t W, 
    image_t M, 
    int M_subsampling, 
    r2_t M_position, 
    image_t I, 
    int I_subsampling, 
    r2_t I_position,
    bool_t debug, 
    double *alpha, 
    double *beta, 
    double *corr );
/* Computes the coefficients {alpha} and {beta} such that 
  {alpha*M + beta} best approximates the image {I} in the least squares sense.
  The image {M} is implicitly translated so that point {M_position} goes to
  the origin, and then scaled by {M_subsampling}.  Likewise, the {I} image 
  is implicitly translated so that point {I_position} goes to
  the origin, and then scaled by {I_subsampling}.
  The images are then weighted by the mask {W}. */

/*---------------- Auxiliary functions ----------------------*/

double ** image_mdouble_alloc (int nrows, int ncols);
/* Returns a double matrix with {nrows}x{ncols} dimension */

void image_mdouble_free (double **m, int nrows);
/* Disallocates a double matrix {m} with {nrows} */

void image_deform(image_t M, image_t U, r2_t *u_i, r2_t *v_i, int subsampling, image_t *N, image_t *V);
/* Returns a deformed copy {D} of image {I}.  Let {Inx,Iny} be the
  size of {I}.  The center of {I} is mapped to the center of {D}. The
  displacement vector {(+Inx/2,0)} is mapped to the displacement
  vector {*u_i} times {subsampling}. The displacement vector {(0,-Iny/2)} is mapped to the
  displacement vector {*v_i} times {subsampling}. The image {D} is large enough to contain
  the distorted copy of {M}.

  If {useAvg} is true, the original pixels are averaged, else
  they are added.  If {wholePixels} is true, a reduced pixel 
  is 0 if any original pixel falls outside the image; else
  the pixels outside the image are taken to be 0.

*/


r3x3_t image_map_from_shape_vectors (int Mnx,
                                  int Mny,
                                  int Gnx,
                                  int Gny,
                                  r2_t *u_i,
                                  r2_t *v_i,
                                  int subsampling);

/*Computes an affine 3x3 matrix that maps an original image {M} to a
 deformed image {G}. Requires the dimensions {Mnx,Mny} of {M} and the
dimensions {Gnx,Gny} of {G}.  The center {(Mnx/2,Mny/2)} of {M} is mapped to
center {(Gnx/2,Gny/2)} of {G}. The horizontal displacement vector {(+Mnx/2,0)}
in {M} is mapped to the displacement vector {subsampling*(*u_i)} in G.  The vertical
displacement vector {(0,-Mny/2)} in {M} is mapped to the displacement vector
{subsampling*(*v_i)} in G. */


void image_subtract_mark (image_t I, image_t M);
/* Subtracts the image {M}, centered, from image {I}. The sizes must differ by even integer. */

void image_multiply_by_normalized_weight (image_t I, image_t W);
/* Multiplies the image {W}, centered, into image {I}.
   The sizes must differ by even integer. The image {W} is implicitly 
   normalized {[0_wtmax] -> [0_1]} and extended with zeros if needed. */

r2_t image_extrapolate_position(r2_t p_2, r2_t p_1);
/* Estimates the position of an object in the current frame from its positions 
  in the next-to-last frame ({p_2}) and in the last frame ({p_1}). */
 
r2_t image_snd_order_extrapolate_position (r2_t p_3, r2_t p_2, r2_t p_1);
/* Estimates the position of an object in the current frame from its three past positions 
  ({p_1, p_2, p_3}). Thus performing a second order extrapolation. */


void image_apply_color_correction (image_t img, double alpha, double beta);

image_t image_extract_subimage_by_pos (image_t img, r2_t pi, int radius_x, int radius_y, int magnify);

image_t image_magnify (image_t img, int magnify);

void image_resize_size (int nx, int ny, int redFactor, int *rnx, int *rny);

#endif
