#include "interpolate.h"

void interpol_linear 
( double t,
  double v[],
  double *s,
  double *ds,
  double *d2s 
 ){
  if (t == 0) {
    if (s != NULL) *s = v[0];
    if (ds != NULL) *ds = 0.0; /* Should be NaN */
  } else if (t == 1) {
    if (s != NULL) *s = v[1];
    if (ds != NULL) *ds = 0.0; /* Should be NaN */
  } else {
    if (s != NULL) *s = (1 - t)*v[0] + t*v[1];
    if (ds != NULL) *ds = v[1] - v[0];
  }
  if (d2s != NULL) *d2s = 0.0; /* Should be NaN */
}

void interpol_bilinear 
( double xfrac, 
  double yfrac, 
  double v00, 
  double w00, 
  double v10, 
  double w10, 
  double v01, 
  double w01, 
  double v11, 
  double w11, 
  double *s
  ){
  /* Should consider special cases {xfrac,yfrac == 0,1} */
  w00 *= (1 - xfrac) * (1 - yfrac);
  w10 *= xfrac * (1 - yfrac);
  w01 *= (1 - xfrac) * yfrac;
  w11 *= xfrac * yfrac;
  *s =  (w00*v00 + w10*v10 + w01*v01 + w11*v11)/(w00 + w10 + w01 + w11 + 1.0e-200);
}

void interpol_spline_2_B 
( double t,
  double v[],
  double *s,
  double *ds,
  double *d2s 
 ){
  if (t == 0) {
    if (s != NULL) *s = 0.5*v[0] + 0.5*v[1];
    if (ds != NULL) *ds = v[1] - v[0];
    if (d2s != NULL) *d2s = 0.0; /* Should be NaN */
  } else if (t == 1) {
    if (s != NULL) *s = 0.5*v[1] + 0.5*v[2];
    if (ds != NULL) *ds = v[2] - v[1];
    if (d2s != NULL) *d2s = 0.0; /* Should be NaN */
  } else {
    double B0 = 0.5*(1-t)*(1-t); double dB0 = t - 1;   double d2B0 = +1.0;
    double B1 = 0.5 + t*(1-t);   double dB1 = 1 - 2*t; double d2B1 = -2.0;
    double B2 = 0.5*t*t;         double dB2 = t;       double d2B2 = +1.0;
    if (s != NULL) *s = B0*v[0] + B1*v[1] + B2*v[2];
    if (ds != NULL) *ds = dB0*v[0] + dB1*v[1] + dB2*v[2];
    if (d2s != NULL) *d2s = d2B0*v[0] + d2B1*v[1] + d2B2*v[2];
  }
}

void interpol_spline_3_A 
( double t,
  double v[],
  double *s,
  double *ds,
  double *d2s 
 ){
  /* Should consider special cases {t == 0,1} */
  double F3,F2,F1,F0;
  F3 = -v[0] + (3.0 * v[1]) - (3.0 * v[2]) + (v[3]);
  F2 = (2.0*v[0]) - (5.0*v[1]) + (4.0*v[2]) - v[3];
  F1 = -v[0] + v[2];
  F0 = 2.0 * v[1];
  if (s != NULL) *s = (F0 + t * (F1 + t * (F2 + t * (F3))))/2.0;
  if (ds != NULL) *ds = ((F1 + t * (2*F2 + t * (3*F3))))/2.0;
  if (d2s != NULL) *d2s = F2 + t * (F3);
}

void interpol_spline_3_B 
( double t,
  double v[],
  double *s,
  double *ds,
  double *d2s 
 ){
  /* Should consider special cases {t == 0,1} */
  double F3,F2,F1,F0;
  F3 = -v[0]/6.0 + v[1]/2.0 - v[2]/2.0 + v[3]/6.0;
  F2 =  v[0]/2.0 - v[1] + v[2]/2.0;
  F1 = -v[0]/2.0 + v[2]/2.0;
  F0 =  v[0]/6.0 + 2.0/3.0 * v[1] + v[2]/6.0;
  if (s != NULL) *s   = F0 + t * (F1 + t * (F2 + t * (F3)));
  if (ds != NULL) *ds  = F1 + t * (2*F2 + t * (3*F3));
  if (d2s != NULL) *d2s = 2*F2 + t * (6*F3);
}

void interpol_spline_3_C 
( double t,
  double v[],
  double *s,
  double *ds,
  double *d2s 
 ){
  /* Should consider special cases {t == 0,1} */
  double a = -0.5;
  double t0 = t+1, t1 = t, t2 = 1-t, t3 = 2-t;

  double K0 = (-4*a) + t0*((8*a) + t0*((-5*a) + t0*(a))); 
  double K1 = 1.0 + t1*t1*((-a-3) + t1*(a+2)); 
  double K2 = 1.0 + t2*t2*((-a-3) + t2*(a+2)); 
  double K3 = (-4*a) + t3*((8*a) + t3*((-5*a) + t3*(a)));

  double dK0 = (8*a) + 2*t0*((-5*a) + 3*t0*(a)); 
  double dK1 = t1*(2*(-a-3) + 3*t1*(a+2)); 
  double dK2 = t2*(2*(-a-3) - 3*t2*(a+2)); 
  double dK3 = (8*a) - 2*t3*((-5*a) - 3*t3*(a)); 

  double d2K0 = 2*((-5*a) + 6*t0*(a)); 
  double d2K1 = 2*(-a-3) + 6*t1*(a+2); 
  double d2K2 = 2*(-a-3) + 6*t2*(a+2); 
  double d2K3 = 2*((-5*a) + 6*t3*(a)); 

  if (s != NULL) *s = K0*v[0] + K1*v[1] + K2*v[2] + K3*v[3];
  if (ds != NULL) *ds = dK0*v[0] + dK1*v[1] + dK2*v[2] + dK3*v[3];
  if (d2s != NULL) *d2s = d2K0*v[0] + d2K1*v[1] + d2K2*v[2] + d2K3*v[3];
}

