#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <image.h>

void plot_1D_spline_and_derivatives (char *name, int degree, char type, int ns, double s[]);

void interpol_1D (int degree, char type, double t, double *s, int ns, double *v, double *dv, double *d2v);
  /* Given samples {s[0..ns-1]}, interpolates them at position {t} with the
    spline of given {degree} and {type}. */

int main (int argc, char *argv[])
{
  struct mallinfo info = mallinfo();
  int MemDinInicial = info.uordblks;

  /* Degree of spline: */
  int degree = atoi(argv[1]);

  /* Type of spline: */
  char type = argv[2][0];

  /* A sample array: */
  int ns = degree + 1 + degree%2;
  double s[ns]; 
  int i;
  for (i = 0; i < ns; i++) { s[i] = (i == ns/2 ? 1.0 : 0.0); }

  plot_1D_spline_and_derivatives ("out/spline1D", degree, type, ns, s);

  info = mallinfo();
  int MemDinFinal = info.uordblks;
  if (MemDinInicial!=MemDinFinal)
    fprintf(stderr, "** main: memory leak - %d blocks\n", MemDinFinal-MemDinInicial);
  return 0;
} 

void plot_1D_spline_and_derivatives (char *name, int degree, char type, int ns, double s[])
{
  int r;
  double step = 0.01; /* Numeric differentiation step. */

  fprintf(stderr, "plotting 1D spline of degree %d and type %c\n", degree, type);

  char *fname = NULL;
  asprintf(&fname, "%s_%d_%c.txt", name, degree, type);
  fprintf(stderr, "writing %s ...", fname); 
  FILE *fplot = fopen(fname, "w");
  assert(fplot != NULL);

  int N = 20*ns; /* Number of plot intervals. */
  int h = (ns+1)/2;

  for (r = -h; r <= N+h; r++) {
    double t = ((double)r/N)*ns;

    double v,d_v,d2_v,dn_v,dn_d_v,d2n_v;
    double vp, vn, d_vp, d_vn;

    interpol_1D (degree, type, t, s, ns, &v, &d_v, &d2_v); 
    interpol_1D (degree, type, t+step, s, ns, &vp, &d_vp, NULL); 
    interpol_1D (degree, type, t-step, s, ns, &vn, &d_vn, NULL); 

    dn_v = (vp - vn)/2/step; 
    dn_d_v = (d_vp - d_vn)/2/step; 
    d2n_v = (vp - 2*v + vn)/step/step;

    fprintf(fplot, "%12.8f %12.8f", t, v);
    fprintf(fplot, "  %12.8f %12.8f", d_v, dn_v );
    fprintf(fplot, "  %12.8f %12.8f %12.8f", d2_v, d2n_v, dn_d_v);
    fprintf(fplot, "\n"); 

  } 
  fclose(fplot);
  fprintf(stderr, "done\n"); 
  free(fname);
}

void interpol_1D (int degree, char type, double t, double *s, int ns, double *v, double *dv, double *d2v) 
{
  switch(degree) {
  case 1:
    { /* Odd degree: breaks at half-integer {t} values: */
      t -= 0.5;
      /* Get relevant samples: */
      int it = floor(t);
      double tfrac = t - it;

      double p[2];
      p[0] = (((it+0) >= 0) && ((it+0) < ns) ? s[it+0]:0);
      p[1] = (((it+1) >= 0) && ((it+1) < ns) ? s[it+1]:0);

      assert(type == 'B');
      interpol_linear (tfrac, p, v, dv, d2v);
      break;
    }
  case 2:
    { /* Even degree: breaks at integer {t} values: */
      t -= 0.0;
      /* Get relevant samples: */
      int it = floor(t);
      double tfrac = t - it;

      double p[3];
      p[0] = (((it-1) >= 0) && ((it-1) < ns) ? s[it-1]:0);
      p[1] = (((it+0) >= 0) && ((it+0) < ns) ? s[it+0]:0);
      p[2] = (((it+1) >= 0) && ((it+1) < ns) ? s[it+1]:0);

      assert(type == 'B');
      interpol_spline_2_B (tfrac, p, v, dv, d2v);
      break;
    }
  case 3:
    { /* Odd degree: breaks at half-integer {t} values: */
      t -= 0.5;
      /* Get relevant samples: */
      int it = floor(t);
      double tfrac = t - it;
     
      double p[4];
      p[0] = (((it-1) >= 0) && ((it-1) < ns) ? s[it-1]:0);
      p[1] = (((it+0) >= 0) && ((it+0) < ns) ? s[it+0]:0);
      p[2] = (((it+1) >= 0) && ((it+1) < ns) ? s[it+1]:0);
      p[3] = (((it+2) >= 0) && ((it+2) < ns) ? s[it+2]:0);

      switch (type) {
      case 'A': interpol_spline_3_A (tfrac, p, v, dv, d2v); break;
      case 'B': interpol_spline_3_B (tfrac, p, v, dv, d2v); break;
      case 'C': interpol_spline_3_C (tfrac, p, v, dv, d2v); break;
      default: assert(FALSE); 
      }
      break;
    }
  default:
    { fprintf(stderr, "bad degree %d\n", degree);
      assert(0);
    }
  }
}


