#include "lk.h"
#include "numeric.h"

#define DEBUG_LOST_MARKS 1
#define DEBUG_ALL 0

void lk_mark_position_adjust
  ( char *debug_name,
    image_t I, 
    image_t M, 
    image_t W,
    int subsampling,  
    double alpha,
    double beta,
    r2_t *p_i,
    r2_t *p_i_dev,
    lk_error_function_t error_func, 
    numeric_minimizer_t minimize,
    double step_factor,
    int niter,
    double *errorSqr
  )
{

   /*The point falled outside the image*/
   if (p_i->c[0] < 0 || p_i->c[0] >= I->nx || p_i->c[1] < 0 || p_i->c[1] >= I->ny) {
      *errorSqr = 1.0;
      if (DEBUG_LOST_MARKS) {
        fprintf(stderr, "*** mark is outside the frame domain (%s)\n", debug_name);
      }
      return;
   }

   int N = 30; /*number of plot steps*/
   int nz = 2;   /*number of parameters*/
   double z[nz]; /*current paramter vector*/ 
   
   /* Max displacement allowed from initial guess [pixels]: */
   double max_d = 2.0;
   double max_disp[nz];
   int k;
   for (k = 0; k < nz; k++) { max_disp[k] = max_d; } 
   double min_step = 0.01/max_d; /* Convergence step length (rel to {max_disp}): */
  
   if (DEBUG_ALL) {
     fprintf(stderr, "enter lk_mark_position_adjust\n");
     fprintf(stderr, "alpha = %8.6f, beta = %8.6f, ", alpha, beta);
     fprintf(stderr, "p_i = (%8.2f, %8.2f)\n", p_i->c[0], p_i->c[1]); 
   }

   if (DEBUG_GIVEN_MARK_IMAGE) {       
      /* Write the feature (mark) image being searched: */
      image_t color_corrected_M;
      char *M_color_corrected_name = NULL;
      asprintf(&M_color_corrected_name, "%s_Q.pgm", debug_name);
      color_corrected_M = image_copy(M);
      image_apply_color_correction (color_corrected_M, alpha, beta);
      image_write_pgm (color_corrected_M, PGM_RAW_MAGIC, MAXVAL, M_color_corrected_name);
      image_free(color_corrected_M);
      free(M_color_corrected_name);

      char *M_name = NULL;
      asprintf(&M_name, "%s_M.pgm", debug_name);
      image_write_pgm (M, PGM_RAW_MAGIC, MAXVAL, M_name);
      free(M_name);
   }

   if (DEBUG_WEIGHT_IMAGE) {
      /* Write the weight mask: */
      char *W_name = NULL;
      asprintf(&W_name, "%s_W", debug_name);
      image_change_interval_and_write (W, PGM_RAW_MAGIC, MAXVAL, W_name, "pgm", 0, 0, 0, 1);
      free(W_name);
   }

   if (DEBUG_GUESS_CONTEXT_IMAGE) {   
     /* Write window {K} extracted from image {I} around guessed point {p_i} before adjustment */
     int dnx = (int)ceil(M->nx*(DEBUG_CONTEXT_IMAGE_REL_SIZE));
     int dny = (int)ceil(M->ny*(DEBUG_CONTEXT_IMAGE_REL_SIZE));
     int minx = (int)floor(p_i->c[0] - 0.5*dnx);
     int miny = (int)floor(p_i->c[1] - 0.5*dny);
     int maxx = minx + dnx - 1;
     int maxy = miny + dny - 1;

     lk_write_target_image_region (I, minx, maxx, miny, maxy, debug_name, "K");
   }

   auto void goal_func(int nz, double z[], double *Q, double *J, double *H);

   void goal_func(int nz, double z[], double *Q, double *J, double *H)
     { assert(nz == 2);
       r2_t q_i;
       lk_scatter_params (nz, z, &q_i);

       error_func(I, M, W, subsampling, alpha, beta, q_i, Q, J, H);
       lk_add_a_priori_error(&q_i, p_i, p_i_dev, nz, Q, J, H);
     }

   /* Pack {p_i} as a minimization parameter vector {z[0..nz-1]}: */
   lk_gather_params (nz, z, *p_i);
 
   if (TEST_ERROR) {
      double diff_step[nz];
      lk_choose_num_diff_steps(nz, diff_step); 
      numeric_test_goal_func_derivatives (goal_func, nz, z, diff_step);
   }

   if (PLOT_ERROR) {
     double plot_step[nz];
     lk_choose_plot_steps(nz, plot_step);
     double diff_step[nz];
      lk_choose_num_diff_steps(nz, diff_step); 
      numeric_plot_all_goal_funcs_and_derivatives (debug_name, goal_func, nz, z, diff_step, plot_step, N);
   }

   int ok = minimize(nz, z, goal_func, step_factor, max_disp, niter, min_step);

   if (DEBUG_ALL) {
     fprintf(stderr, "optimized mark position: z[0] = %f, z[1] = %f\n", z[0], z[1]);
   }
   
   if (ok) {

     /* Compute error function value {*error} at final point {z[0..nz-1]}: */
     goal_func(nz, z, errorSqr, NULL, NULL);

     /* Unpack {p_i} from the minimum parameter vector {z[0..nz-1]}: */
     lk_scatter_params (nz, z, p_i);
     
   } else {

     /* Set {*errorSqr} to a large value, and leave {p_i} unchanged: */
     *errorSqr = 1.0;
     if (DEBUG_LOST_MARKS) {
       fprintf(stderr, "** optimization of mark's position failed (%s)\n", debug_name);
     }
   }
   
   if (DEBUG_ADJUSTED_MARK_IMAGE) {   
     /* Extract and write the occurrence {L} of the mark image as found in {I}, color-matched to {M}: */
     lk_write_mark_extracted_from_image (I, M->nx, M->ny, subsampling, *p_i, alpha, beta, debug_name, "L");
     /* Extract and write the occurrence {N} of the mark image as found in {I}: */
     lk_write_mark_extracted_from_image (I, M->nx, M->ny, subsampling, *p_i, 1.0, 0.0, debug_name, "N");
     /* Extract and write the difference {D} between the mark image and its instance in {I}: */
     lk_write_mark_difference_with_image (I, M, W, subsampling, *p_i, alpha, beta, debug_name, "D");
   }

   if (DEBUG_ADJUSTED_CONTEXT_IMAGE) {  
     /* Write window {K} extracted from image {I} around guessed point {p_i} before adjustment */
     int dnx = (int)ceil(M->nx*(DEBUG_CONTEXT_IMAGE_REL_SIZE));
     int dny = (int)ceil(M->ny*(DEBUG_CONTEXT_IMAGE_REL_SIZE));
     int minx = (int)floor(p_i->c[0] - 0.5*dnx);
     int miny = (int)floor(p_i->c[1] - 0.5*dny);
     int maxx = minx + dnx - 1;
     int maxy = miny + dny - 1;
     lk_write_target_image_region (I, minx, maxx, miny, maxy, debug_name, "R");
   }
}

void lk_add_a_priori_error(r2_t *q_i, r2_t *p_i, r2_t *p_i_dev, int nz, double *Q, double *J, double *H)
{
   double e_x = 1.0/p_i_dev->c[0];
   double e_y = 1.0/p_i_dev->c[1];
   double d_x = e_x*(p_i->c[0] - q_i->c[0]);
   double d_y = e_y*(p_i->c[1] - q_i->c[1]);
 
   /*adding error in Q*/
   if (Q != NULL) {
      *Q += d_x*d_x + d_y*d_y;
   }   

   /*adding error in J*/
   if (J != NULL) {
      J[0] += 2*e_x*d_x;
      J[1] += 2*e_y*d_y;
   }	

   /*adding error in H*/
   if (H != NULL) {
     H[0] += 2.0*e_x*e_x;
     H[3] += 2.0*e_y*e_y;
   }	    

}


void lk_write_target_image_region
  ( image_t I, 
    int minx, 
    int maxx, 
    int miny, 
    int maxy, 
    char *debug_name, 
    char *tag
  )
{
  int nx = maxx - minx + 1;
  int ny = maxy - miny + 1;

  /* Extract the sub-image {R} from {I}: */
  image_t R = image_alloc (nx, ny); 
  image_extract_subimage(I, R, minx, miny);
    
  char *R_name = NULL;
  asprintf(&R_name, "%s_%s.pgm", debug_name, tag);
  image_write_pgm (R, PGM_RAW_MAGIC, MAXVAL, R_name);
  image_free (R);
  free(R_name);
}

void lk_write_mark_extracted_from_image
  ( image_t I, 
    int nx, 
    int ny, 
    int subsampling, 
    r2_t p_i, 
    double alpha, 
    double beta, 
    char *debug_name, 
    char *tag
  )
{
    /* Extract and write the occurrence {R} of the mark image as found in {I}: */
    image_t R = image_alloc (nx, ny);
    if (DEBUG_ALL) {
      fprintf(stderr, "%s: alpha = %16.15f  beta = %16.15f", __FUNCTION__, alpha, beta);
      fprintf(stderr, "  p = ( %16.15f %16.15f )", p_i.c[0], p_i.c[1]);
      fprintf(stderr, " sub = %d  size = ( %d x %d )\n", subsampling, nx, ny);
    }
    image_extract_translated_mark(I, R, subsampling, alpha, beta, p_i);
    
    char *R_name = NULL;
    asprintf(&R_name, "%s_%s.pgm", debug_name, tag);
    image_write_pgm (R, PGM_RAW_MAGIC, MAXVAL, R_name);
    image_free (R);
    free(R_name);
}

void lk_write_mark_difference_with_image
  ( image_t I, 
    image_t M, 
    image_t W, 
    int subsampling, 
    r2_t p_i, 
    double alpha, 
    double beta, 
    char *debug_name, 
    char *tag
  )
{
    /* Extract and write the occurrence {R} of the mark image as found in {I}: */
    image_t R = image_alloc (M->nx, M->ny); 
    image_extract_translated_mark(I, R, subsampling, alpha, beta, p_i);
    image_subtract_mark (R, M);
    if (W != NULL) image_multiply_by_normalized_weight (R, W);
    image_change_interval (R, -1, +1, 0, 1);
    
    char *R_name = NULL;
    asprintf(&R_name, "%s_%s.pgm", debug_name, tag);
    image_write_pgm (R, PGM_RAW_MAGIC, MAXVAL, R_name);
    image_free (R);
    free(R_name);
}

void lk_choose_plot_steps(int nz, double *step)
{
   assert(nz == 2);
   //step[0] = 0.1;     /* X position. */ 
   step[0] = 0.25;     /* X position. */ 
   //step[1] = 0.1;     /* Y position. */ 
   step[1] = 0.25;     /* Y position. */ 
}

void lk_choose_num_diff_steps(int nz, double *step)
{
   assert(nz == 2);
   step[0] = 0.02;     /* X position. */ 
   step[1] = 0.02;     /* Y position. */ 
}

void lk_gather_params (int nz, double *z, r2_t p_i) 
{
   assert(nz==2);
   z[0] = p_i.c[0];
   z[1] = p_i.c[1];
}

void lk_scatter_params (int nz, double *z, r2_t *p_i) 
{
   assert(nz==2);
   p_i->c[0] = z[0];
   p_i->c[1] = z[1];
}

double lk_compute_confidence_A (double errorSqr, double alpha, double beta, double csigma)
{
  double noise = 0.005; /* Nominal magnitude of pixel value noise. */
  double error = sqrt(errorSqr + noise*noise); 
  double contrast_factor = hypot(alpha, noise);
  double r = error/contrast_factor;
  return exp(-(r*r)/(2*csigma*csigma));
}

