#ifndef __lk_H__
#define __lk_H__

#include <image.h>
#include <r2x2.h>
#include "numeric.h"
#include <r3.h>

#define PLOT_ERROR 0 
/* Define as 1 to get a plot of the quadratic error {lk_Q}. */

#define TEST_ERROR 0 
/* Define as 1 to compare the algebraic and numerical
  derivatives of the quadratic error {lk_Q}. */

#define DEBUG_GIVEN_MARK_IMAGE 1 
/* Write the given mark image, as given ("M") and with {alpha/beta} correction ("Q"). */
 
#define DEBUG_WEIGHT_IMAGE 1
/* Write the weight mask image ("W"). */
  
#define DEBUG_GUESS_CONTEXT_IMAGE 1
/* Write image with a region of the frame image around the guessed point ("K"). */
 
#define DEBUG_ADJUSTED_MARK_IMAGE 1
/* Write the given mark image as found in the frame image ("N"), 
  with {alpha/beta} un-correction ("L"), and the
  discrepancy image ("D"). */
 
#define DEBUG_ADJUSTED_CONTEXT_IMAGE 1
/* Write image with a region of the frame image around adjusted point ("R"). */

#define DEBUG_CONTEXT_IMAGE_REL_SIZE 2.0
/* Relative size of region that is written out when {DEBUG_GUESS_CONTEXT_IMAGE} 
  or {DEBUG_ADJUSTED_CONTEXT_IMAGE} are set. This value is multiplied by
  the size of the target image. */

/* In all these functions, the image {W} is a weight mask that
  defines the relative importance of each mark pixel in the 
  quadratic error function {lk_Q} (and in its derivatives). The
  weight mask need not be normalized; */

typedef void lk_error_function_t
  ( image_t I,
    image_t M,
    image_t W,
    int subsampling,
    double alpha,
    double beta,
    r2_t p_i,
    double *Q,
    double *J, 
    double *H
  );
/* Type of a procedure that computes the quadratic error {*Q} between the template
  {M(u,v)}, translated by {p_i},, and the image {I(x,y)},
  weighted by {W(u,v)}.  The parameters
  {alpha, beta} are the color correction coefficients that must be
  applied to {M} in order to match its occurence in {I}.

  Also computes the gradient {J[0..1]} of {*Q}, and the Hessian
  {H[0..3]} of {*Q}, with respect to {p_i.c[0..1]}. The results Q, J and H
  are computed only if the arguments are not NULL*/

void lk_mark_position_adjust 
  ( char *fname,
    image_t I,
    image_t M,
    image_t W,
    int subsampling,
    double alpha,
    double beta,
    r2_t *p_i,
    r2_t *p_i_dev,
    lk_error_function_t error_func, 
    numeric_minimizer_t minimize,
    double step_factor,
    int niter,
    double *errorSqr
  );
/* the Lucas-Kanade (LK) algorithm with generic error function and
  minimization method.

  This procedure locates a mark image {M} in an image {I} by
  minimizing the quadratic discrepancy {Q} with a weight mask {W},
  computed by calling
  {error_func(I,M,W,subsampling,alpha,beta,p_i,Q,J,H)}.  The
  parameters {alpha, beta} are the color correction coefficients that
  must be applied to {M} in order to match its occurence in {I}.

  The minimum is found by calling the given {minimize} procedure with
  parameters {step_factor} and {niter}.  

  The location of {M} in {I} is defined by a point {*p_i} which is the
  position of the center of {M} in the domain of {I}, and a reduction
  factor {1/subsampling}.  On input, {*p_i} must contain the initial
  estimate for this point. On output, it will contain the optimal
  position found by the procedure; and {*errorSqr} will contain the
  quadratic discrepancy of {Q} of the match.

  The procedure assumes that the `a priori' probability distribution
  for the target's position along axis {i} is a Gaussian with mean
  {p_i.c[i]} and standard deviation {p_i_dev.c[i]}.

  The string {fname} is used to compose file names for debugging.  In
  particular, the procedure writes the following image files
   
     "{fname}_M.pgm" -- the mark {M};

     "{fname}_K.pgm" -- the context of the input (guessed) position {*p_i} in {I};

     "{fname}_L.pgm" -- the context of the output (adjusted) position {*p_i} in {I};

     "{fname}_D.pgm" -- the difference between the {M} and {L} images, weighted by {W}.

   */

void lk_choose_num_diff_steps(int nz, double *step);
/* Chooses a suitable numeric differentiation 
  step {step[k]} for each optimization parameter {z[k]},
  for {k} in {0..nz-1}. */

void lk_choose_plot_steps(int nz, double *step);
/* Chooses a suitable plotting 
  step {step[k]} for each optimization parameter {z[k]},
  for {k} in {0..nz-1}. */

void lk_write_target_image_region
  ( image_t I, 
    int minx, 
    int maxx, 
    int miny, 
    int maxy, 
    char *debug_name, 
    char *tag
  );
/* Extracts a subimage from {I} comprising columns {minx..maxx}
  and rows {miny..maxy}, and writes it file
  "{debug_name}_{tag}.ppm". Pixel values outside the image are 
  extapolated from the nearest existing pixel. */   

void lk_write_mark_extracted_from_image
  ( image_t I, 
    int nx, 
    int ny, 
    int subsampling, 
    r2_t p_i, 
    double alpha, 
    double beta, 
    char *debug_name, 
    char *tag
  );
/* Extracts a subimage of size {nx} by {ny} from {I}, centered at
   {p_i}, magnified by {subsampling} and writes it file
   "{debug_name}_{tag}.ppm". Pixel values are interpolated and
   corrected by {(v-beta)/alpha}.*/   


void lk_write_mark_difference_with_image
  ( image_t I, 
    image_t M, 
    image_t W, 
    int subsampling, 
    r2_t p_i, 
    double alpha, 
    double beta, 
    char *debug_name, 
    char *tag
  );
/* Extracts a subimage from {I} with the same size as {M}, centered at
  {p_i}, and subtracts from it the image {M}. If {W} is not NULL
  multiplies {W}, normalized o maximum 1.0 and magnified by
  {subsampling}, into the result image. Pixel values of {I} are
  implicitly corrected by {(v-beta)/alpha}. Writes the result (scaled
  from [-1,+1] to [0,255]) in file "{debug_name}_{tag}.ppm". */   

void lk_gather_params ( int nz, double *z, r2_t p_i );
/* Converts the point {p} into a generic parameter vector {z[0..nz-1]} 
  suitable for the Newton solver, plotting, etc. */

void lk_scatter_params ( int nz, double *z, r2_t *p_i );
/* Extracts the point {p} from the generic parameter vector {z[0..nz-1]}. */

double lk_compute_confidence_A (double errorSqr, double alpha, double beta, double csigma);
/* Computes the confidene weight for a mark, given the mean quadratic error {errorSqr} 
  between the reference image and the target image.  Assumes that the reference 
  image was color-adjusted by {alpha*v + beta} before the comparison.

  The parameter {csigma} should be the expected root-mean-square pixel error 
  that distinguishes a good match from a bad match, when there is no 
  color correction (i.e. when {alpha=1} and {beta=0}.) */


void lk_add_a_priori_error
  ( r2_t *q_i, 
    r2_t *p_i,
    r2_t *p_i_dev,
    int nz, 
    double *Q, 
    double *J, 
    double *H );
/* Adds a kalman error in {Q,J,H} proportional to how far is the deviation of {q_i} to {p_i}. 
   The matrix {E} has the shape of the weights.*/



#endif
