#include "lk_error_generic.h"
#include "lk.h"
#include "numeric.h"


void lk_error_generic 
   ( image_t I, 
     image_t M, 
     image_t W, 
     int subsampling,
     double alpha,  
     double beta,
     int color_flag,
     r2_t p_i,
     double *Q,
     double *J,
     double *H,
     lk_interpolation_function func_interpol,
     int iord,
     double *gw,                                /*smoothing weights*/
     int hgn                                    /*half width of smoothing weight table*/ 
   )
{
    int k;
    int Mnx = M->nx;
    int Mny = M->ny;
   
    assert(subsampling == 1); /* Generalize later? */

    int nz = 2;

    int jx, jy;
    if (Q != NULL) (*Q) = 0.0;
    if (J != NULL) for (k = 0; k < nz; k++) { J[k] = 0.0; }
    if (H != NULL) for (k = 0; k < nz*nz; k++) { H[k] = 0.0; }

   
    double sum_UW = 1.0e-200; 
    for (jx = 0; jx < M->nx; jx++) {
        for (jy = 0; jy < M->ny; jy++) {
            /* Get the mark value {M_val} and its weight {W_val} at the pixel center */
	    double gx, gy; /* Center of pixel {[jy][jx]}. */
            image_pixel_center (jx, jy, &gx, &gy); 
            double M_val, W_val, Mx_val, My_val, Mxx_val, Mxy_val, Myy_val; 

            image_eval_generic (  gx, 
                                  gy, 
                                  M,  
                                  &M_val, 
                                  &Mx_val, 
                                  &My_val, 
                                  &Mxx_val, 
                                  &Mxy_val, 
                                  &Myy_val, 
                                  W,  
                                  &W_val, 
                                  func_interpol, 
                                  iord, 
                                  gw, 
                                  hgn  );

            /* Get the image value {I_val} and its weight {U_val} at the corresponding point */
            double px, py; /* Point of {I} corresponding to {gx,gy}. */
            image_map_mark_coords_to_image_coords (gx, gy, Mnx, Mny, subsampling, p_i,  &px, &py);
            double I_val, U_val, Ix_val, Iy_val, Ixx_val, Ixy_val, Iyy_val;

            image_eval_generic (  px, 
                                  py, 
                                  I, 
                                  &I_val,
                                  &Ix_val, 
                                  &Iy_val, 
                                  &Ixx_val, 
                                  &Ixy_val, 
                                  &Iyy_val, 
                                  NULL, 
                                  &U_val, 
                                  func_interpol, 
                                  iord, 
                                  gw, 
                                  hgn  );
   
            /* The weight {w} of this pixel in {Q} is the mark weight times the image weight: */
	    double UW = U_val * W_val;
	    sum_UW += UW;
            
            /* Apply color correction to the mark value and gradient: */
            M_val = alpha * M_val + beta;
            Mx_val = alpha * Mx_val;
            My_val = alpha * My_val;

            /* Evaluate the color adjusted error term {E_val} */
            double E_val = M_val - I_val;

            /* Q */
            if (Q != NULL) {
               (*Q) += UW * E_val*E_val;
            }

            /* J */
            if (J != NULL) { 
               J[0] += - 2.0 * UW * E_val * (Mx_val + Ix_val);
               J[1] += - 2.0 * UW * E_val * (My_val + Iy_val);
            }
            
            /* H  */
            if (H != NULL) {
               H[0] += UW * pow(Mx_val + Ix_val, 2);
               H[1] += UW * (My_val + Iy_val) * (Mx_val + Ix_val);
               H[2] += UW * (My_val + Iy_val) * (Mx_val + Ix_val);
               H[3] += UW * pow(My_val + Iy_val, 2);
            }
            
     
	}
    }
    /* Normalize the quadratic error and derivatives to be pixel means: */
    if (Q != NULL) (*Q) /= sum_UW;
    if (J != NULL) for (k = 0; k < nz; k++) { J[k] /= sum_UW; }
    if (H != NULL) for (k = 0; k < nz*nz; k++) { H[k] /= sum_UW; }
}

void image_eval_generic
  ( double x, 
    double y, 
    image_t V, 
    double *v, 
    double *dx_v, 
    double *dy_v, 
    double *dxx_v,
    double *dxy_v,
    double *dyy_v,
    image_t W,
    double *w,
    lk_interpolation_function func_interpol,
    int iord,
    double *gw, /*smoothing weights*/
    int hgn     /*half width of smoothing weight table*/
    )
{
  /* This procedure performs a biquadratic interpolation of 
    pixel values of an implicitly smoothed image, then differentiates
    the interpolation. */

  /* What do we need to compute: */
  int v_ok = (v != NULL)||(dx_v != NULL)||(dxx_v != NULL);
  int dy_v_ok = (dy_v != NULL)||(dxy_v != NULL);
  int dyy_v_ok = (dyy_v != NULL);

  int w_ok = (w != NULL);
  if (! (v_ok | dy_v_ok | dyy_v_ok | w_ok)) { return; }

  /* Adjust the coordinates: */ 
  double rx = x - 0.5 * iord;
  double ry = y - 0.5 * iord;

  /* Compute the indices {ix,iy} of the first relevant sample: */
  int ix = floor(rx);
  int iy = floor(ry);

  /* Compute the local interpolation coordinates {xfrac,yfrac}: */
  double xfrac = rx - ix;
  double yfrac = ry - iy;

  //fprintf(stderr, "image_eval_generic fracs = (%8.6f %8.6f)\n", xfrac, yfrac);

  /* Extract a window with {npoints}x{npoints} starting at {ix,iy} with smoothing: */
  int npoints = iord + 1;
  double vtxy[npoints*npoints]; /* Pixel values. */
  double wtxy[npoints*npoints]; /* Reciprocals of pixel weights. */
  int jx, jy;
  for (jx = 0; jx < npoints; jx++) {
    for (jy = 0; jy < npoints; jy++) {
      int ind = jx*npoints + jy;
      double vt, wt;
      image_eval_aux_generic (V, W, ix+jx, iy+jy, &vt, &wt, gw, hgn);
      vtxy[ind] = vt;
      wtxy[ind] = (wt == 0 ? +HUGE_VAL : 1/wt);
      //fprintf(stderr, "[%5.3f %5.3f --> %8.1f]", vtxy[ind], wt, wtxy[ind]);
    }
    //fprintf(stderr, "\n");
  }

  /* Interpolate vertically: */
  double vtx[npoints];
  double dy_vtx[npoints];
  double dyy_vtx[npoints];
  double wtx[npoints];
  for (jx = 0; jx < npoints; jx++) {
    //for (jy = 0; jy < npoints; jy++)
    // {fprintf(stderr, "[%5.3f %8.1f]", vtxy[jx*npoints + jy], wtxy[jx*npoints + jy])};

    func_interpol
      ( yfrac, 
        &(vtxy[jx*npoints]), 
        &(vtx[jx]), 
        &(dy_vtx[jx]),
        &(dyy_vtx[jx])
      );
    func_interpol
      ( yfrac, 
        &(wtxy[jx*npoints]), 
        &(wtx[jx]), 
        NULL,
        NULL
      );
    //fprintf(stderr, " = [%5.3f %8.1f]\n", vtx[jx], wtx[jx]);
  }

  /* Interpolate horizontally: */
  
  if (v_ok) func_interpol (xfrac, vtx, v, dx_v, dxx_v);
  if (dy_v_ok) func_interpol (xfrac, dy_vtx, dy_v, dxy_v, NULL);
  if (dyy_v_ok) func_interpol (xfrac, dyy_vtx, dyy_v, NULL, NULL);


  /* Compute weight {*w} of result: */
  if (w_ok) {
    double wtxy;
    func_interpol (xfrac, wtx, &wtxy, NULL, NULL);
    (*w) = (wtxy == +HUGE_VAL ? 0.0 : 1/wtxy);
    //fprintf(stderr, " %8.1f -> %5.3f ]\n", wtxy, (*w));
  }
}


void image_eval_aux_generic
  ( image_t V, 
    image_t W, 
    int ix, 
    int iy, 
    double *v,
    double *w,
    double *gw, /*smoothing weights*/
    int hgn     /*half width of smoothing weight table*/
  )
{
  int nx = V->nx; if (W != NULL) { assert(nx == W->nx); }
  int ny = V->ny; if (W != NULL) { assert(ny == W->ny); }

  int jx, jy;

  /* Half-width and width of weight mask for Gaussian-smoothed derivative. */
  //int gn = (2*hgn + 1);

  double sum_wv = 0.0;
  double sum_w = 0.0;
  for(jx = -hgn; jx <= +hgn; jx++){
    if ((ix+jx >= 0) && (ix+jx < nx)) {
      int indx = jx + hgn;
      for(jy = -hgn; jy <= +hgn; jy++){
        if ((iy+jy >= 0) && (iy+jy < ny)) {
          int indy = jy + hgn;
          double wxy = (W == NULL ? 1.0 : W->val[iy+jy][ix+jx]) * gw[indx] * gw[indy];
          double vxy = V->val[iy+jy][ix+jx];
          sum_wv += wxy*vxy;
          sum_w += wxy;
            
	}
      }
    }
  }
  
  if (v != NULL) { (*v) = (sum_wv + 0.5e-200)/(sum_w + 1.0e-200); }
  if (w != NULL) { (*w) = sum_w; }
}

