#ifndef __lk_error_generic_H__
#define __lk_error_generic_H__

#include <image.h>
#include <r2x2.h>
#include <r3.h>

typedef void lk_interpolation_function
  ( double frac,
    double v[],
    double *z,
    double *dz,
    double *d2z
  );
  /* Type of a procedure that computes a goal function {*Q} depending on
    parameters {z[0..nz-1]}.  It shoudl also compute the gradient
    {J[0..nz-1]} of {*Q}, and the Hessian {H[0..nz*nz-1]} of {*Q}, with
    respect to {z[0..nz-1]}. The results {Q}, {J} and {H} should be
    computed if and only if the corresponding arguments are not NULL*/



void lk_error_generic 
   ( image_t I, 
     image_t M, 
     image_t W, 
     int subsampling,
     double alpha,  
     double beta,
     int color_flag,
     r2_t p_i,
     double *Q,
     double *J,
     double *H,
     lk_interpolation_function func_interpol,
     int iord,
     double *gw,                                /*smoothing weights*/
     int hgn                                    /*half width of smoothing weight table*/ 
   );
  /* Computes the error {*Q} between the template {M(u,v)}, 
     translated to point {p_i}, and the image {I(x,y)}, weighted by
    {W(u,v)}.

    The parameters {alpha, beta} are the color correction coefficients
    that must be applied to {M} in order to match its occurence in {I}.

    Also computes the gradient {J[0..1]} of {*Q}, and the Hessian
    {H[0..3]} of {*Q}, with respect to {p_i}. The results Q, J and H
    are computed only if the arguments are not NULL. */

void image_eval_generic
  ( 
     double x, 
     double y, 
     image_t V, 
     double *v, 
     double *dx_v, 
     double *dy_v, 
     double *dxx_v,
     double *dxy_v,
     double *dyy_v,
     image_t W,
     double *w,
     lk_interpolation_function func_interpol,
     int iord,
     double *gw, /*smoothing weights*/
     int hgn     /*half width of smoothing weight table*/
  );

  /* Evaluates the image (target or mark) {V} at the fractional point {x,y}.  Assumes
    that pixel centers are at half-integer coordinates.  The result is
    C0-continuous (a biquadratic B-spline).
    Returns the result in {*v}.  Also returns in {*dx_v,*dy_v}
    the spatial derivatives of the value. 

    This procedure ignores {W} and uses bilinear interpolation of
    pixel values of an implicitly smoothed image and pixel values of
    an implicitly smoothed derivative image.  It returns {*w = 1}. */

  /* Evaluates the image (target or mark) {V} at the fractional point {x,y}.  Assumes
    that pixel centers are at half-integer coordinates.  The result is
    C0-continuous (a biquadratic B-spline).
    Returns the result in {*v}.  Also returns in {*dx_v,*dy_v}
    the spatial derivatives of the value. 

    If {W} is not NULL, it must be an image with the same size as {V},
    containing a non-negative weight (``confidence'', ``importance'',
    etc.)  for each pixel.  Pixels outside the domain of {V} and {W}
    are assumed to have weight 0.  If {W} is NULL, the weights are
    assumed to be 1 inside the domain, 0 outside.
    
    The weight of each pixel is taken into account while smoothing 
    the image.

    The procedure returns in {*w} the weight of the result {*v}, defined
    as the the minimum of the weights of all pixels of {V}
    that were used in the computation of {*v}.  */

void image_eval_aux_generic
  ( 
     image_t V, 
     image_t W, 
     int ix, 
     int iy, 
     double *v,
     double *w,
     double *gw, /*smoothing weights*/
     int hgn     /*half width of smoothing weight table*/
  );

  /* Evaluates the image {V} at the pixel {[iy][ix]} with Gaussian smoothing.
    If the mask {W} is not null, each pixel of {V} is also weighted with the 
    corresponding pixel of {W} (besides the smoothing kernel weight).
    If {W} is NULL, assumes all its elements are 1.  Pixels outside the domain
    of {V} and {W} are assumed to have zero weight.  In any case, the procedure
    also returns in {*w} the sum of all pixel weights (kernel times mask) 
    that entered into the result {*v}.  In particular, if the whole kernel
    falls outside the mask and/or the domain of {V}, the result {*v}
    will be indeterminate, and {*w} will be 0. */    

#endif
