#ifndef __numeric_H__
#define __numeric_H__

#define _GNU_SOURCE
#include <stdio.h>
#include <rn.h>
#include <malloc.h>
#include <math.h>
#include <jsmath.h>

void numeric_diff
  ( double func (int nz, double *z),
    int nz,
    double *z_c,
    double *step, 
    double *J
  );
/* Given a function {func} with {nz} parameters {z[0..nz-1]}, computes the gradient
   {J[0..nz-1]} of {func} at the point {*z_c}, using weighted numerical
  differencing with step {step[i]} on each coordinate {z[i]}. */

void numeric_diff2
  ( double func (int nz, double *z),
    int nz,
    double *z_c, 
    double *step,
    double *H
  );
/* Given a function {func} with {nz} parameters {z[0..nz-1]}, computes the Hessian 
   {H[0..nz*nz-1]} of {func} at the point {*z_c}, using weighted numerical
  differencing with step {step[i]} on each coordinate {z[i]}. Stores in {H[nz*i+j]}
  the second derivative of {func} with respect to {z[i]} and {z[j]}.*/

double numeric_norm_vector ( double *a, int m );
/* Returns the Euclidean norm of {a[0..m-1]}*/

/* BASIC NEWTON STEP */

int numeric_Newton_step (int m, double *H, double *J, double *d, double factor);
/* Given a {m}x{m} matrix H and a {m} vector J, computes an m vector
   {d} such that {H} * {d} = -{J}.  If {H} is the Hessian matrix of
   quadratic positive definite function {Q}, and {J} is the gradient
   of Q at some point {u} then {u + d} the point of minimum of {Q}.
   
   Returns 0 (failure) if the determinant of {H} is zero.
   The computed step is multiplied by the given {factor}
   (1.0 gives the standard Newton step; 0.5 gives Pundlik's 
   modified Newton step).  */

double trim_step(int nz, double z0[], double z1[], double z_min[], double z_max[]);
/* Pulls {z1[0..nz-1]} towards {z0[0..nz-1]}
   until it lies within the rectangle defined by {zmin} and {zmax}. 
   Assumes that {z0} is within that region. Returns the reduction factor
   that was applied to the step (between 0 and 1). */ 

/* MINIMIZATION TOOLS */

typedef void numeric_goal_function_t  
  ( int nz,
    double z[],
    double *Q,
    double *J, 
    double *H
  );
  /* Type of a procedure that computes a goal function {*Q} depending on
    parameters {z[0..nz-1]}.  It shoudl also compute the gradient
    {J[0..nz-1]} of {*Q}, and the Hessian {H[0..nz*nz-1]} of {*Q}, with
    respect to {z[0..nz-1]}. The results {Q}, {J} and {H} should be
    computed if and only if the corresponding arguments are not NULL*/

typedef int numeric_minimizer_t
   ( int nz, 
     double z[], 
     numeric_goal_function_t goal_func,
     double step_factor, 
     double max_disp[], 
     int niter,
     double min_step
   );
  /* Type of an error mimizing procedure that modifies the parameter
    vector {z[0..nz-1]} so as to minimize the goal function
    {goal_func}.  

    The {niter} parameters should control the amount of effort done by
    the minimizer (e.g. the number of iterations of its internal
    loop); setting {niter == 0} causes the procedure to return without
    changing {z}.

    The {step_factor} parameter should control the length of the
    internal steps.  The normal behavior should be obtained with
    {step_factor==1}; a smaller value should slow down the speed of
    convergence, but hopefully enlarge the region of convergence.
    
    The parameters {max_disp[0..nz-1]} should bound the displacement
    between the input and output {z} vectors, in each coordinate.

    The {min_step} parameter specifies a convergence criterion
    (e.g. the minimum step between iterations), relatve to {max_disp}.

    The procedure should return 1 if it believes that the returned {z}
    is indeed a minimum of the function (e.g. because the gradient is
    zero or its internal steps were too small). 

    The procedure should return 0 if the returned {z} may not be a
    minimum (e.g. because it exceeded {niter} internal steps with no
    convergence).

    In any case, the procedure should never return a {z} that
    is worse (has a higher {Q} value) than the initial one. 

    !!! A {minimizer_t} should also return the value of {Q} and
    possibly of {J,H} at the returned {z} point. Thus it must do at
    least one call to {goal_func}, even if {niter} is
    zero. Alternatively, we should require that the client gives
    {Q,J,H} for the initial point {z}. !!!  */

#define DEBUG_CONV 1
/* If 1, minimizers will print a character to indicate the outcome. */

#define TRACE_CONV(c) do { if (DEBUG_CONV) { fputc(c, stderr); } } while(0)

int numeric_iterated_Newton
   ( int nz, 
     double z[], 
     numeric_goal_function_t goal_func,
     double step_factor, 
     double max_disp[], 
     int niter,
     double min_step
   );
/* A {numeric_minimizer_t} procedure that iterates the basic Newton
  step.  It stops with success when a Newton step moves less than
  {min_step} times {max_disp[k]} along every axis {k} in {0..nz-1}.
  It stops with failure when the Newton step returns 0, or after
  {niter} iterations, or when a coordiante {z[k]} differs by
  {max_dist[k]} or more from the starting point, for any {k}. */

int numeric_new_safe_Newton
  ( int nz, 
     double z[], 
     numeric_goal_function_t goal_func,
     double step_factor, 
     double max_disp[], 
     int niter,
     double min_step
   );
  /* At each iteration, applies one Newton step to the current guess
    {z[0..nz-1]}; but trims or reverses the step if necessary, to 
    ensure that its length does not exceed {max_disp}, and that the 
    goal function does not increase.  Otherwise, the behavior is 
    similar to that of {numeric_iterated_Newton}. */

int numeric_STB_safe_Newton
  ( int nz, 
     double z[], 
     numeric_goal_function_t goal_func,
     double step_factor, 
     double max_disp[], 
     int niter,
     double min_step
   );
  /* Old minimization method (was inside {lk_adjust_mark_position}).

    At each iteration, applies {numeric_STB_safe_Newton_step} to the current guess
    {z[0..nz-1]}, with the parameter {Dmax} decreasing in geometric progression.
    It ignores {step_factor,max_disp,min_step}. */

double numeric_STB_safe_Newton_step
  ( numeric_goal_function_t goal_func, 
    double *z, 
    int nz, 
    double *Q, 
    double *J, 
    double *H, 
    double Dmax
  );
/* Given a {m}x{m} matrix H and a {m} vector J, computes an m vector
   {d} such that {H} * {d} = -{J}.  If {H} is the Hessian matrix of
   quadratic positive definite function {Q}, and {J} is the gradient
   of Q at some point {u} then {u + d} the point of minimum of {Q}.

   Then adds {dz} to the argument vector {z[0..nz-1]}, but trims the
   step to maximum length {Dmax}.  Recomputes {Q,J,H} at the new {z},
   and then tries a few more points, looking for a {z} that has
   smaller function value than the original {Q}.  If it fails, leaves
   {z,Q,J,H} unchanged. If it succeeds, updates those values with the
   best point found.  Returns the determinant of {H}.  */

/* PLOTTING AND TESTING GOAL FUNCTIONS */

void numeric_plot_goal_func_and_derivatives   
  ( char *name,
    char *tag,
    numeric_goal_function_t goal_func, 
    int nz,
    double z[],
    double diff_step[],
    double dz[],
    double plot_step[],
    int N);
/* Writes to a plot file called "{name}_{tag}.txt" the value {Q} of
  function {goal_func} and its algebraic and numerical derivatives, for various
  parameter vectors {z_m[0..nz-1]} that differ from the given posiiton
  {z} in each coordinate {k} by {r*dz[k]*plot_step[k]}, where {r}
  varies over {-N..N}. 

  Each line contains the following fields
  
     (1) the distance between {z} and {z_m},
     (2) the value of {numeric_Q(z_m)}, 
     (3,4) the algebraic and numeric first derivs of {numeric_Q} in the direction {dz};
     (5,6) the algebraic and numeric second derivs  {numeric_Q} in the direction {dz}.  
     (7..7+nz-1) the coordinates {z_m}.

  The numeric derivatives are computed using the steps {diff_step[0..nz-1]}.  */ 

void numeric_plot_all_goal_funcs_and_derivatives
  ( char *name,
    numeric_goal_function_t goal_func, 
    int nz,
    double z[],
    double diff_step[],
    double plot_step[],
    int N
  );
/* Calls {numeric_plot_goal_func_and_derivatives} for several
  direction vectors {dz}, including all axial directions and the space
  diagonals {+1,+1}, and, {+1,-1}.  The numeric derivatives are
  computed using the steps {diff_step[0..nz-1]} */

void numeric_plot_2D_goal_func_and_derivatives
  ( char *outname,
    int nz,
    double z[],
    double plot_step[],
    numeric_goal_function_t goal_func, 
    numeric_minimizer_t minimize,
    double step_factor,
    double max_disp[],
    int niter,
    double min_step, 
    int N
  );
/* Writes to a plot file called "{outname}.txt" the value {Q} of
  function {goal_func} and its algebraic
  derivatives, as computed by the client-given {error_func},
  for various positions {z_m} around {z[0..nz-1]}.  Assumes
  that {nz == 2} for now.

  Each position {z_m} differs from the given position
  {z} in each coordinate {k} by {i[k]*plot_step[k]},
  where {i[k]} varies over {-N..N}, and {step}
  is the step vector defined by {numeric_choose_plot_steps}. 

  Also tests whether a client-given minimizer algorithm {minimize}
  converges when applied to the given {goal_func} starting at {z_m},
  with parameters {max_disp}, {niter}, {step_factor}, and {min_step}.

  Currently assumes that {nz==2}.  Each line then contains the following fields
  
     (1,2) the difference between {z} and {z_m};  
     (3,4) the parameter vector {z_m};
     (5) the value of {Q(z_m)}; 
     (6,7) the algebraic gradient of {Q} at {z_m};
     (8,9,10,11) the algebraic Hessian (second derivs) {Q} at {z_m};
     (12,13) the position {z_u} resulting from {minimize} at {z_m}, with {niter=1}. 
     (14) minimizer's convergence flag (0 = didn't, 1 = did);
     (15,16) the position {z_s} resulting from {minimize} at {z_m}, with the given {niter}.  */ 

void numeric_test_goal_func_derivatives
  ( numeric_goal_function_t goal_func, 
    int nz,
    double z[],
    double diff_step[]
  );
/* Writes to a plot file called "{fname}_{KK}.dat" the values of
  {numeric_Q} and its algebraic and numerical derivatives, for various
  parameter vectors {z_m[0..nz-1]} that differ from the given vector
  {[0..nz-1]} in coordinate {k}. The numeric derivatives are computed
  using the steps {diff_step[0..nz-1]}. */

void numeric_dn_goal_func
  ( numeric_goal_function_t goal_func, 
    int nz,
    double z[],
    double diff_step[],
    double J_num[]
 );
/* Computes the gradient {J_num[0..nz-1]} of the error function at
  argument {z[0..nz-1]} by numeric differentiation of the error, as
  computed {goal_func}. The numeric derivatives are computed using the
  steps {diff_step[0..nz-1]}. */

void numeric_dn_d_goal_func
  ( numeric_goal_function_t goal_func, 
    int nz,
    double z[],
    double diff_step[],
    double H[]
 );
/* Computes the Hessian {H[0..nz*nz-1]} of the error function at
  argument {z[0..nz-1]} by numeric differentiation of the Jacobian, as
  computed by {goal_func}. The numeric derivatives are computed using
  the steps {diff_step[0..nz-1]}. */

void numeric_d2n_goal_func
  ( numeric_goal_function_t goal_func, 
    int nz,
    double z[],
    double diff_step[], 
    double H[]
   );
/* Computes the Hessian {H[0..nz*nz-1]} of the error function at
  argument {z[0..nz-1]} by double numeric differentiation of the error
  value, as computed by {goal_func}. The numeric derivatives are
  computed using the steps {diff_step[0..nz-1]}. */

void numeric_debug(double x);
/* A bogus procedure for {gdb} breakpoints. */

#endif
