#include "numeric.h"
#include "numeric_min_nwgrad.h"
#include <assert.h>
#include <rmxn.h>
#include <rn.h>

int numeric_grad_Newton
  ( int nz, 
     double z[], 
     numeric_goal_function_t goal_func,
     double step_factor, 
     double max_disp[], 
     int niter,
     double min_step
  )
  { int k;
    double Q; 
    double J[nz];
    double H[nz*nz];

    // fprintf(stderr, "numeric grad newton\n");

    /* Boundaries of search area: */
    double z_min[nz], z_max[nz];
    for (k = 0; k < nz; k++) {
      z_min[k] = z[k] - max_disp[k] - 0.0001;
      z_max[k] = z[k] + max_disp[k] + 0.0001;
    }

    if (niter == 0) { TRACE_CONV('0'); return 0; }

    /* Evaluate the goal function at {z}: */
    goal_func(nz, z, &Q, J, H);

    /* Iterate the mixed Newton-gradient step: */
    int iter = 0;
    while (iter < niter) {
      /* Compute the mixed gradient-Newton's step: */
      double grad_step = 15.0 * min_step;
      int improved = numeric_grad_Newton_step
        ( nz, z, goal_func,
          &Q, J, H, step_factor,
          z_min, z_max, grad_step, min_step
        );
      if (! improved) { /* Looks like a minimum: */ TRACE_CONV('!'); return 1; }
      iter++;
    }
    /* We gave up before reaching a minimum: */
    TRACE_CONV((iter <= niter ? '0' + iter : '*')); 
    return 0;
  }

int numeric_grad_Newton_step
  ( int nz, 
    double z[], 
    numeric_goal_function_t goal_func,
    double *Q, 
    double *J, 
    double *H,
    double step_factor, 
    double z_min[], 
    double z_max[],
    double grad_step, 
    double min_step
  )
  {
    double dz[nz]; /* Step to take from {z}. */

    double z_t[nz];    /* Trial point (end of step). */
    double Q_t;        /* Function value at {z_t}. */
    double J_t[nz];    /* Function gradient at {z_t}. */
    double H_t[nz*nz]; /* Function Hessian at {z_t}. */
    int k;

    int strategy = 0; /* What strategy to try next. */

    while (1) {

      /* Apply the next {strategy} to obtain a new trial step. Set {ok} if succeeded. */
      int step_ok = 0; /* Step {dz} is a good *direction*. */
      
      switch (strategy) {
        case 0:
         { /* Try the Newton step: */
	   // fprintf(stderr, "caso 0\n");
           step_ok = numeric_Newton_step (nz, H, J, dz, step_factor);
	   // for (k = 0; k < nz; k++) fprintf(stderr, "%f, %f\n", z[k], dz[k]);
         } break;

        case 1:
	  { /* Try a gradient descent step: */
	    // fprintf(stderr, "caso 1\n");
	    double Jnorm = rn_norm(nz, J);
	    if (Jnorm == 0) { /* Can't use this direction: */ step_ok = 0; break; }
	    for (k = 0; k < nz; k++) { dz[k] = - grad_step * J[k]/Jnorm; }
	    // for (k = 0; k < nz; k++) { fprintf(stderr, "%f, %f\n", z[k], dz[k]); }
	    step_ok = 1; 
            // fprintf(stderr, "descendo pelo gradiente\n");
	  } break;

        case 2:
	  { /* Reduce the current step by half: */
	    // fprintf(stderr, "caso 2\n");
	    for (k = 0; k < nz; k++) { dz[k] *= 0.5; }
	    step_ok = 1;
	    // for (k = 0; k < nz; k++) { fprintf(stderr, "%f, %f\n", z[k], dz[k]); }
	  } break;

        case 3:
         { /* Give up: */
           /* !!! Perhaps a brute force search could be advisable here. */
           /* Do not change {z,Q,J,H} but say that step "converged": */
	   // fprintf(stderr, "caso 3\n");
           return 0; 
         } break;

        default:
          { assert(0); }
      }

      if (step_ok) {
        /* Direction of {dz} is OK, but it may be too long or too short. */
        /* Try to use this step, and exit the procedure if successful. */
        /* If the attempt fails, set {step_ok} appropriately. */ 
	// fprintf(stderr, "Step ok\n");
        
        /* Apply step {dz} from {z} to {z_t}: */
	for (k = 0; k < nz; k++) { z_t[k] = z[k] + dz[k]; }
	// for (k = 0; k < nz; k++) { fprintf(stderr, "z_t[%d] = %f\n", k, z_t[k]); }
	/* Trim the step to the valid search region: */
	(void)trim_step(nz, z, z_t, z_min, z_max);
        /* Evaluate the function at the end of the step: */
        goal_func(nz, z_t, &Q_t, J_t, H_t);
	// for (k = 0; k < nz; k++) { fprintf(stderr, "novo z_t[%d] = %f\n", k, z_t[k]); }
        if (Q_t < (*Q))
          { /* Success! Move {z} to {z_t} and update {Q,J,H}: */
            (*Q) = Q_t;
            for (k = 0; k < nz; k++) { z[k] = z_t[k]; J[k] = J_t[k]; }
            for (k = 0; k < nz*nz; k++) { H[k] = H_t[k]; }
	    // fprintf(stderr, "melhorou ainda nao pare\n");
            return 1; /* Improved, so we did not converge yet! */
          }
        /* Check step length: */ 
        double dzlen = rn_dist(nz, z, z_t);
	// fprintf(stderr, "dzlen %f min_step %f\n", dzlen, min_step);
	if (dzlen < min_step) { /* Step is useless now: */ step_ok = 0; }
      }

      /* We are still at the starting point. */
      /* Choose the strategy to generate the next step: */
      switch (strategy) 
        {
	case 0: 
          /* Step {dz} was generated by Newton's method. */
	  /* There is no point in retrying Newton or shortening the step. */
	  /* Try gradient descent instead: */
	  // fprintf(stderr, "voltar ao caso 0\n");
	  strategy = 1; break;

	case 1:
	  /* Step was generated by gradient descent.  There is no point in that */
	  /* strategy, but we may try shortening the step. */
	  // fprintf(stderr, "voltar ao caso 1\n");
	  strategy = 2; break;

	case 2:
	  /* Step was generated by shortening the grad step. */
	  /* We can retry if the step is still valid (and long enough). */
	  /* Otherwise we try the next strategy. */
	  // fprintf(stderr, "voltar ao caso 2 ou 3\n");
	  strategy = (step_ok ? 2 : 3); break;

	default:
	  assert(0);
	}
    }
  }
