
#define PROG_NAME "test_calib"
#define PROG_DESC "tests the Tsai camera calibration algorithm"

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <image.h>
#include <image_list.h>
#include <r2x2.h>
#include <r2.h>
#include <tf_camera.h>
#include <tf_targets.h>
#include <tf_calib.h>

void at_estimate_alpha_beta
  ( image_t I,
    image_t M,
    image_t W,
    int subsampling,
    r2_t p_i,
    double *alpha,
    double *beta
  );

int main (int argc, char *argv[])
{
  /*Verify memory*/
  void *trash = malloc(1);
  struct mallinfo info;
  int MemDinInicial, MemDinFinal;
  /*End*/

  free(trash);
  info = mallinfo();
  MemDinInicial = info.uordblks;

  /*#*/char *image_pos_fname = argv[3]; // pi[k]
  /*#*/char *weights_fname = argv[4];   // wgts[k]
  /*#*/char *target_info_fname = argv[5];  // type[k], u[k], v[k]
  /*#*/char *output_image_name = argv[6];
  /*#*/char *input_image_name = argv[1];

  char *world_coords_fname = argv[2]; // pw[k]
  int n_target_types = atoi(argv[7]); /* Number of distinct targets: */
  char **target_image_names = &(argv[8]); 
  
  image_t I = image_read_pgm (input_image_name);
  image_t I_with_marks = image_copy (I);

  image_t OUT = image_alloc(I->nx, I->ny);
 
  /* Choose the central parameters: */
  r2_t p_i;
  p_i.c[0] = I->nx/2.0;
  p_i.c[1] = I->ny/2.0;

  /* Choose the number of plotting steps: */ 
  double alpha = 1.0;
  double beta = 0.0;

  /* Read the generic camera specs: */
  camera_specs_t cspec = tf_get_svga_povray_specs ();
  fprintf(stderr, "Generic camera parameters:\n");
  tf_show_camera_specs (cspec, stderr);

  /* Get the default values for the camera parameters: */
  camera_parameters_t cpar = tf_get_mean_camera_parameters (cspec);
  fprintf(stderr, "Default camera parameters:\n");
  tf_show_camera_parameters (cpar, stderr);

  /* Read a set of camera calibration data: */
  calibration_data_t cdat = tf_calibration_data_read
    ( world_coords_fname, 
      image_pos_fname, 
      weights_fname );
  fprintf(stderr, "Calibration data:\n");
  tf_calibration_data_print(cdat, stderr);

  /* Read the target images and descriptions (for output visualization) */
  targets_data_t tdat = tf_targets_data_read 
    ( world_coords_fname, 
      target_info_fname, 
      image_pos_fname, 
      weights_fname );
  demand(tdat->ntargets == cdat->nmarks, "inconsistent counts tdat/cdat");

  /* Read the ideal target images and create their weight masks: */
  int useGaussianWeights = 0; 
  image_list_t M_list = NULL; /* Target images. */
  image_list_t W_list = NULL; /* Weight masks for target images. */
  image_list_t T_list = image_list_read (n_target_types, target_image_names);
  M_list = image_list_copy_selected (T_list, tdat->ntargets, tdat->type);
  image_list_free(T_list);
  /* Create masks (uniform for STB error, Gaussian otherwise): */
  W_list = image_list_create_weight_masks (M_list, useGaussianWeights);

  image_fill (OUT, 1.0);

  /*camera calibration algorithm to find the camera position*/
  int mark, subsampling = 2;
  FILE *ferr;
  ferr = fopen("ferr.txt", "w");
  tf_calib1_tsai (cdat, cspec, cpar, ferr);
  fclose(ferr); 

  fprintf(stderr, "Calibrated camera parameters:\n");
  tf_show_camera_parameters (cpar, stderr);

  for (mark = 0; mark < cdat->nmarks; mark++) {
 
      r2_t computed_p_i;
      tf_mark_estimate_position_and_shape 
      (   cpar,
          tdat->p_w[mark], tdat->u_w[mark], tdat->v_w[mark],
          &computed_p_i, &(tdat->u_i[mark]), &(tdat->v_i[mark])
      );
      
      image_t G, V;

      image_deform(M_list->imgs[mark], W_list->imgs[mark], &(tdat->u_i[mark]), &(tdat->v_i[mark]), subsampling, &G, &V);

      at_estimate_alpha_beta (I, G, V, subsampling, computed_p_i, &alpha, &beta);

      image_paint_translated_mark (G, V, OUT, subsampling, alpha, beta, computed_p_i);

      if (mark == 11) {
         int magnify = 4;
         int radius_x = 15;
         int radius_y = 15;

         image_t feature_img = image_extract_subimage_by_pos (I, tdat->p_i[mark], radius_x, radius_y, magnify);

         char *N_name = NULL;
         char *M_name = NULL;
         char *G1_name = NULL;
         char *G2_name = NULL;
         char *W1_name = NULL;
 
         asprintf(&N_name, "test_N_%02d.pgm", mark);
         asprintf(&M_name, "test_M_%02d.pgm", mark);
         asprintf(&G1_name, "test_G1_%02d.pgm", mark);
         asprintf(&G2_name, "test_G2_%02d.pgm", mark);
         asprintf(&W1_name, "test_W1_%02d.pgm", mark);

         image_write_pgm (feature_img, PGM_RAW_MAGIC, MAXVAL, N_name);
         image_write_pgm (M_list->imgs[mark], PGM_RAW_MAGIC, MAXVAL, M_name);
     
         image_t tmp0 = image_magnify (G, 2);
         image_t tmp1 = image_magnify (G, 1);

         image_write_pgm (tmp0, PGM_RAW_MAGIC, MAXVAL, G1_name);
     
         image_apply_color_correction (tmp0, alpha, beta);

         image_write_pgm (tmp0, PGM_RAW_MAGIC, MAXVAL, G2_name);
         image_write_pgm (tmp1, PGM_RAW_MAGIC, MAXVAL, W1_name);
 
         //image_t OUT2 = image_alloc(I->nx, I->ny);

         free(N_name);
         free(M_name);
         free(G1_name);
         free(G2_name);
         free(W1_name);
         image_free(feature_img);

         image_free(tmp0);
         image_free(G);
         image_free(V);
      }

      image_print_cross (I_with_marks, 5, computed_p_i);

  }

  image_write_pgm (OUT, PGM_RAW_MAGIC, MAXVAL, output_image_name); 
  image_write_pgm (I_with_marks, PGM_RAW_MAGIC, MAXVAL, "marked.pgm"); 

  image_free (I_with_marks);
  image_free (I);
  image_free (OUT);

  info = mallinfo();
  MemDinFinal = info.uordblks;
  if (MemDinInicial!=MemDinFinal)
    fprintf(stderr, "\n\nMain loop: Dinamic memory was not completely deallocated (%d, %d)\n", MemDinInicial,MemDinFinal);
  return 0;
}

void at_estimate_alpha_beta
  ( image_t I,
    image_t M,
    image_t W,
    int subsampling,
    r2_t p_i,
    double *alpha,
    double *beta
  )
{
  int M_nx = M->nx;
  int M_ny = M->ny;

  /* Compute mean and average of mark in the reference image {M}: */
  r2_t trivial_p_i = (r2_t){{ M_nx/2.0, M_ny/2.0 }};

  double Mavg, Mvar;
  image_weighted_avg_var (M, W, 1, trivial_p_i, &Mavg, &Mvar);
  fprintf(stderr, "Mavg %f Mvar %f\n", Mavg, Mvar);

  /* Compute mean and average of mark in the frame image {I}: */

  double Iavg, Ivar;
  image_weighted_avg_var (I, W, subsampling, p_i, &Iavg, &Ivar);
  fprintf(stderr, "Iavg %f Ivar %f\n", Iavg, Ivar);
    
  *alpha = sqrt (Ivar/Mvar);
  if (*alpha < 0.1) {*alpha = 0.1;}
  if (*alpha > 1.0) {*alpha = 1.0;}
    
  *beta = Iavg - (*alpha)*Mavg;
    
  if (*beta < 0.0) {*beta = 0.0;}

  fprintf(stderr, "Estimated alpha = %f beta = %f\n", *alpha, *beta);
}
 
