
#define PROG_NAME "test_calib"
#define PROG_DESC "tests the Tsai camera calibration algorithm"

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <image.h>
#include <image_list.h>
#include <r2x2.h>
#include <r2.h>
#include <jsfile.h>
#include <tf_camera.h>
#include <tf_errors.h>
#include <tf_targets.h>
#include <tf_calib.h>
#include <tf_calib_guess1.h>
#include <tf_calib_guess2.h>
#include <tf_calib_refine2.h>
#include <sys/stat.h>
#include <sys/types.h>


void at_estimate_alpha_beta
  ( image_t I,
    image_t M,
    image_t W,
    int subsampling,
    r2_t p_i,
    double *alpha,
    double *beta
  );

void tf_show_all_marks
  ( camera_parameters_t cpar, 
    targets_data_t tdat,
    image_t img,
    image_t out_mark_img,
    image_t out_cross_img,
    image_list_t M_list,
    image_list_t W_list,
    bool_t cross /*TRUE paints crosses, FALSE paints dots.*/
  );

int main (int argc, char *argv[])
{
  /*Verify memory*/
  void *trash = malloc(1);
  struct mallinfo info;
  int MemDinInicial, MemDinFinal;
  /*End*/

  free(trash);
  info = mallinfo();
  MemDinInicial = info.uordblks;

  /* Where to find the images (for visualization): */
  char *frame_prefix = argv[1];
  int fbegin = atoi(argv[2]);
  int fend = atoi(argv[3]);

  char *camera_name = argv[4];
  char *image_data_prefix = argv[5]; /* Prefix for frame {p_i.txt,p_wgt.txt,true.cpar}. */
  char *world_coords_fname = argv[6]; /* Name of file with makr {p_w}s. */
  char *world_shapes_fname = argv[7];  /* Name of file with mark types and world {u,v}s. */
  
  char *out_dir = argv[8]; /* Directory for all output files. */
  int which_calib = atoi(argv[9]); /* Version of {tf_calib} to use. */
  int nl_optimization = (atoi(argv[10]) != 0); /* 0 to skip the NL optimization. */
  int n_target_types = atoi(argv[11]); /* Number of distinct targets. */
  char **target_image_names = &(argv[12]);
  fprintf(stderr, "Starting:\n");
  
  /* Get the fixed camera parameters: */
  camera_specs_t cspec;
  
  if (strcmp(camera_name,"optura") == 0) {
    cspec = tf_get_canon_optura_specs ();
  }
  else if (strcmp(camera_name,"svga") == 0) {
    cspec = tf_get_svga_povray_specs ();
  }
  else if (strcmp(camera_name,"hvga") == 0) {
    cspec = tf_get_hvga_povray_specs ();
  }
  else {
    fprintf(stderr, "error: define the camera specifications\n");
    exit(1);
  }

  fprintf(stderr, "Generic camera parameters:\n");
  tf_show_camera_specs (cspec, stderr);

  /* A priori uncertainty on mark image positions (for error computation): */
  r2_t p_i_dev = (r2_t){{ 5.0, 5.0 }};  /* Should be a parameter... */

  /* Shall we include the {cpar} Kalman error terms in the optimization? */
  bool_t use_cpar_dev_terms = FALSE;

  /*Output data directory*/
  mkdir(out_dir, 0777);

  camera_parameters_t cpar = tf_get_mean_camera_parameters(cspec); /* Calibrated. */
  camera_parameters_t tpar = tf_get_mean_camera_parameters(cspec); /* True. */

  /*Calibration debug messages*/
  int frame;
  for (frame = fbegin; frame <= fend; frame++) {

    char *frame_name = NULL;
    asprintf(&frame_name, "%s%05d.pgm", frame_prefix, frame);
    image_t img = image_read_pgm (frame_name); 

    char *image_points_fname = NULL;
    asprintf(&image_points_fname, "%s%05d/p_i.txt", image_data_prefix, frame);
    char *image_weights_fname = NULL;
    asprintf(&image_weights_fname, "%s%05d/p_wgt.txt", image_data_prefix, frame);
    char *image_tpar_fname = NULL;
    asprintf(&image_tpar_fname, "%s%05d/true.cpar", image_data_prefix, frame);

    /* Read a set of camera calibration data: */
    calibration_data_t cdat;
    cdat = tf_calibration_data_read (world_coords_fname, image_points_fname, image_weights_fname);
    fprintf(stderr, "calibration data:\n");
    tf_calibration_data_print(cdat, stderr); 
    fprintf(stderr, "-----------------------------\n");

    /* Read the target images and descriptions (for output visualization) */
    targets_data_t tdat;
    tdat = tf_targets_data_read
      ( world_coords_fname, world_shapes_fname, image_points_fname, image_weights_fname);

    /* Read the true {cpar}, if present: */
    FILE *ftpar = fopen(image_tpar_fname, "r");
    bool_t has_tpar = (ftpar != NULL);
    if (has_tpar) {
      tf_read_camera_parameters(ftpar, tpar);
      fclose(ftpar);
    }

    /* Output file with calibration errors: */
    char *ferr_fname = NULL;
    asprintf(&ferr_fname, "%s/out%05d.ferr", out_dir, frame);
    FILE *ferr = open_write(ferr_fname, TRUE);

    /* Output file with calibrated parameters: */
    char *fcpar_fname = NULL;
    asprintf(&fcpar_fname, "%s/out%05d.cpar", out_dir, frame);
    FILE *fcpar = open_write(fcpar_fname, TRUE);

    if (which_calib == 1)
      { assert(FALSE); }//tf_calib1_tsai (cdat, cspec, cpar, TRUE, nl_optimization); }
    else if (which_calib == 2)
      { tf_calib_guess2_initial_camera_parameters
          (cdat, cspec, cpar, FALSE, TRUE);
        if (nl_optimization) {
          tf_calib_generic
	    ( cdat, cspec, p_i_dev, use_cpar_dev_terms, cpar,
              tf_calib_refine2_gather_optimization_params,
              tf_calib_refine2_scatter_optimization_params,
              TRUE);
	}
      }
    if (fcpar != NULL) { tf_write_camera_parameters(fcpar, frame, cpar); fclose(fcpar); }
    if (ferr != NULL) { tf_calib_summarize_errors(cdat, cpar, ferr); fclose(ferr); }

    fprintf(stderr, "Calibrated parameters:\n");
    tf_show_camera_parameters (cpar, stderr);

    fprintf(stderr, "True parameters:\n");
    tf_show_camera_parameters (tpar, stderr);

    /* Read the ideal target images and create their weight masks: */
    int useGaussianWeights = 0; 
    image_list_t M_list = NULL; /* Target images. */
    image_list_t W_list = NULL; /* Weight masks for target images. */
    image_list_t T_list = image_list_read (n_target_types, target_image_names);
    M_list = image_list_copy_selected (T_list, tdat->ntargets, tdat->type);
    image_list_free(T_list);
    /* Create masks (uniform for STB error, Gaussian otherwise): */
    W_list = image_list_create_weight_masks (M_list, useGaussianWeights);  

    image_t out_mark_img = image_copy(img);
    image_t out_cross_img = image_copy(img);

    if (has_tpar) {
      tf_show_all_marks(tpar, tdat, img, out_mark_img, out_cross_img, M_list, W_list, FALSE);
    }
    tf_show_all_marks(cpar, tdat, img, out_mark_img, out_cross_img, M_list, W_list, TRUE);

    char *out_mark_fname = NULL;
    asprintf(&out_mark_fname, "%s/out%05d_mark.pgm", out_dir, frame);
    image_write_pgm (out_mark_img, PGM_RAW_MAGIC, MAXVAL, out_mark_fname);
    
    char *cross_fname = NULL;
    asprintf(&cross_fname, "%s/out%05d_cross.pgm", out_dir, frame);
    image_write_pgm (out_cross_img, PGM_RAW_MAGIC, MAXVAL, cross_fname);
    
    free(out_mark_fname);
    free(cross_fname);

    free(ferr_fname);
    free(fcpar_fname);

    free(image_points_fname);
    free(image_weights_fname);
    free(image_tpar_fname);

    image_free(out_mark_img);
    image_free(out_cross_img);
    image_free(img);
    image_list_free(M_list);
    image_list_free(W_list);
    tf_targets_data_free(tdat); 
  }

  tf_free_camera_parameters_structure (cpar);

  /*Missing dealocattes cspec structure, to create */

  info = mallinfo();
  MemDinFinal = info.uordblks;
  if (MemDinInicial!=MemDinFinal)
    fprintf(stderr, "\n\nMain loop: Dinamic memory was not completely deallocated (%d, %d)\n", MemDinInicial,MemDinFinal);
  return 0;
}

void tf_show_all_marks
  ( camera_parameters_t cpar, 
    targets_data_t tdat,
    image_t img,
    image_t out_mark_img,
    image_t out_cross_img,
    image_list_t M_list,
    image_list_t W_list,
    bool_t cross /*TRUE paints crosses, FALSE paints dots.*/
  )
{    
    int mark;
    for (mark = 0; mark < tdat->ntargets; mark++) {

      r2_t p_i;
      tf_mark_estimate_position_and_shape
	( cpar, tdat->p_w[mark], tdat->u_w[mark], tdat->v_w[mark], 
          &p_i, &(tdat->u_i[mark]), &(tdat->v_i[mark]) );

      image_t G, V;

      int subsampling = 2;
      image_deform
        ( M_list->imgs[mark], W_list->imgs[mark], 
          &(tdat->u_i[mark]), &(tdat->v_i[mark]), 
          subsampling, &G, &V);
   
      double alpha, beta;
      at_estimate_alpha_beta (img, G, V, subsampling, p_i, &alpha, &beta);
 
      image_paint_translated_mark (G, V, out_mark_img, subsampling, alpha, beta, p_i); 

      int mark_radius = 5;
      fprintf(stderr, "mark %d cross in %f %f image position\n", mark, p_i.c[0], p_i.c[1]);
      if (cross) {
        image_print_cross (out_cross_img, mark_radius, p_i);
      } else {
        image_print_dot (out_cross_img, mark_radius, p_i);
      }

      image_free(G);
      image_free(V);
    }
}

void at_estimate_alpha_beta
  ( image_t I,
    image_t M,
    image_t W,
    int subsampling,
    r2_t p_i,
    double *alpha,
    double *beta
  )
{
  int M_nx = M->nx;
  int M_ny = M->ny;

  /* Compute mean and average of mark in the reference image {M}: */
  r2_t trivial_p_i = (r2_t){{ M_nx/2.0, M_ny/2.0 }};

  double Mavg, Mvar;
  image_weighted_avg_var (M, W, 1, trivial_p_i, &Mavg, &Mvar);
  fprintf(stderr, "Mavg %f Mvar %f\n", Mavg, Mvar);

  /* Compute mean and average of mark in the frame image {I}: */

  double Iavg, Ivar;
  image_weighted_avg_var (I, W, subsampling, p_i, &Iavg, &Ivar);
  fprintf(stderr, "Iavg %f Ivar %f\n", Iavg, Ivar);
    
  *alpha = sqrt (Ivar/Mvar);
  if (*alpha < 0.1) {*alpha = 0.1;}
  if (*alpha > 1.0) {*alpha = 1.0;}
    
  *beta = Iavg - (*alpha)*Mavg;
    
  if (*beta < 0.0) {*beta = 0.0;}

  fprintf(stderr, "Estimated alpha = %f beta = %f\n", *alpha, *beta);
}
 
