
#define PROG_NAME "test_calib_guess"
#define PROG_DESC "tests the initial camera calibration guess algorithms"

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <image.h>
#include <image_list.h>
#include <r2x2.h>
#include <r2.h>
#include <pswr.h>
#include <jsfile.h>
#include <tf_calib_refine2.h>
#include <tf_camera.h>
#include <tf_camera_plot.h>
#include <tf_errors.h>
#include <tf_targets.h>
#include <tf_calib.h>
#include <tf_calib_guess1.h>
#include <tf_calib_guess2.h>
#include <sys/stat.h>
#include <sys/types.h>

void tf_write_guess_plots
  ( camera_specs_t cspec,
    calibration_data_t cdat,
    camera_parameters_t affine_cpar,
    camera_parameters_t guess_cpar,
    camera_parameters_t true_cpar,
    char *out_dir
  );

int main (int argc, char *argv[])
{
  char *camera_name = argv[1];
  char *in_dir = argv[2]; /* Directory where to find {p_i.txt,p_wgt.txt,true.cpar}. */
  char *out_dir = argv[3]; /* Directory for all output files. */
  int which_calib_guess = atoi(argv[4]); /* Version of {tf_calib_guessN} to use. */

  fprintf(stderr, "Starting:\n");
  
  /* Get the fixed camera parameters: */
  camera_specs_t cspec;
  
  if (strcmp(camera_name,"optura") == 0) {
    cspec = tf_get_canon_optura_specs ();
  }
  else if (strcmp(camera_name,"svga") == 0) {
    cspec = tf_get_svga_povray_specs ();
  }
  else if (strcmp(camera_name,"hvga") == 0) {
    cspec = tf_get_hvga_povray_specs ();
  }
  else if (strcmp(camera_name,"sony") == 0) {
    cspec = tf_get_sony_dv40_specs ();
  }
  else {
    fprintf(stderr, "error: define the camera specifications\n");
    exit(1);
  }

  /* A priori uncertainty on mark image positions (for error computation): */
  r2_t p_i_dev = (r2_t){{ 5.0, 5.0 }};  /* Should be a parameter... */

  fprintf(stderr, "Generic camera parameters:\n");
  tf_show_camera_specs (cspec, stderr);

  /*Output data directory*/
  mkdir(out_dir, 0777);

  camera_parameters_t true_cpar = tf_get_mean_camera_parameters(cspec); /* True. */

  /* Read the true {cpar}, if present: */
  char *true_cpar_fname = NULL;
  asprintf(&true_cpar_fname, "%s/true.cpar", in_dir);
  FILE *true_cpar_file = fopen(true_cpar_fname, "r");
  if (true_cpar_file != NULL) {
    tf_read_camera_parameters(true_cpar_file, true_cpar);
    fclose(true_cpar_file);
  } else {
    free(true_cpar); true_cpar = NULL;
  }
  free(true_cpar_fname);

  /* Read a set of camera calibration data: */
  char *image_coords_fname = NULL;
  asprintf(&image_coords_fname, "%s/p_i.txt", in_dir);
  char *point_weights_fname = NULL;
  asprintf(&point_weights_fname, "%s/p_wgt.txt", in_dir);
  char *world_coords_fname = NULL;
  asprintf(&world_coords_fname, "%s/p_w.txt", in_dir);
  calibration_data_t cdat = tf_calibration_data_read
    ( world_coords_fname, image_coords_fname, point_weights_fname );
  fprintf(stderr, "calibration data:\n");
  tf_calibration_data_print(cdat, stderr); 
  fprintf(stderr, "-----------------------------\n");

  /* Allocate camera parameter structures: */
  camera_parameters_t affine_cpar = tf_get_mean_camera_parameters(cspec);
  camera_parameters_t guess_cpar = tf_get_mean_camera_parameters(cspec);

  if (which_calib_guess == 1)
    { free(affine_cpar); affine_cpar = NULL;
      tf_calib_guess1_initial_camera_parameters 
        ( cdat, cspec, p_i_dev, guess_cpar, TRUE );
    }
  else if (which_calib_guess == 2)
    { 
      tf_calib_guess2_initial_camera_parameters 
        ( cdat, cspec, affine_cpar, TRUE, TRUE );
      tf_calib_guess2_initial_camera_parameters 
        ( cdat, cspec, guess_cpar, FALSE, TRUE );
    }

  tf_write_cpar_and_errors("true", cdat, cspec, p_i_dev, FALSE, true_cpar, out_dir); 
  tf_write_cpar_and_errors("affine", cdat, cspec, p_i_dev, FALSE, affine_cpar, out_dir); 
  tf_write_cpar_and_errors("guess", cdat, cspec, p_i_dev, FALSE, guess_cpar, out_dir); 

  tf_write_guess_plots(cspec, cdat, affine_cpar, guess_cpar, true_cpar, out_dir);

  tf_calibration_data_free(cdat); 

  tf_free_camera_parameters_structure (guess_cpar);
  tf_free_camera_parameters_structure (true_cpar);

  return 0;
}

void tf_write_guess_plots
  ( camera_specs_t cspec,
    calibration_data_t cdat,
    camera_parameters_t affine_cpar,
    camera_parameters_t guess_cpar,
    camera_parameters_t true_cpar,
    char *out_dir
  )
{
  /* Camera image dimensions and aspect ratio: */
  double Nx = guess_cpar->Npx;
  double Ny = guess_cpar->Npy;
  double aspect = Nx/Ny;

  /* Postscript figure sizeand margin, in pt: */
  double hSize = 400.0*sqrt(aspect);
  double vSize = 400.0/sqrt(aspect);
  double hMrg = 4;
  double vMrg = 3;

  /* Open the output Postscript file: */
  char *ps_prefix = NULL;
  asprintf(&ps_prefix, "%s/", out_dir);
  PSStream *ps = pswr_new_stream(TRUE, ps_prefix, NULL, NULL, hSize + 2*hMrg, vSize + 2*vMrg);

  /* Plot the figures: */
  int nmarks = cdat->nmarks;
  r2_t *p_i = cdat->image_coords;
  r3_t *p_w = cdat->world_coords;

  camera_parameters_t fig1[2] = { true_cpar, affine_cpar };
  camera_parameters_t fig2[2] = { true_cpar, guess_cpar   };

  tf_plot_cameras(ps, "true_affine", nmarks, p_w, 2, fig1, p_i, hSize, vSize, hMrg, vMrg, Nx, Ny, 1.5, 0);
  tf_plot_cameras(ps, "true_guess",  nmarks, p_w, 2, fig2, p_i, hSize, vSize, hMrg, vMrg, Nx, Ny, 1.5, 0);

  /* Finish the postscript file: */
  pswr_close_stream(ps);
}
