
#define PROG_NAME "test_calib_guess"
#define PROG_DESC "tests the initial camera calibration guess algorithms"

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <image.h>
#include <image_list.h>
#include <r2x2.h>
#include <r2.h>
#include <pswr.h>
#include <jsfile.h>
#include <tf_calib_refine2.h>
#include <tf_calib_guess2.h>
#include <tf_camera.h>
#include <tf_camera_plot.h>
#include <tf_errors.h>
#include <tf_targets.h>
#include <tf_calib.h>
#include <sys/stat.h>
#include <sys/types.h>

void tf_write_opt_plots
  ( calibration_data_t cdat,
    camera_parameters_t guess_cpar,
    camera_parameters_t opt1_cpar,
    camera_parameters_t opt2_cpar,
    camera_parameters_t true_cpar,
    char *out_dir
  );

void get_cdat_p_i_from_cpar
  (
    calibration_data_t cdat,
    camera_parameters_t true_cpar
  );

int main (int argc, char *argv[])
{
  char *camera_name = argv[1];
  char *in_dir = argv[2]; /* Directory where to find {p_i.txt,p_w.txt,p_wgt.txt,true.cpar}. */
  char *guess_cpar_name = argv[3];
  char *out_dir = argv[4]; /* Directory for all output files. */

  /* Shall we include the {cpar} Kalman error terms in the optimization? */
  bool_t use_cpar_dev_terms = FALSE;

  fprintf(stderr, "Starting:\n");
  
  /* Get the fixed camera parameters: */
  camera_specs_t cspec;
  
  if (strcmp(camera_name,"optura") == 0) {
    cspec = tf_get_canon_optura_specs ();
  }
  else if (strcmp(camera_name,"svga") == 0) {
    cspec = tf_get_svga_povray_specs ();
  }
  else if (strcmp(camera_name,"hvga") == 0) {
    cspec = tf_get_hvga_povray_specs ();
  }
  else if (strcmp(camera_name,"sony") == 0) {
    cspec = tf_get_sony_dv40_specs ();
  }
  else {
    fprintf(stderr, "error: define the camera specifications\n");
    exit(1);
  }

  fprintf(stderr, "Generic camera parameters:\n");
  tf_show_camera_specs (cspec, stderr);

  /* A priori uncertainty on mark image positions (for error computation): */
  r2_t p_i_dev = (r2_t){{ 1.0, 1.0 }};  /* Should be a parameter... */

  /*Output data directory*/
  mkdir(out_dir, 0777);

  /* Read a set of camera calibration data: */
  char *image_coords_fname = NULL;
  asprintf(&image_coords_fname, "%s/p_i.txt", in_dir);
  char *point_weights_fname = NULL;
  asprintf(&point_weights_fname, "%s/p_wgt.txt", in_dir);
  char *world_coords_fname = NULL;
  asprintf(&world_coords_fname, "%s/p_w.txt", in_dir);
  calibration_data_t cdat = tf_calibration_data_read
    ( world_coords_fname, image_coords_fname, point_weights_fname );
  fprintf(stderr, "calibration data:\n");
  tf_calibration_data_print(cdat, stderr); 
  fprintf(stderr, "-----------------------------\n");

  int nmarks = cdat->nmarks;
  r2_t *p_i = cdat->image_coords;
  r3_t *p_w = cdat->world_coords;
  double *p_wgt = cdat->weight;

  /* Read the true {cpar}, if present: */
  camera_parameters_t true_cpar = NULL; /* True. */
  char *true_cpar_fname = NULL;
  asprintf(&true_cpar_fname, "%s/true.cpar", in_dir);
  FILE *true_cpar_file = fopen(true_cpar_fname, "r");
  if (true_cpar_file != NULL) {
    true_cpar = tf_new_camera_parameters();
    tf_read_camera_parameters(true_cpar_file, true_cpar);
    fclose(true_cpar_file);
  }
  free(true_cpar_fname);

  tf_write_cpar_and_errors("true", cdat, cspec, p_i_dev, use_cpar_dev_terms, true_cpar, out_dir); 
  
  /* Read the guess {cpar}: */
  FILE *f1_cpars = fopen(guess_cpar_name, "r");
  camera_parameters_t guess_cpar = tf_new_camera_parameters();
  tf_read_camera_parameters (f1_cpars, guess_cpar);
  fclose(f1_cpars);

  tf_write_cpar_and_errors("guess", cdat, cspec, p_i_dev, use_cpar_dev_terms, guess_cpar, out_dir); 

  bool_t debug = TRUE;

  // !!! Should copy {guess_cpar} instead of reading the file again.
  camera_parameters_t opt1_cpar = tf_copy_camera_parameters(guess_cpar);

  /* First optimization pass: */
  tf_calib_generic
    ( cdat, cspec, p_i_dev, use_cpar_dev_terms, opt1_cpar,
      tf_calib_refine2_gather_optimization_params,
      tf_calib_refine2_scatter_optimization_params,
      debug ); 

  tf_write_cpar_and_errors("opt1", cdat, cspec, p_i_dev, use_cpar_dev_terms, opt1_cpar, out_dir); 

  /* Adjust weights based on position discrepancies: */
  /* Estimate the standard deviation {dev_bad} of outliers: */
  r2_t dev_bad;
  double cook1 = 5.0; /* Guessed ratio of outlier-dev/inlier-dev. */
  r2_scale(cook1, &p_i_dev, &dev_bad);
  /* Estimate the standard deviation {dev_gud} of inliers: */
  r2_t dev_gud; /* Standard deviation for inliers. */
  double cook2 = 0.10; /* Guessed relative effect of outliers on predicted {p_i}s. */
  r2_mix(1.0, &p_i_dev, cook2, &dev_bad, &dev_gud);
  tf_recompute_target_weights(opt1_cpar, nmarks, p_w, p_i, p_wgt, dev_gud, dev_bad, TRUE);

  /* Second optimization pass: */
  camera_parameters_t opt2_cpar = tf_copy_camera_parameters(opt1_cpar);
  tf_calib_generic
    ( cdat, cspec, p_i_dev, use_cpar_dev_terms, opt2_cpar,
      tf_calib_refine2_gather_optimization_params,
      tf_calib_refine2_scatter_optimization_params,
      debug ); 

  tf_write_cpar_and_errors("opt2", cdat, cspec, p_i_dev, use_cpar_dev_terms, opt2_cpar, out_dir); 

  tf_write_opt_plots(cdat, guess_cpar, opt1_cpar, opt2_cpar, true_cpar, out_dir);

  tf_calibration_data_free(cdat); 

  tf_free_camera_parameters_structure (opt1_cpar);
  tf_free_camera_parameters_structure (opt2_cpar);
  tf_free_camera_parameters_structure (true_cpar);
  tf_free_camera_parameters_structure (guess_cpar);

  return 0;
}

void tf_write_opt_plots
  ( calibration_data_t cdat,
    camera_parameters_t guess_cpar,
    camera_parameters_t opt1_cpar,
    camera_parameters_t opt2_cpar,
    camera_parameters_t true_cpar,
    char *out_dir
  )
{
  /* Camera image dimensions and aspect ratio: */
  double Nx = guess_cpar->Npx;
  double Ny = guess_cpar->Npy;
  double aspect = Nx/Ny;

  /* Postscript figure sizeand margin, in pt: */
  double hSize = 400.0*sqrt(aspect);
  double vSize = 400.0/sqrt(aspect);
  double hMrg = 4;
  double vMrg = 3;

  /* Output Postscript file: */
  char *ps_prefix = NULL;
  asprintf(&ps_prefix, "%s/", out_dir);
  PSStream *ps = pswr_new_stream(TRUE, ps_prefix, NULL, NULL, hSize + 2*hMrg, vSize + 2*vMrg);

  /* Plot the figures: */
  int nmarks = cdat->nmarks;
  r2_t *p_i = cdat->image_coords;
  r3_t *p_w = cdat->world_coords;

  camera_parameters_t fig1[2] = { true_cpar, guess_cpar };
  camera_parameters_t fig2[2] = { true_cpar, opt1_cpar  };
  camera_parameters_t fig3[2] = { true_cpar, opt2_cpar  };

  tf_plot_cameras(ps, "true_guess",  nmarks, p_w, 2, fig1, p_i, hSize, vSize, hMrg, vMrg, Nx, Ny, 1.5, 0);
  tf_plot_cameras(ps, "true_opt1",   nmarks, p_w, 2, fig2, p_i, hSize, vSize, hMrg, vMrg, Nx, Ny, 1.5, 0);
  tf_plot_cameras(ps, "true_opt2",   nmarks, p_w, 2, fig3, p_i, hSize, vSize, hMrg, vMrg, Nx, Ny, 1.5, 0);

  pswr_close_stream(ps);
}
