#include "tf_calib_data.h"
#include <affirm.h>

calibration_data_t tf_calibration_data_read (char *p_w_fname, char *p_i_fname, char *wgt_fname)
{
  int n_p_w, n_p_i, n_wgt;
    
  calibration_data_t cdat = (calibration_data_t)notnull(malloc(sizeof(struct _calibration_data_t)), "no mem");
  tf_read_world_positions (p_w_fname, &n_p_w, &(cdat->world_coords));
  tf_read_image_positions (p_i_fname, &n_p_i, &(cdat->image_coords));
  demand(n_p_i == n_p_w, "inconsistent mark counts (world_coords vs. image_coords)");
  if (wgt_fname != NULL) {
    /* read confidence weights from file */
    tf_read_weights (wgt_fname, &n_wgt, &(cdat->weight));
    demand(n_wgt == n_p_w, "inconsistent mark counts (world_coords vs. weights)");
  } 
  else {
    /* assume unit weight for all marks */
    cdat->weight = (double *)notnull(malloc(n_p_w * sizeof(double)), "no mem");
    int k;
    for (k = 0; k < n_p_w; k++) { cdat->weight[k] = 1.0; }
  }
  cdat->nmarks = n_p_w;
  return cdat; 
}

void tf_calibration_data_print (calibration_data_t cdat, FILE *ferr)
{
    int i;
    fprintf(ferr, "--------- calibration data ---------------------------------\n");
    for (i = 0; i < cdat->nmarks; i++) {
        r3_t p_w = cdat->world_coords[i];
        r2_t p_i = cdat->image_coords[i];
        double wgt = cdat->weight[i];
        fprintf(ferr, "[%03d]", i);
        fprintf(ferr, "  p_w = ( %14.6f %14.6f %14.6f )", p_w.c[0], p_w.c[1], p_w.c[2]);
        fprintf(ferr, "  p_i = ( %9.3f %9.3f )", p_i.c[0], p_i.c[1]);
        fprintf(ferr, "  wgt = %11.9f\n", wgt);
    }
    fprintf(ferr, "------------------------------------------------------------\n");
}

void tf_2D_points_print (int nmarks, r2_t p[], FILE *ferr)
{
    int i;
    fprintf(ferr, "--------- point coordinates --------------------------------\n");
    for (i = 0; i < nmarks; i++) {
        r2_t pi = p[i];
        fprintf(ferr, "[%03d]", i);
        fprintf(ferr, "  p = ( %9.3f %9.3f )\n", pi.c[0], pi.c[1]);
    }
    fprintf(ferr, "------------------------------------------------------------\n");
}

void tf_read_world_positions (char *fname, int *nmarks, r3_t **world_coords)
{
  int i, n;
  FILE *f = open_read(fname, TRUE);
  if (f == NULL) {
    fprintf(stderr, "error: the file %s doesn't exist\n", fname);
    exit(1);
  }
  int ok = fscanf(f, "%d ", &n);
  if ((ok != 1) || (n < 1) || (n > MAX_POINTS)) {
    fprintf(stderr, "%s", "error: number of marks missing or invalid\n");
    exit (-1);
  } 
  (*world_coords) = (r3_t *)notnull(malloc(n * sizeof(r3_t)),"no mem"); 
  for (i = 0; i < n; i++)
    {
      fscanf(f, "%lf ", &((*world_coords)[i].c[0]));
      fscanf(f, "%lf ", &((*world_coords)[i].c[1]));
      fscanf(f, "%lf ", &((*world_coords)[i].c[2]));
      tf_calib_data_skip_comment (f);
    }
  fclose(f);
  (*nmarks) = n;
}

void tf_read_image_positions (char *fname, int *nmarks, r2_t **image_coords)
{
  int i, n;
  FILE *f = open_read(fname, TRUE);
  if (f == NULL) {
    fprintf(stderr, "error: the file %s doesn't exist\n", fname);
    exit(1);
  }
  int ok = fscanf(f, "%d ", &n);
  if ((ok != 1) || (n < 1) || (n > MAX_POINTS)) {
    fprintf(stderr, "error: number of marks missing or invalid\n");
    exit (-1);
  } 
  (*image_coords) = (r2_t *)notnull(malloc(n * sizeof(r2_t)),"no mem"); 
  for (i = 0; i < n; i++)
    { fscanf(f, "%lf ", &((*image_coords)[i].c[0]));
      fscanf(f, "%lf ", &((*image_coords)[i].c[1]));
      tf_calib_data_skip_comment (f);
    }
  fclose(f);
  (*nmarks) = n;
}

void tf_read_weights (char *fname, int *nmarks, double **weight)
{
  int i, n;
  FILE *f = open_read(fname, TRUE);
  if (f == NULL) {
    fprintf(stderr, "error: the file %s doesn't exist\n", fname);
    exit(1);
  }
  int ok = fscanf(f, "%d ", &n);
  if ((ok != 1) || (n < 1) || (n > MAX_POINTS)) {
    fprintf(stderr, "error: number of marks missing or invalid\n");
    exit (-1);
  } 
  (*weight) = (double *)notnull(malloc(n * sizeof(double)),"no mem");  
  for (i = 0; i < n; i++)
    { 
      fscanf(f, "%lf ", &((*weight)[i]));
      tf_calib_data_skip_comment (f);
    }
  fclose(f);
  (*nmarks) = n;
}

void tf_print_weights (int ntargets, double wgts[], FILE *ferr)
{
    int i;
    fprintf(ferr, "%s\n", "Weights:");
    for (i = 0; i < ntargets; i++) {
        fprintf(ferr, "%9.9f \n", (wgts == NULL ? 1.0 : wgts[i]));
    }
    fprintf(ferr, "\n");
}

void  tf_calib_data_skip_comment (FILE *f)
{
  int ch;
  while ((ch=fgetc(f)) == ' ') { }
  if (ch == '#') { while (((ch=fgetc(f)) != '\n') && (ch != EOF)) { } }
  if (ch != EOF) { ungetc(ch, f);}
}

void tf_calib_set_weights(calibration_data_t cdat, double weight[])
{
  int k;
  for (k = 0; k < cdat->nmarks; k++) {
    cdat->weight[k] = weight[k];
  }
}

void tf_calib_set_weights_uniform(calibration_data_t cdat)
{
  int k;
  for (k = 0; k < cdat->nmarks; k++) {
    cdat->weight[k] = 1.0;
  }
}

void tf_calibration_data_free (calibration_data_t cdat)
{
    free(cdat->world_coords);
    free(cdat->image_coords);
    free(cdat->weight);
    free(cdat);
}


