#ifndef tf_calib_data_H
#define tf_calib_data_H

/* General-purpose data and functions for calibration routines. */
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <jsfile.h>
#include <r2.h>
#include <r3.h>

/* Max number of world-image data pairs: */
#define MAX_POINTS 500

/* ---------------------------------------------------------------------- */
/* CALIBRATION DATA STRUCTURE */

typedef struct _calibration_data_t {
    int    nmarks;             /* number of marks */
    r3_t   *world_coords;      /* world coordinates of each mark [mm] */
    r2_t   *image_coords;      /* observed image coordinates of each mark [pixels] */
    double *weight;            /* confidence weight of (world,image) data pair. */
} *calibration_data_t;
  /* A data set for weighted camera calibration. */

/* ---------------------------------------------------------------------- */
/* CALIBRATION DATA I/O */

calibration_data_t tf_calibration_data_read (char *p_w_fname, char *p_i_fname, char *wgt_fname);
/* Reads the world coordinates, image coordinates, and confidence
  weights of the marks from files {p_w_fname}, {p_i_fname}, and
  {wgt_name}, respectively.  If {wgt_name} is NULL, sets all weights
  to 1. 

  In each file, the first line must be the number of marks {nmarks}.
  The following lines must contain the mark data, one mark per line,
  consisting of three, two, and one real numbers, respectively.
  The line may end optionally in a '#' followed by an arbitrary comment. */

void tf_calibration_data_free(calibration_data_t cdat);
/* Reclaims the record {*cdat} and its internal vectors. */

void tf_calibration_data_print (calibration_data_t cdat, FILE *ferr);
/* Prints all the parameters of the calibration data structure. */

void tf_2D_points_print (int nmarks, r2_t p[], FILE *ferr);
/* Prints {p[0..nmarks-1]} to {ferr}. */

void tf_read_world_positions (char *fname, int *nmarks, r3_t **world_coords);
/* Reads a file containing a number of marks {*nmarks} followed by {*nmarks}
  world coordinates {(*world_coords)[0..*nmarks-1]}, each being three real numbers. 
  Allocates the vector {*world_coords}. */

void tf_read_image_positions (char *fname, int *nmarks, r2_t **image_coords);
/* Reads a file containing a number of marks {*nmarks} followed by {*nmarks}
  image coordinates {(*image_coords)[0..*nmarks-1]}, each being two real numbers. 
  Allocates the vector {*image_coords}. */

/* ---------------------------------------------------------------------- */
/* DATA PAIR WEIGHTS */

void tf_read_weights (char *fname, int *nmarks, double **weight);
/* Reads a file containing a number of marks {*nmarks} followed by {*nmarks}
  image coordinates {(*weight)[0..*nmarks-1]}, each being one real number. 
  Allocates the vector {*weight}. */

void tf_print_weights (  int ntargets, double wgts[], FILE *ferr  );
/* This routine print the weights. */

void tf_calib_set_weights(calibration_data_t cdat, double weight[]);
/* Sets {cdat->weight[k] = weight[k]} for all {k}. */

void tf_calib_set_weights_uniform(calibration_data_t cdat);
/* Sets {cdat->weight[k] = 1} for all {k}. */

void  tf_calib_data_skip_comment (FILE *f);
/* Skips spaces until the next EOF or non-space character,
   if that character is '#' skips it and everthing else
   on that line (except the final '\n'). */

#endif
